/*
 * SANE - Scanner Access Now Easy.
 * coolscan9k.c
 *
 * This file implements a SANE backend for Nikon Coolscan film scanners.
 *
 * coolscan9k.c is based on coolscan3.c, Copyright:
 *
 * Copyright (C) 2007-08 Tower Technologies
 * Author: Alessandro Zummo <a.zummo@towertech.it>
 *
 * Copyright (C) 2008 J. Robert Sims, III
 *
 * LS-9000 ED added by Rob Sims <sane@robsims.com>, June 2008
 *   - Support both RGBI and two-pass infrared handling
 *   - Improved use of capability bits to mask options
 *   - Support for non-static device parameters (changeable
 *     film holders)
 *   - Remove a lot of miniscule mallocs
 *   - Support "thumbnail" mode where entire holder content is
 *     previewed
 *   - Exposure controls now in stops rather than linear
 *   - Support for user defineable frame locations (primarily
 *     intended for medium format support)
 *   - Support for frame grids rather than linear-only layout;
 *     some holders have parallel strips
 *   - A lot noisier in debug output
 *   - Support for non 8-bit capable scanners
 *   - Gather exposure info from inside multiple frames
 *   - ADF mode supports full-frame or partial frame scans; selection on
 *     thumbnail is translated.
 *   - Support for exposure analysis.
 *   - Support for infrared exposure analysis.
 *
 * This file is part of the SANE package.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

/* ========================================================================= */

#include <sane/config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>

#include <sane/sane.h>
#include <sane/sanei.h>
#include <sane/saneopts.h>
#include <sane/sanei_scsi.h>
#include <sane/sanei_usb.h>
#include <sane/sanei_debug.h>
#include <sane/sanei_config.h>

#define BACKEND_NAME coolscan9k
#include <sane/sanei_backend.h>	/* must be last */

#define CS9K_VERSION_MAJOR 0
#define CS9K_VERSION_MINOR 5
#define CS9K_REVISION 0
#define CS9K_CONFIG_FILE "coolscan9k.conf"

#define WSIZE (sizeof (SANE_Word))

#define CS9K_HOLDER_NAME_MAX     (30)
#define CS9K_NEXTSCAN_MAX        (64)

/* ========================================================================= */
/* typedefs */

typedef enum
{
  CS9K_TYPE_UNKOWN,
  CS9K_TYPE_LS_X,
  CS9K_TYPE_LS30,
  CS9K_TYPE_LS40,
  CS9K_TYPE_LS50,
  CS9K_TYPE_LS2000,
  CS9K_TYPE_LS4000,
  CS9K_TYPE_LS5000,
  CS9K_TYPE_LS8000,
  CS9K_TYPE_LS9000
}
cs9k_type_t;

typedef enum
{
  CS9K_INTERFACE_UNKNOWN,
  CS9K_INTERFACE_SCSI,		/* includes IEEE1394 via SBP2 */
  CS9K_INTERFACE_USB
}
cs9k_interface_t;

typedef enum
{
  CS9K_PHASE_NONE = 0x00,
  CS9K_PHASE_STATUS = 0x01,
  CS9K_PHASE_OUT = 0x02,
  CS9K_PHASE_IN = 0x03,
  CS9K_PHASE_BUSY = 0x04
}
cs9k_phase_t;

typedef enum
{
  CS9K_INQ_SAME_VALUE = 0,
  CS9K_INQ_NEW_VALUE = 1,
  CS9K_INQ_BAD_OFFSET = 1 << 7
}
cs9k_inq_val_t;

typedef enum			/* static hardware capabilities and properties */
{
  CS9K_CAPS_PAD_BLOCK = 1,
  CS9K_CAPS_PAD_WORD = 1 << 1,
  CS9K_CAPS_HOLDER = 1 << 2,
  CS9K_CAPS_NEGATIVE = 1 << 3,
  CS9K_CAPS_INDEP_XY = 1 << 4,
  CS9K_CAPS_CTL_VS10 = 1 << 5
}
cs9k_caps_t;

typedef enum
{
  CS9K_COORD_BASE_RES_CONTINUOUS = 0,
  CS9K_COORD_BASE_RES_PITCH = 1,
  CS9K_COORD_BASE_RES_LINEGAP = 2,
  CS9K_COORD_BASE_RES_MASK = 3,

  CS9K_COORD_BASE_X_REVERSED = 1 << 2,
  CS9K_COORD_BASE_Y_REVERSED = 1 << 3,
  CS9K_COORD_BASE_THUMB_REVERSED = 1 << 4,
  CS9K_COORD_BASE_RESERVED = 1 << 5,
  CS9K_COORD_BASE_ADDL_COORD = 1 << 6,
  CS9K_COORD_BASE_EXTEND = 1 << 7
}
cs9k_coord_base_t;

typedef enum
{
  CS9K_SCANKIND_IMAGE = 1,
  CS9K_SCANKIND_THUMBNAIL = 1 << 1,
  CS9K_SCANKIND_PRESCAN = 1 << 2,
  CS9K_SCANKIND_RESERVED1 = 1 << 3,
  CS9K_SCANKIND_RESERVED2 = 1 << 4,
  CS9K_SCANKIND_AESCAN = 1 << 5,
  CS9K_SCANKIND_AEWBSCAN = 1 << 6,
  CS9K_SCANKIND_EXTEND = 1 << 7
}
cs9k_skind_t;

typedef enum
{
  CS9K_SCANMODE_HIGHQUAL = 1,
  CS9K_SCANMODE_NORMQUAL = 1 << 1,
  CS9K_SCANMODE_HIGHSPEED = 1 << 2,
  CS9K_SCANMODE_RESERVED1 = 1 << 3,
  CS9K_SCANMODE_MULTISCAN = 1 << 4,
  CS9K_SCANMODE_RESERVED2 = 1 << 5,
  CS9K_SCANMODE_REVERSESCAN = 1 << 6,
  CS9K_SCANMODE_EXTEND = 1 << 7
}
cs9k_smode_t;

typedef enum
{
  CS9K_CLRCMP_GRAYSCALE = 1,
  CS9K_CLRCMP_DROPOUT = 1 << 1,
  CS9K_CLRCMP_RGB = 1 << 2,
  CS9K_CLRCMP_CMY = 1 << 3,
  CS9K_CLRCMP_RESERVED1 = 1 << 4,
  CS9K_CLRCMP_RESERVED2 = 1 << 5,
  CS9K_CLRCMP_RESERVED3 = 1 << 6,
  CS9K_CLRCMP_EXTEND = 1 << 7
}
cs9k_clrcmp_t;

typedef enum
{
  CS9K_COOP_THUMBNAIL = 1 << 8,
  CS9K_COOP_AVG_MULT = 1 << 9,
  CS9K_COOP_REG_GAP = 1 << 10,
  CS9K_COOP_DARK_DATA = 1 << 11,
  CS9K_COOP_SHADE_CAL = 1 << 12,
  CS9K_COOP_AUTOFOCUS = 1 << 13,
  CS9K_COOP_SHADE_CORR = 1 << 14,
  CS9K_COOP_MULTI_LINE = 1 << 0,
  CS9K_COOP_PITCH_MAIN = 1 << 1,
  CS9K_COOP_TRUNCATE = 1 << 2,
  CS9K_COOP_CCD_DATA = 1 << 3
}
cs9k_host_coop_t;

typedef enum
{
  CS9K_EXE_80_RESET = 1 << 8,
  CS9K_EXE_81_GOTO_ORIGIN = 1 << 9
}
cs9k_exe_8x_t;

typedef enum
{
  CS9K_EXE_90_CHANGE_UNIT = 1 << 8,
  CS9K_EXE_91_AF_AUTO_EXECUTE = 1 << 9,
  CS9K_EXE_92_CAL_AUTO_EXE = 1 << 10
}
cs9k_exe_9x_t;

typedef enum
{
  CS9K_EXE_A0_AF = 1 << 8,
  CS9K_EXE_A1_COLOR_AF = 1 << 9
}
cs9k_exe_ax_t;

typedef enum
{
  CS9K_EXE_B0_SETUP_SHADING = 1 << 8,
  CS9K_EXE_B1_DARK_I_CORR = 1 << 9,
  CS9K_EXE_B2_OFFSET_CORR = 1 << 10,
  CS9K_EXE_B3_UNIT_SPECIFIC = 1 << 11,
  CS9K_EXE_B4_AUTO_UNLOAD_TIME = 1 << 12
}
cs9k_exe_bx_t;

typedef enum
{
  CS9K_EXE_C0_STAGE_MOVE = 1 << 8,
  CS9K_EXE_C1_FOCUS_MOVE = 1 << 9
}
cs9k_exe_cx_t;


typedef enum
{
  CS9K_EXE_D0_UNLOAD = 1 << 8,
  CS9K_EXE_D1_LOAD = 1 << 9,
  CS9K_EXE_D2_ABS_POS = 1 << 10,
  CS9K_EXE_D3_REL_POS = 1 << 11,
  CS9K_EXE_D4_ROTATE = 1 << 12
}
cs9k_exe_dx_t;

typedef enum
{
  CS9K_SENDREAD_02_READ_HALFTONE = 1 << 16,
  CS9K_SENDREAD_02_WRITE_HALFTONE = 1 << 17,
  CS9K_SENDREAD_03_READ_GAMMA_FUNC = 1 << 18,
  CS9K_SENDREAD_03_WRITE_GAMMA_FUNC = 1 << 19,
  CS9K_SENDREAD_80_READ_HISTOGRAM = 1 << 20,
  CS9K_SENDREAD_81_READ_MAX_VALUE = 1 << 21,
  CS9K_SENDREAD_82_READ_MATRIX = 1 << 8,
  CS9K_SENDREAD_82_WRITE_MATRIX = 1 << 9,
  CS9K_SENDREAD_83_READ_FILTER = 1 << 10,
  CS9K_SENDREAD_83_WRITE_FILTER = 1 << 11,
  CS9K_SENDREAD_84_READ_SHADING = 1 << 12,
  CS9K_SENDREAD_84_WRITE_SHADING = 1 << 13,
  CS9K_SENDREAD_85_READ_DARK_V = 1 << 0,
  CS9K_SENDREAD_85_WRITE_DARK_V = 1 << 1,
  CS9K_SENDREAD_86_READ_MAGNETIC = 1 << 2,
  CS9K_SENDREAD_86_WRITE_MAGNETIC = 1 << 3,
  CS9K_SENDREAD_87_READ_COOP = 1 << 4,
  CS9K_SENDREAD_88_READ_BOUNDARY = 1 << 5,
  CS9K_SENDREAD_88_WRITE_BOUNDARY = 1 << 6
}
cs9k_sendread_cap1_t;

typedef enum
{
  CS9K_SENDREAD_89_READ_ANALOG_GAMMA = 1 << 8,
  CS9K_SENDREAD_8A_READ_ANALOG_GAIN = 1 << 9,
  CS9K_SENDREAD_8B_READ_DIGITAL_GAIN = 1 << 10,
  CS9K_SENDREAD_8C_READ_WB_EXPOSURE = 1 << 11,
  CS9K_SENDREAD_8D_READ_SETUP = 1 << 12,
  CS9K_SENDREAD_8D_WRITE_SETUP = 1 << 13,
  CS9K_SENDREAD_8E_READ_PERFORATION = 1 << 14,
  CS9K_SENDREAD_8F_READ_BOUNDARY2 = 1 << 0,
  CS9K_SENDREAD_8F_WRITE_BOUNDARY2 = 1 << 1,
  CS9K_SENDREAD_90_READ_INIT_WB_EXPOSURE = 1 << 2,
  CS9K_SENDREAD_91_READ_CCD = 1 << 3,
  CS9K_SENDREAD_92_READ_DRIVER_VER = 1 << 4,
  CS9K_SENDREAD_92_WRITE_DRIVER_VER = 1 << 5,
  CS9K_SENDREAD_93_READ_LEAK = 1 << 6
}
cs9k_sendread_cap2_t;

typedef enum
{
  CS9K_SCAN_NORMAL,
  CS9K_SCAN_AE,
  CS9K_SCAN_AE_WB
}
cs9k_scan_t;

typedef enum
{
  CS9K_INFRARED_OFF,
  CS9K_INFRARED_4PLANE,
  CS9K_INFRARED_IN,
  CS9K_INFRARED_OUT,
  CS9K_INFRARED_ONLY
}
cs9k_infrared_t;

typedef enum
{
  CS9K_THUMB_SEL_NOFRAME,
  CS9K_THUMB_SEL_INFRAME,
  CS9K_THUMB_SEL_CROSSFRAME
}
cs9k_thumb_sel_descr_t;

typedef enum
{
  CS9K_STATUS_READY = 0,
  CS9K_STATUS_BUSY = 1,
  CS9K_STATUS_NO_DOCS = 2,
  CS9K_STATUS_PROCESSING = 4,
  CS9K_STATUS_ERROR = 8,
  CS9K_STATUS_REISSUE = 16,
  CS9K_STATUS_ALL = 31		/* sum of all others */
}
cs9k_status_t;

typedef enum
{
  CS9K_FBF_EXAMINED = 1,
  CS9K_FBF_PRESENT = 1 << 1
}
cs9k_frame_bounds_flags_t;

typedef enum
{
  CS9K_HIST_ALG_NONE = 0,
  CS9K_HIST_ALG_UNDEREXPOSED,
  CS9K_HIST_ALG_HIGH_RANGE,
  CS9K_HIST_ALG_HIGH_SAT,
  CS9K_HIST_ALG_HEDGE_DET,
  CS9K_HIST_ALG_LEDGE_DET,
  CS9K_HIST_ALG_MULTI_FRAME
}
cs9k_hist_alg_t;

typedef struct
{
  SANE_Int offset_x;
  SANE_Int offset_y;
  SANE_Int width;
  SANE_Int height;
  SANE_Fixed adj_exp;
  SANE_Fixed exposure;
  cs9k_hist_alg_t adj_exp_type;
  cs9k_frame_bounds_flags_t flags;
}
cs9k_frame_bounds_t;

typedef struct
{
  SANE_String name;
  SANE_Int sf_origin_x;
  SANE_Int sf_origin_y;
  SANE_Int sf_stride_x;
  SANE_Int sf_stride_y;
  SANE_Int sf_width;
  SANE_Int sf_height;
}
cs9k_frame_preset_t;

typedef enum
{
  CS9K_OPTION_NUM = 0,

  CS9K_OPTION_PREVIEW,

  CS9K_OPTION_DEPTH,

  CS9K_OPTION_GROUP_INFRARED,
  CS9K_OPTION_INFRARED,
  CS9K_OPTION_INFRARED_FMT,

  CS9K_OPTION_GROUP_EXPOSURE,
  CS9K_OPTION_NEGATIVE,
  CS9K_OPTION_EXPOSURE,
  CS9K_OPTION_EXPOSURE_R,
  CS9K_OPTION_EXPOSURE_G,
  CS9K_OPTION_EXPOSURE_B,
  CS9K_OPTION_EXPOSURE_I,
  CS9K_OPTION_EXPOSURE_PER_FRAME,
  CS9K_OPTION_ADJUST_EXPOSURE_TEXT,
  CS9K_OPTION_ADJUST_EXPOSURE,
  CS9K_OPTION_ADJUST_EXPOSURE_TEXT_IR,
  CS9K_OPTION_ADJUST_EXPOSURE_IR,
  CS9K_OPTION_ADJUST_EXPOSURE_TARGET,
  CS9K_OPTION_ADJUST_EXPOSURE_TARGET_IR,
  CS9K_OPTION_SCAN_AE,
  CS9K_OPTION_SCAN_AE_WB,

  CS9K_OPTION_GROUP_LUT,
  CS9K_OPTION_LUT_R,
  CS9K_OPTION_LUT_G,
  CS9K_OPTION_LUT_B,

  CS9K_OPTION_GROUP_RES,
  CS9K_OPTION_RES_INDEPENDENT,
  CS9K_OPTION_RES,
  CS9K_OPTION_RESY,
  CS9K_OPTION_RESX,
  CS9K_OPTION_PREVIEW_RESOLUTION,
  CS9K_OPTION_THUMB_RESOLUTION,

  CS9K_OPTION_GROUP_FRAME,
  CS9K_OPTION_SOURCE,
  CS9K_OPTION_HOLDER,
  CS9K_OPTION_NEXTSCAN,
  CS9K_OPTION_SCAN_THUMBNAILS,
  CS9K_OPTION_FRAME,
  CS9K_OPTION_SUBFRAME_X,
  CS9K_OPTION_SUBFRAME_Y,
  CS9K_OPTION_XMIN,
  CS9K_OPTION_XMAX,
  CS9K_OPTION_YMIN,
  CS9K_OPTION_YMAX,

  CS9K_OPTION_SET_FRAME_GROUP,
  CS9K_OPTION_SET_FRAME_USER,
  CS9K_OPTION_SET_FRAME_ORIGIN_X,
  CS9K_OPTION_SET_FRAME_ORIGIN_Y,
  CS9K_OPTION_SET_FRAME_STRIDE_X,
  CS9K_OPTION_SET_FRAME_STRIDE_Y,
  CS9K_OPTION_SET_FRAME_WIDTH,
  CS9K_OPTION_SET_FRAME_HEIGHT,
  CS9K_OPTION_SET_FRAME_PRESET,

  CS9K_OPTION_GROUP_ACTION,
  CS9K_OPTION_REFRESH,
  CS9K_OPTION_LOAD,
  CS9K_OPTION_AUTOLOAD,
  CS9K_OPTION_EJECT,
  CS9K_OPTION_RESET,

  CS9K_OPTION_GROUP_FOCUS,
  CS9K_OPTION_AUTOFOCUS,
  CS9K_OPTION_FOCUS,
  CS9K_OPTION_FOCUS_ON_CENTRE,
  CS9K_OPTION_FOCUSX,
  CS9K_OPTION_FOCUSY,

  CS9K_N_OPTIONS		/* must be last -- counts number of enum items */
}
cs9k_option_t;

typedef int cs9k_pixel_t;

#define HIST_STEPS (20)
#define HIST_FRAME_MARGIN (10)
#define CS9K_COLOR_MAX 10	/* 9 + 1, see cs9k_colors */
#define CS9K_ADJUST_EXPOSURE_STRING_MAX (60)

typedef struct
{
  /* interface */
  cs9k_interface_t interface;
  int fd;
  SANE_Byte *send_buf, *recv_buf;
  size_t send_buf_size, recv_buf_size;
  size_t n_cmd, n_send, n_recv;

  /* static device characteristics */
  char vendor_string[9], product_string[17], revision_string[5];
  cs9k_type_t type;
  SANE_Int maxbits;
  SANE_Int *depth_list;
  SANE_Int resx_max, *resx_list;
  SANE_Int resy_max, *resy_list;
  SANE_Int res_thumb, *res_thumb_list;
  SANE_Word host_coop, scankinds, scanmodes, scancolors, coord_base_info;
  cs9k_exe_8x_t exe_8x;
  cs9k_exe_9x_t exe_9x;
  cs9k_exe_ax_t exe_ax;
  cs9k_exe_bx_t exe_bx;
  cs9k_exe_cx_t exe_cx;
  cs9k_exe_dx_t exe_dx;
  cs9k_sendread_cap1_t sendread_cap1;
  cs9k_sendread_cap2_t sendread_cap2;
  SANE_Int unit_dpi;
  double unit_mm;
  SANE_Word caps;
  SANE_Int focus_min, focus_max;
  SANE_Int ccd_width, linegapcount;
  SANE_Int exposure_min, exposure_max;
  double exposure_zero;

  /* dynamic device characteristics */

  SANE_Int source, thumb_sel_descr, mode;
  SANE_Int holder_id;
  char holder_string[CS9K_HOLDER_NAME_MAX];
  SANE_Int swax_min, swax_max;	/* Limits on frame setup */
  SANE_Int sway_min, sway_max;
  SANE_Int framex_pos, framey_pos, frame_width, frame_height;
  cs9k_frame_bounds_t *frame_bounds;
  SANE_Int n_frames, frame_bounds_size;
  SANE_Bool medium_changed;

  /* settings */
  SANE_Bool preview, negative, autoload, autofocus, ae, aewb, thumbnail,
    thumbnail_done;
  SANE_Int thumb_pos_in, thumb_pos_out;
  SANE_Word infrared, infrared_fmt;
  SANE_Char infrared_txt[10];
  SANE_Char exposure_txt[20];
  SANE_Int infrared_index;
  SANE_Int depth, real_depth, bytes_per_pixel, shift_bits, n_colors_in,
    n_colors_out;
  cs9k_pixel_t n_lut, n_lut_out;
  cs9k_pixel_t *lut_r, *lut_g, *lut_b, *lut_neutral;
  SANE_Int resx, resy, res;
  SANE_Bool res_independent;
  SANE_Int res_preview;
  SANE_Int xmin, xmax, ymin, ymax;
  int i_frame, c_frame, t_frame;
  double subframe_x, subframe_y;

  SANE_Int real_resx, real_resy, real_pitchx, real_pitchy;
  SANE_Int real_xoffset, real_yoffset, real_width, real_height,
    logical_width, logical_height;
  int odd_padding;
  int block_padding;

  SANE_Bool sf_user;
  SANE_Int sf_origin_x, sf_origin_y;
  SANE_Int sf_stride_x, sf_stride_y;
  SANE_Int sf_width, sf_height;

  SANE_Fixed exposure, exposure_r, exposure_g, exposure_b, exposure_i;	/* Control setting */
  SANE_Int real_exposure[CS9K_COLOR_MAX];	/* Scanner setting */
  SANE_Fixed exposure_per_frame;

  SANE_Bool focus_on_centre;
  SANE_Int focusx, focusy, real_focusx, real_focusy;
  SANE_Int focusz;

  SANE_Fixed adj_exp, adj_exp_ir;
  SANE_Fixed adj_exp_tgt, adj_exp_tgt_ir;
  cs9k_hist_alg_t adj_exp_type, adj_exp_type_ir;
  SANE_Word thresh;

  SANE_Byte *infrared_buf;
  SANE_Byte *thumbnail_ibuf, *thumbnail_obuf;
  SANE_Int infrared_buf_size, thumbnail_buf_size, thumb_padx, thumb_pady;

  /* status */
  SANE_Word option_flags;
  SANE_Bool scanning;
  SANE_Byte *line_buf;
  ssize_t line_buf_n, line_buf_i;
  SANE_Int line_buf_size;
  unsigned long sense_key, sense_asc, sense_ascq, sense_info, sense_sks;
  unsigned long sense_code;
  cs9k_status_t status;
  size_t xfer_position, xfer_bytes_total;
  SANE_Bool init_done;
  SANE_Int *histogram[4], histogram_size, histogram_alloc;
  SANE_Int old_focusx, old_focusy;
  SANE_Char adjust_exposure_title[CS9K_ADJUST_EXPOSURE_STRING_MAX],
    adjust_exposure_ir_title[CS9K_ADJUST_EXPOSURE_STRING_MAX];
  /* SANE stuff */
  SANE_Option_Descriptor option_list[CS9K_N_OPTIONS];
  SANE_Range ranges[CS9K_N_OPTIONS];
}
cs9k_t;

/* ========================================================================= */
/* prototypes */

static SANE_Word cs9k_adjustWord2list (SANE_Word * val,
				       const SANE_Word * list, char *tag);
static SANE_Word cs9k_adjustWord2range (SANE_Word * val, const SANE_Word min,
					const SANE_Word max, char *tag);
static SANE_Word cs9k_adjustWord2optrange (cs9k_t * s, SANE_Word * val,
					   const cs9k_option_t n);
static SANE_Word cs9k_convertString2val (SANE_String in,
					 const SANE_String_Const * list,
					 SANE_Int * val);
static SANE_Word cs9k_set_range (cs9k_t * s, cs9k_option_t n, SANE_Word min,
				 SANE_Word max, SANE_Word quant);
static SANE_Word cs9k_set_visible (cs9k_t * s, cs9k_option_t n,
				   SANE_Bool visible);
static SANE_Int cs9k_field2val (unsigned char *field, size_t size);
static cs9k_inq_val_t cs9k_get_inquiry_value (cs9k_t * s, int offset,
					      size_t size, SANE_Word * val);
static SANE_Status cs9k_open (const char *device, cs9k_interface_t interface,
			      cs9k_t ** sp);
static void cs9k_close (cs9k_t * s);
static SANE_Status cs9k_attach (const char *dev);
static SANE_Status cs9k_scsi_sense_handler (int fd, u_char * sense_buffer,
					    void *arg);
static SANE_Status cs9k_parse_sense_data (cs9k_t * s);
static void cs9k_init_buffer (cs9k_t * s);
static SANE_Status cs9k_pack_byte (cs9k_t * s, SANE_Byte byte);
static void cs9k_pack_3byte (cs9k_t * s, SANE_Word word);
static void cs9k_pack_long (cs9k_t * s, unsigned long val);
static void cs9k_pack_word (cs9k_t * s, unsigned long val);
static SANE_Status cs9k_parse_cmd (cs9k_t * s, char *text);
static SANE_Status cs9k_grow_send_buffer (cs9k_t * s);
static SANE_Status cs9k_issue_cmd (cs9k_t * s);
static cs9k_phase_t cs9k_phase_check (cs9k_t * s);
static SANE_Status cs9k_scanner_ready (cs9k_t * s, int flags);
static SANE_Word cs9k_option_set_value (cs9k_t * s, cs9k_option_t n, void *v);
static SANE_Status cs9k_page_inquiry (cs9k_t * s, int page);
static SANE_Word cs9k_set_frame_defs (cs9k_t * s, SANE_Word * flags);
static SANE_Status cs9k_read_frame_defs (cs9k_t * s, SANE_Word * flags);
static SANE_Status cs9k_full_inquiry (cs9k_t * s, SANE_Int * flags);
static SANE_Status cs9k_mode_select (cs9k_t * s);
static SANE_Status cs9k_reserve_unit (cs9k_t * s);
static SANE_Status cs9k_release_unit (cs9k_t * s);
static SANE_Status cs9k_execute (cs9k_t * s);
static SANE_Status cs9k_load (cs9k_t * s);
static SANE_Status cs9k_eject (cs9k_t * s);
static SANE_Status cs9k_reset (cs9k_t * s);
static SANE_Status cs9k_set_focus (cs9k_t * s);
static SANE_Status cs9k_autofocus (cs9k_t * s);
static SANE_Status cs9k_autoexposure (cs9k_t * s, int wb);
static SANE_Status cs9k_get_all_windows (cs9k_t * s);
static SANE_Status cs9k_get_exposure (cs9k_t * s);
static SANE_Status cs9k_set_window (cs9k_t * s, cs9k_scan_t type);
static SANE_Status cs9k_thumb2real (cs9k_t * s, SANE_Int xmin, SANE_Int xmax,
				    SANE_Int ymin, SANE_Int ymax);
static SANE_Status cs9k_convert_options (cs9k_t * s);
static SANE_Status cs9k_scan (cs9k_t * s, cs9k_scan_t type);
static SANE_Status cs9k_analyze_histogram (SANE_Int ** hists, SANE_Int layers,
					   SANE_Int levels,
					   double edge_target,
					   cs9k_hist_alg_t * type,
					   SANE_Fixed * offset);
static void cs9k_set_pixel (cs9k_t * s, SANE_Int x, SANE_Int y,
			    SANE_Int color);
static SANE_Status cs9k_draw_bounds (cs9k_t * s);
static SANE_Status cs9k_seq_begin (cs9k_t * s);
static SANE_Status cs9k_set_adjust_message (cs9k_t * s, SANE_Int option,
					    SANE_Fixed val, SANE_Word flags);
static void *cs9k_xmalloc (size_t size);
static void *cs9k_xrealloc (void *p, size_t size);
static void cs9k_xfree (const void *p);
static void cs9k_decode_sense (const unsigned long sk,
			       const unsigned long asc,
			       const unsigned long ascq,
			       const unsigned long info,
			       const unsigned long sks);

/* ========================================================================= */
/* global variables */

static int cs9k_colors[] = { 1, 2, 3, 9 };

static SANE_Device **device_list = NULL;
static int n_device_list = 0;
static cs9k_interface_t try_interface = CS9K_INTERFACE_UNKNOWN;
static int open_devices = 0;
static SANE_Word cs9k_bit_depths[] = { 1, 8, 10, 12, 14, 16, 0, 0 };

static SANE_String_Const infrared_fmt_list[] = {
  "3 layer plus 1 layer image pair",
  "4-layer single image",
  "Infrared only",
  0
};

#define infrared_fmt_size (32)

static SANE_String_Const source_list[] = {
  "Transparency Adapter",
  "Automatic Document Feeder",
  0
};

#define source_size (26)

static SANE_String_Const refresh_list[] = {
  "Refresh...",
  "Reload options",
  "Refresh parameters from scanner",
  0
};

#define refresh_size (32)

cs9k_frame_preset_t cs9k_frame_presets[] = {
  {"Pick frame...", 0, 0, 0, 0, 0, 0},
  {"6 x 4.5 cm", 250, 200, 8964, 7138, 8964, 6602},
  {"6 x 6 cm", 250, 200, 8964, 9500, 8964, 8964},
  {"6 x 7 cm", 250, 200, 8964, 11075, 8964, 10539},
  {"6 x 8 cm", 250, 200, 8964, 12650, 8964, 12114},
  {0, 0, 0, 0, 0, 0, 0}
};
static SANE_String_Const cs9k_frame_preset_list[6];
#define frame_preset_size (14)

/* For drawing on thumbnails */
SANE_Byte bordercolors1[][3] = {
  {255, 0, 0},		/* Even # positives */
  {0, 255, 0},		/*  Odd # positives */
  {0, 255, 255},	/* Even # negatives */
  {255, 0, 255} 	/*  Odd # negatives */
};
SANE_Byte bordercolors2[][6] = {
  {255, 255, 0, 0, 0, 0},		/* Even # positives */
  {0, 0, 255, 255, 0, 0},       	/*  Odd # positives */
  {0, 0, 255, 255, 255, 255},   	/* Even # negatives */
  {255, 255, 0, 0, 255, 255}    	/*  Odd # negatives */
};


/* ========================================================================= */
/* Utilities */

/* Round up to nearest value if possible, else down. */
SANE_Word
cs9k_adjustWord2list (SANE_Word * val, const SANE_Word * list, char *tag)
{
  /* Assumes high to low sort order */
  int i;

  for (i = list[0]; i >= 1; i--)
    {
      if (*val == list[i])
	return 0;
      if (*val < list[i])
	{
	  DBG (20,
	       "Adjusting 0x%x up to 0x%x to match constraint list entry %d for %s.\n",
	       *val, list[i], i, tag);
	  *val = list[i];
	  return SANE_INFO_INEXACT;
	}
    }
  DBG (20, "Adjusting 0x%x down to 0x%x to match constraint list.\n",
       *val, list[i]);
  *val = list[1];
  return SANE_INFO_INEXACT;
}

SANE_Word
cs9k_adjustWord2range (SANE_Word * val, const SANE_Word min,
		       const SANE_Word max, char *tag)
{
  if (*val < min)
    {
      DBG (20, "Adjusting 0x%x up to 0x%x to meet minimum for %s.\n", *val,
	   min, tag);
      *val = min;
      return SANE_INFO_INEXACT;
    }
  if (*val > max)
    {
      DBG (20, "Adjusting 0x%x down to 0x%x to meet maximum for %s.\n", *val,
	   max, tag);
      *val = max;
      return SANE_INFO_INEXACT;
    }
  return 0;
}

SANE_Word
cs9k_adjustWord2optrange (cs9k_t * s, SANE_Word * val, const cs9k_option_t n)
{
  if (*val < s->ranges[n].min)
    {
      DBG (20, "Adjusting 0x%x up to 0x%x to meet minimum for option %s.\n",
	   *val, s->ranges[n].min, s->option_list[n].title);
      *val = s->ranges[n].min;
      return SANE_INFO_INEXACT;
    }
  if (*val > s->ranges[n].max)
    {
      DBG (20, "Adjusting 0x%x down to 0x%x to meet maximum for option %s.\n",
	   *val, s->ranges[n].max, s->option_list[n].title);
      *val = s->ranges[n].max;
      return SANE_INFO_INEXACT;
    }
  return 0;
}

SANE_Word
cs9k_convertString2val (SANE_String in, const SANE_String_Const * list,
			SANE_Int * val)
{
  SANE_Int index;

  for (index = 0; list[index]; index++)
    {
      DBG (20, "Checking '%s' against '%s'\n", in, list[index]);
      if (!strcmp (list[index], in))
	{
	  *val = index;
	  DBG (20, "Found matching entry (%d)\n", *val);
	  return 0;
	}
    }
  *val = 0;
  return SANE_INFO_INEXACT;
}

SANE_Word
cs9k_set_range (cs9k_t * s, cs9k_option_t n, SANE_Word min, SANE_Word max,
		SANE_Word quant)
{
  SANE_Word flags = 0;

  DBG (20, "%s: %s: %x to %x, step %x\n", __func__, s->option_list[n].title,
       min, max, quant);

  if (s->ranges[n].min != min)
    {
      s->ranges[n].min = min;
      flags |= SANE_INFO_RELOAD_OPTIONS;
    }
  if (s->ranges[n].max != max)
    {
      s->ranges[n].max = max;
      flags |= SANE_INFO_RELOAD_OPTIONS;
    }
  if (s->ranges[n].quant != quant)
    {
      s->ranges[n].quant = quant;
      flags |= SANE_INFO_RELOAD_OPTIONS;
    }
  return flags;
}

/* Set control visibility */
SANE_Word
cs9k_set_visible (cs9k_t * s, cs9k_option_t n, SANE_Bool visible)
{
  if (visible && (s->option_list[n].cap & SANE_CAP_INACTIVE))
    {
      s->option_list[n].cap &= ~SANE_CAP_INACTIVE;
      DBG (12, "%s: Set %s active\n", __func__, s->option_list[n].title);
      return SANE_INFO_RELOAD_OPTIONS;
    }
  else if (!visible && ((s->option_list[n].cap & SANE_CAP_INACTIVE) == 0))
    {
      s->option_list[n].cap |= SANE_CAP_INACTIVE;
      DBG (12, "%s: Set %s inactive\n", __func__, s->option_list[n].title);
      return SANE_INFO_RELOAD_OPTIONS;
    }
  return 0;
}

/* Convert SCSI field 1-4 bytes to SANE_Int */
SANE_Int
cs9k_field2val (unsigned char *field, size_t size)
{
  SANE_Word val;
  size_t i;

  val = field[0];
  for (i = 1; i < size; i++)
    {
      val <<= 8;
      val |= field[i];
    }
  return (val);
}

/* Convert field in inquiry page to SANE_Int, with change and
 * error detection.
 */
cs9k_inq_val_t
cs9k_get_inquiry_value (cs9k_t * s, int offset, size_t size, SANE_Int * val)
{
  SANE_Word oldval;

  if ((offset + size - 4) > s->recv_buf[3])
    {
      DBG (1,
	   "Tried to read %d byte value at offset %d, but page "
	   "0x%02x has only %d bytes\n",
	   size, offset, s->recv_buf[1], s->recv_buf[3] + 4);
      return CS9K_INQ_BAD_OFFSET;
    }
  if ((offset + size - 4) > s->n_recv)
    {
      DBG (1,
	   "Tried to read %d byte value at offset %d, but read only %d bytes\n",
	   size, offset, s->n_recv);
      return CS9K_INQ_BAD_OFFSET;
    }
  oldval = *val;
  *val = cs9k_field2val (s->recv_buf + offset, size);
  DBG (10, "Page 0x%02x, value %d:%d: %0*lx (%lu)\n", s->recv_buf[1],
       offset, size, size * 2, (unsigned long) *val, (unsigned long) *val);
  if (oldval == *val)
    return CS9K_INQ_SAME_VALUE;
  return CS9K_INQ_NEW_VALUE;
}

/* ========================================================================= */
/* SANE entry points */

SANE_Status
sane_init (SANE_Int * version_code, SANE_Auth_Callback authorize)
{
  DBG_INIT ();
  DBG (1, "coolscan9k backend, version %i.%i.%i initializing.\n",
       CS9K_VERSION_MAJOR, CS9K_VERSION_MINOR, CS9K_REVISION);

  authorize = authorize;	/* to shut up compiler */

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, 0);

  sanei_usb_init ();

  return SANE_STATUS_GOOD;
}

void
sane_exit (void)
{
  int i;

  DBG (10, "%s\n", __func__);

  for (i = 0; i < n_device_list; i++)
    {
      cs9k_xfree (device_list[i]->name);
      cs9k_xfree (device_list[i]->vendor);
      cs9k_xfree (device_list[i]->model);
      cs9k_xfree (device_list[i]);
    }
  cs9k_xfree (device_list);
}

SANE_Status
sane_get_devices (const SANE_Device *** list, SANE_Bool local_only)
{
  char line[PATH_MAX], *p;
  FILE *config;

  local_only = local_only;	/* to shut up compiler */

  DBG (10, "%s\n", __func__);

  if (device_list)
    DBG (6,
	 "sane_get_devices(): Device list already populated, not probing again.\n");
  else
    {
      if (open_devices)
	{
	  DBG (4,
	       "sane_get_devices(): Devices open, not scanning for scanners.\n");
	  return SANE_STATUS_IO_ERROR;
	}

      config = sanei_config_open (CS9K_CONFIG_FILE);
      if (config)
	{
	  DBG (4, "sane_get_devices(): Reading config file.\n");
	  while (sanei_config_read (line, sizeof (line), config))
	    {
	      p = line;
	      p += strspn (line, " \t");
	      if (strlen (p) && (p[0] != '\n') && (p[0] != '#'))
		cs9k_open (line, CS9K_INTERFACE_UNKNOWN, NULL);
	    }
	  fclose (config);
	}
      else
	{
	  DBG (4, "sane_get_devices(): No config file found.\n");
	  cs9k_open ("auto", CS9K_INTERFACE_UNKNOWN, NULL);
	}

      DBG (6, "%s: %i device(s) detected.\n", __func__, n_device_list);
    }

  *list = (const SANE_Device **) device_list;

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const name, SANE_Handle * h)
{
  SANE_Status status;
  cs9k_t *s;
  int i_option;
  SANE_Option_Descriptor o;
  int alloc_failed = 0;

  DBG (10, "%s\n", __func__);

  status = cs9k_open (name, CS9K_INTERFACE_UNKNOWN, &s);
  if (status != SANE_STATUS_GOOD)
    return status;

  *h = (SANE_Handle) s;

  for (i_option = 0; i_option < CS9K_N_OPTIONS; i_option++)
    {
      o.name = o.title = o.desc = NULL;
      o.type = o.unit = o.cap = o.constraint_type = o.size = 0;
      o.constraint.range = NULL;	/* only one union member needs to be NULLed */
      switch (i_option)
	{
	case CS9K_OPTION_NUM:
	  o.name = "";
	  o.title = SANE_TITLE_NUM_OPTIONS;
	  o.desc = SANE_DESC_NUM_OPTIONS;
	  o.type = SANE_TYPE_INT;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_GROUP_INFRARED:
	  o.title = "Infrared / Dropout controls";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_GROUP_EXPOSURE:
	  o.title = "Exposure";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_GROUP_LUT:
	  o.title = "Response curves";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_GROUP_RES:
	  o.title = "Resolution controls";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_GROUP_FRAME:
	  o.title = "Frame definition and selection";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_GROUP_ACTION:
	  o.title = "Immediate commands";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_GROUP_FOCUS:
	  o.title = "Focus controls";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_SET_FRAME_GROUP:
	  o.title = "Frame definition";
	  o.type = SANE_TYPE_GROUP;
	  break;
	case CS9K_OPTION_PREVIEW:
	  o.name = "preview";
	  o.title = "Preview mode";
	  o.desc = "Preview mode";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_NEGATIVE:
	  o.name = "negative";
	  o.title = "Negative";
	  o.desc = "Negative film - hint to scanner; does not invert colors";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;

	case CS9K_OPTION_INFRARED:
	  o.name = "infrared";
	  o.title = "Read infrared channel";
	  o.desc = "Read infrared channel in addition to scan colors";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;

	case CS9K_OPTION_INFRARED_FMT:
	  o.name = "infrared-fmt";
	  o.title = "Infrared data format";
	  o.desc = "Infrared output format";
	  o.type = SANE_TYPE_STRING;
	  o.size = infrared_fmt_size;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_STRING_LIST;
	  o.constraint.string_list = infrared_fmt_list;
	  break;

	case CS9K_OPTION_DEPTH:
	  o.name = "depth";
	  o.title = "Bit depth per channel";
	  o.desc = "Number of bits output by scanner for each channel";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_WORD_LIST;
	  o.constraint.word_list = s->depth_list;
	  break;
	case CS9K_OPTION_EXPOSURE:
	  o.name = "exposure";
	  o.title = "Overall exposure [stops]";
	  o.desc = "Common exposure for visible channels - higher is brighter";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_EXPOSURE_R:
	  o.name = "red-exposure";
	  o.title = "Red offset [stops]";
	  o.desc = "Exposure time offset for red channel";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_EXPOSURE_G:
	  o.name = "green-exposure";
	  o.title = "Green offset [stops]";
	  o.desc = "Exposure time offset for green channel";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_EXPOSURE_B:
	  o.name = "blue-exposure";
	  o.title = "Blue offset [stops]";
	  o.desc = "Exposure time offset for blue channel";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_EXPOSURE_I:
	  o.name = "ir-exposure";
	  o.title = "IR exposure [stops]";
	  o.desc = "Exposure time offset for dropout channel";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_EXPOSURE_PER_FRAME:
	  o.name = "adjust-exposure-per-frame";
	  o.title = "Limit for exposure adjustment per frame";
	  o.desc = "No adjustment is done if suggested delta larger than this value";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT | SANE_CAP_ADVANCED;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TARGET:
	  o.name = "adjust-exposure-target";
	  o.title =
	    "Target for upper edge of histogram below saturation [stops]";
	  o.desc = "Exposure adjustment determined by previous scan";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap =
	    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT | SANE_CAP_ADVANCED;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TARGET_IR:
	  o.name = "adjust-exposure-target-ir";
	  o.title =
	    "Target for upper edge of IR histogram below saturation [stops]";
	  o.desc = "Exposure adjustment determined by previous scan";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap =
	    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT | SANE_CAP_ADVANCED;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TEXT:
	  o.name = "adjust-exposure-text";
	  o.title = "Overall exposure info";
	  o.desc = "Suggested changes based on previous scan";
	  o.type = SANE_TYPE_STRING;
	  o.size = CS9K_ADJUST_EXPOSURE_STRING_MAX;
	  o.cap = SANE_CAP_HARD_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE:
	  o.name = "adjust-exposure";
	  o.title = "Adjust Overall Exposure";
	  o.desc = "Exposure adjustment determined by previous scan";
	  o.type = SANE_TYPE_BUTTON;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TEXT_IR:
	  o.name = "adjust-exposure-text-ir";
	  o.title = "IR exposure info";
	  o.desc = "Suggested changes based on previous scan";
	  o.type = SANE_TYPE_STRING;
	  o.size = CS9K_ADJUST_EXPOSURE_STRING_MAX;
	  o.cap = SANE_CAP_HARD_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_IR:
	  o.name = "adjust-exposure-ir";
	  o.title = "Adjust IR Exposure";
	  o.desc = "Exposure adjustment determined by previous scan";
	  o.type = SANE_TYPE_BUTTON;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_LUT_R:
	  o.name = "red-gamma-table";
	  o.title = "LUT for red channel";
	  o.desc = "LUT for red channel";
	  o.type = SANE_TYPE_INT;
	  o.size = 0;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_LUT_G:
	  o.name = "green-gamma-table";
	  o.title = "LUT for green channel";
	  o.desc = "LUT for green channel";
	  o.type = SANE_TYPE_INT;
	  o.size = 0;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_LUT_B:
	  o.name = "blue-gamma-table";
	  o.title = "LUT for blue channel";
	  o.desc = "LUT for blue channel";
	  o.type = SANE_TYPE_INT;
	  o.size = 0;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_REFRESH:
	  o.name = "refresh";
	  o.title = "Refresh Options [see new tray]";
	  o.desc = "Trigger re-reading of controls";
	  o.type = SANE_TYPE_STRING;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.size = refresh_size;
	  o.constraint_type = SANE_CONSTRAINT_STRING_LIST;
	  o.constraint.string_list = refresh_list;
	  break;
	case CS9K_OPTION_LOAD:
	  o.name = "load";
	  o.title = "Load";
	  o.desc = "Load next slide";
	  o.type = SANE_TYPE_BUTTON;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_AUTOLOAD:
	  o.name = "autoload";
	  o.title = "Autoload";
	  o.desc = "Autoload slide before each scan";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_EJECT:
	  o.name = "eject";
	  o.title = "Eject";
	  o.desc = "Eject loaded medium";
	  o.type = SANE_TYPE_BUTTON;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_RESET:
	  o.name = "reset";
	  o.title = "Reset scanner";
	  o.desc = "Initialize scanner";
	  o.type = SANE_TYPE_BUTTON;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_RES_INDEPENDENT:
	  o.name = "independent-res";
	  o.title = "Independent x/y resolutions";
	  o.desc =
	    "Enable independent controls for scanning resolution in x and y direction";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap =
	    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT | SANE_CAP_ADVANCED;
	  break;
	case CS9K_OPTION_RESX:
	case CS9K_OPTION_RES:
	case CS9K_OPTION_PREVIEW_RESOLUTION:
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  if (i_option == CS9K_OPTION_PREVIEW_RESOLUTION)
	    {
	      o.name = "preview-resolution";
	      o.title = "Preview resolution";
	      o.desc =
		"Scanning resolution for preview mode in dpi, affecting both x and y directions";
	      o.cap |= SANE_CAP_ADVANCED;
	    }
	  else if (i_option == CS9K_OPTION_RES)
	    {
	      o.name = "resolution";
	      o.title = "Resolution";
	      o.desc =
		"Scanning resolution in dpi, affecting both x and y directions";
	    }
	  else
	    {
	      o.name = "x-resolution";
	      o.title = "X resolution";
	      o.desc =
		"Scanning resolution in dpi, affecting x direction only";
	      o.cap |= SANE_CAP_ADVANCED;
	    }
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_DPI;
	  o.size = WSIZE;
	  o.constraint_type = SANE_CONSTRAINT_WORD_LIST;
	  o.constraint.word_list = s->resx_list;
	  break;
	case CS9K_OPTION_RESY:
	  o.name = "y-resolution";
	  o.title = "Y resolution";
	  o.desc = "Scanning resolution in dpi, affecting y direction only";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_DPI;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT |
	    SANE_CAP_ADVANCED;
	  o.constraint_type = SANE_CONSTRAINT_WORD_LIST;
	  o.constraint.word_list = s->resy_list;
	  break;
	case CS9K_OPTION_THUMB_RESOLUTION:
	  o.name = "thumb-resolution";
	  o.title = "Thumbnail resolution";
	  o.desc = "Scanning resolution in dpi, for overview scans";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_DPI;
	  o.size = WSIZE;
	  o.cap =
	    SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT | SANE_CAP_ADVANCED;
	  o.constraint_type = SANE_CONSTRAINT_WORD_LIST;
	  o.constraint.word_list = s->res_thumb_list;
	  break;
	case CS9K_OPTION_SOURCE:
	  o.name = "source";
	  o.title = "Document Source";
	  o.desc = "Affects auto-increment of frame number";
	  o.type = SANE_TYPE_STRING;
	  o.size = source_size;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_STRING_LIST;
	  o.constraint.string_list = source_list;
	  break;

	case CS9K_OPTION_HOLDER:
	  o.name = "holder";
	  o.title = "Film Holder";
	  o.desc = "Name of inserted film holder";
	  o.type = SANE_TYPE_STRING;
	  o.size = CS9K_HOLDER_NAME_MAX;
	  o.cap = SANE_CAP_HARD_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_NEXTSCAN:
	  o.name = "nextscan";
	  o.title = "Next scan is:";
	  o.desc = "Type and frame number of upcoming scan";
	  o.type = SANE_TYPE_STRING;
	  o.size = CS9K_NEXTSCAN_MAX;
	  o.cap = SANE_CAP_HARD_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_SCAN_THUMBNAILS:
	  o.name = "scan-thumbnails";
	  o.title = "Thumbnail Mode";
	  o.desc = "Show all frames plus boundaries";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_FRAME:
	  o.name = "frame";
	  o.title = "Frame number";
	  o.desc = "Frame number of first image to be scanned";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SUBFRAME_X:
	  o.name = "subframe-x";
	  o.title = "Frame shift X";
	  o.desc = "Fine X position within the selected frame";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_MM;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SUBFRAME_Y:
	  o.name = "subframe-y";
	  o.title = "Frame shift Y";
	  o.desc = "Fine Y position within the selected frame";
	  o.type = SANE_TYPE_FIXED;
	  o.unit = SANE_UNIT_MM;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_XMIN:
	  o.name = "tl-x";
	  o.title = "Left x value of scan area";
	  o.desc = "Left x value of scan area";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_XMAX:
	  o.name = "br-x";
	  o.title = "Right x value of scan area";
	  o.desc = "Right x value of scan area";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_YMIN:
	  o.name = "tl-y";
	  o.title = "Top y value of scan area";
	  o.desc = "Top y value of scan area";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_YMAX:
	  o.name = "br-y";
	  o.title = "Bottom y value of scan area";
	  o.desc = "Bottom y value of scan area";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SET_FRAME_USER:
	  o.name = "user-frame";
	  o.title = "User defined frames";
	  o.desc = "For carriers without defined image areas";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_SET_FRAME_ORIGIN_X:
	  o.name = "frame-origin-x";
	  o.title = "X origin for user-defined frames";
	  o.desc = "Top-left X location";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SET_FRAME_ORIGIN_Y:
	  o.name = "frame-origin-y";
	  o.title = "Y origin for user-defined frames";
	  o.desc = "Top-left Y location";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SET_FRAME_STRIDE_X:
	  o.name = "sf-stride-x";
	  o.title = "X frame-to-frame stride";
	  o.desc = "Top-left to top-left distance";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SET_FRAME_STRIDE_Y:
	  o.name = "sf-stride-y";
	  o.title = "Y frame-to-frame stride";
	  o.desc = "Top-left to top-left distance";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SET_FRAME_WIDTH:
	  o.name = "sf-width";
	  o.title = "Width of a single frame";
	  o.desc = "Width of a single frame";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SET_FRAME_HEIGHT:
	  o.name = "sf-height";
	  o.title = "Height of a single frame";
	  o.desc = "Height of a single frame";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SET_FRAME_PRESET:
	  o.name = "sf-preset";
	  o.title = "Fill in user options with preset values";
	  o.desc = "User values may be adjusted afterward";
	  o.type = SANE_TYPE_STRING;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.size = frame_preset_size;
	  o.constraint_type = SANE_CONSTRAINT_STRING_LIST;
	  o.constraint.string_list = cs9k_frame_preset_list;
	  break;

	case CS9K_OPTION_FOCUS_ON_CENTRE:
	  o.name = "focus-on-centre";
	  o.title = "Use centre of scan area as AF point";
	  o.desc =
	    "Use centre of scan area as AF point instead of manual AF point selection";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_FOCUS:
	  o.name = "focus";
	  o.title = "Focus position";
	  o.desc = "Focus position for manual focus (Z axis)";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_NONE;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_AUTOFOCUS:
	  o.name = "autofocus";
	  o.title = "Autofocus";
	  o.desc = "Perform autofocus before scan";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_FOCUSX:
	  o.name = "focusx";
	  o.title = "X coordinate of AF point";
	  o.desc = "X coordinate of AF point";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_FOCUSY:
	  o.name = "focusy";
	  o.title = "Y coordinate of AF point";
	  o.desc = "Y coordinate of AF point";
	  o.type = SANE_TYPE_INT;
	  o.unit = SANE_UNIT_PIXEL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  o.constraint_type = SANE_CONSTRAINT_RANGE;
	  break;
	case CS9K_OPTION_SCAN_AE:
	  o.name = "ae";
	  o.title = "Auto-exposure";
	  o.desc = "Perform auto-exposure before scan";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	case CS9K_OPTION_SCAN_AE_WB:
	  o.name = "ae-wb";
	  o.title = "Auto-exposure with white balance";
	  o.desc = "Perform auto-exposure with white balance before scan";
	  o.type = SANE_TYPE_BOOL;
	  o.size = WSIZE;
	  o.cap = SANE_CAP_SOFT_SELECT | SANE_CAP_SOFT_DETECT;
	  break;
	default:
	  DBG (1, "BUG: sane_open(): Unknown option number %d.\n", i_option);
	  return SANE_STATUS_INVAL;
	  break;
	}
      if (o.constraint_type == SANE_CONSTRAINT_RANGE)
	o.constraint.range = &s->ranges[i_option];

      s->option_list[i_option] = o;
    }

  /* get device properties */
  status = cs9k_full_inquiry (s, 0);
  if (status != SANE_STATUS_GOOD)
    return status;

  status = cs9k_mode_select (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  if (s->scancolors & CS9K_CLRCMP_RGB)
    {
      s->n_colors_in = 3;
    }
  else if (s->scancolors & CS9K_CLRCMP_GRAYSCALE)
    {
      s->n_colors_in = 1;
    }
  else
    {
      DBG (4, "RGB support not indicated; using it anyway.\n");
      s->n_colors_in = 3;
    }

  if (alloc_failed)
    {
      cs9k_close (s);
      return SANE_STATUS_NO_MEM;
    }

  return cs9k_reserve_unit (s);
}

void
sane_close (SANE_Handle h)
{
  cs9k_t *s = (cs9k_t *) h;

  DBG (10, "%s\n", __func__);

  cs9k_release_unit (s);
  cs9k_close (s);
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle h, SANE_Int n)
{
  cs9k_t *s = (cs9k_t *) h;

  if ((n >= 0) && (n < CS9K_N_OPTIONS))
    {
      DBG (24, "%s: option %i (%s)\n", __func__, n, s->option_list[n].title);
      if (s->option_list[n].constraint_type
	  && (s->option_list[n].constraint.range == NULL))
	{
	  DBG (1, "%s:   constraint type %x requires constraint structure\n",
	       __func__, s->option_list[n].constraint_type);
	  return NULL;
	}
      return &s->option_list[n];
    }
  else
    {
      DBG (1, "%s, option %i (INVALID)\n", __func__, n);
      return NULL;
    }
}

SANE_Status
sane_control_option (SANE_Handle h, SANE_Int n, SANE_Action a, void *v,
		     SANE_Int * i)
{
  cs9k_t *s = (cs9k_t *) h;
  SANE_Int flags = 0;
  cs9k_pixel_t pixel;
  SANE_Option_Descriptor o = s->option_list[n];

  if (s->medium_changed)
    {
      DBG (4, "%s: Medium changed, querying parameters.\n", __func__);
      cs9k_full_inquiry (s, &flags);
      s->medium_changed = SANE_FALSE;
    }
  switch (a)
    {
    case SANE_ACTION_GET_VALUE:

      DBG (22, "%s, option %i (%s), action %i.\n", __func__, n, o.title, a);

      switch (n)
	{
	case CS9K_OPTION_NUM:
	  *(SANE_Word *) v = CS9K_N_OPTIONS;
	  break;
	case CS9K_OPTION_NEGATIVE:
	  *(SANE_Word *) v = s->negative;
	  break;
	case CS9K_OPTION_INFRARED:
	  *(SANE_Word *) v = (s->infrared == CS9K_INFRARED_OFF) ?
	    SANE_FALSE : SANE_TRUE;
	  break;
	case CS9K_OPTION_INFRARED_FMT:
	  strcpy (v, infrared_fmt_list[s->infrared_fmt]);
	  break;
	case CS9K_OPTION_DEPTH:
	  *(SANE_Word *) v = s->depth;
	  break;
	case CS9K_OPTION_PREVIEW:
	  *(SANE_Word *) v = s->preview;
	  break;
	case CS9K_OPTION_AUTOLOAD:
	  *(SANE_Word *) v = s->autoload;
	  break;
	case CS9K_OPTION_EXPOSURE:
	  if (s->negative)
	    *(SANE_Word *) v = s->ranges[n].max - s->exposure;
	  else
	    *(SANE_Word *) v = s->exposure;
	  break;
	case CS9K_OPTION_EXPOSURE_R:
	  if (s->negative)
	    *(SANE_Word *) v = -s->exposure_r;
	  else
	    *(SANE_Word *) v = s->exposure_r;
	  break;
	case CS9K_OPTION_EXPOSURE_G:
	  if (s->negative)
	    *(SANE_Word *) v = -s->exposure_g;
	  else
	    *(SANE_Word *) v = s->exposure_g;
	  break;
	case CS9K_OPTION_EXPOSURE_B:
	  if (s->negative)
	    *(SANE_Word *) v = -s->exposure_b;
	  else
	    *(SANE_Word *) v = s->exposure_b;
	  break;
	case CS9K_OPTION_EXPOSURE_I:
	  if (s->negative)
	    *(SANE_Word *) v = s->ranges[n].max - s->exposure_i;
	  else
	    *(SANE_Word *) v = s->exposure_i;
	  break;
	case CS9K_OPTION_EXPOSURE_PER_FRAME:
	  *(SANE_Word *) v = s->exposure_per_frame;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TARGET:
	  *(SANE_Word *) v = s->adj_exp_tgt;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TARGET_IR:
	  *(SANE_Word *) v = s->adj_exp_tgt_ir;
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TEXT:
	  strcpy (v, s->adjust_exposure_title);
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE:
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_TEXT_IR:
	  strcpy (v, s->adjust_exposure_ir_title);
	  break;
	case CS9K_OPTION_ADJUST_EXPOSURE_IR:
	  break;
	case CS9K_OPTION_LUT_R:
	  if (!(s->lut_r))
	    return SANE_STATUS_INVAL;
	  for (pixel = 0; pixel < s->n_lut; pixel++)
	    ((SANE_Word *) v)[pixel] = s->lut_r[pixel];
	  break;
	case CS9K_OPTION_LUT_G:
	  if (!(s->lut_g))
	    return SANE_STATUS_INVAL;
	  for (pixel = 0; pixel < s->n_lut; pixel++)
	    ((SANE_Word *) v)[pixel] = s->lut_g[pixel];
	  break;
	case CS9K_OPTION_LUT_B:
	  if (!(s->lut_b))
	    return SANE_STATUS_INVAL;
	  for (pixel = 0; pixel < s->n_lut; pixel++)
	    ((SANE_Word *) v)[pixel] = s->lut_b[pixel];
	  break;
	case CS9K_OPTION_EJECT:
	  break;
	case CS9K_OPTION_REFRESH:
	  strcpy (v, refresh_list[0]);
	  break;
	case CS9K_OPTION_LOAD:
	  break;
	case CS9K_OPTION_RESET:
	  break;
	case CS9K_OPTION_SOURCE:
	  strcpy (v, source_list[s->source]);
	  break;
	case CS9K_OPTION_HOLDER:
	  strcpy (v, s->holder_string);
	  DBG (10, "Set holder control to '%s'\n", (char *) v);
	  break;
	case CS9K_OPTION_NEXTSCAN:
	  sprintf (s->infrared_txt, "%s",
		   s->infrared == CS9K_INFRARED_OFF ? "RGB" :
		   s->infrared == CS9K_INFRARED_4PLANE ? "RGBI" :
		   s->infrared == CS9K_INFRARED_IN ? "RGB" :
		   s->infrared == CS9K_INFRARED_OUT ? "GRAY (IR)" :
		   s->infrared == CS9K_INFRARED_ONLY ? "GRAY (IR)" : "WTF");

	  if (s->frame_bounds[s->c_frame - 1].flags & CS9K_FBF_EXAMINED)
	    {
	       if (s->frame_bounds[s->c_frame - 1].flags & CS9K_FBF_PRESENT)
		 {
		   if (s->frame_bounds[s->c_frame - 1].adj_exp <= s->exposure_per_frame)
		     snprintf (s->exposure_txt, 20, ", exp. adj. %.2f", 
			    SANE_UNFIX((s->negative ? -1.0 : 1.0) *
				       s->frame_bounds[s->c_frame - 1].adj_exp));
		   else
		     snprintf (s->exposure_txt, 20, ", ignoring exp.");
		 }
	       else
		 snprintf (s->exposure_txt, 20, ", not present");
	    }
	  else
	    s->exposure_txt[0] = 0;

	  if (s->thumbnail)
	    if (s->source)
	      snprintf (v, CS9K_NEXTSCAN_MAX, "Frame %d, %s%s%s", s->c_frame,
			s->infrared_txt,
			(s->thumb_sel_descr ==
			 CS9K_THUMB_SEL_CROSSFRAME) ? ", using full frames"
			: (s->thumb_sel_descr ==
			   CS9K_THUMB_SEL_INFRAME) ?
			", using translated selection" : "",
			s->exposure_txt);
	    else
	      snprintf (v, CS9K_NEXTSCAN_MAX,
			"Scanning area selected on thumbnails, %s",
			s->infrared_txt);
	  else
	    snprintf (v, CS9K_NEXTSCAN_MAX, "Frame %d, %s format", s->c_frame,
		      s->infrared_txt);
	  break;
	case CS9K_OPTION_SCAN_THUMBNAILS:
	  *(SANE_Word *) v = s->thumbnail;
	  break;
	case CS9K_OPTION_FRAME:
	  *(SANE_Word *) v = s->i_frame;
	  break;
	case CS9K_OPTION_SUBFRAME_X:
	  *(SANE_Word *) v = SANE_FIX (s->subframe_x);
	  break;
	case CS9K_OPTION_SUBFRAME_Y:
	  *(SANE_Word *) v = SANE_FIX (s->subframe_y);
	  break;
	case CS9K_OPTION_RES:
	  *(SANE_Word *) v = s->res;
	  break;
	case CS9K_OPTION_RESX:
	  *(SANE_Word *) v = s->resx;
	  break;
	case CS9K_OPTION_RESY:
	  *(SANE_Word *) v = s->resy;
	  break;
	case CS9K_OPTION_RES_INDEPENDENT:
	  *(SANE_Word *) v = s->res_independent;
	  break;
	case CS9K_OPTION_PREVIEW_RESOLUTION:
	  *(SANE_Word *) v = s->res_preview;
	  break;
	case CS9K_OPTION_THUMB_RESOLUTION:
	  *(SANE_Word *) v = s->res_thumb;
	  break;
	case CS9K_OPTION_XMIN:
	  *(SANE_Word *) v = s->xmin;
	  break;
	case CS9K_OPTION_XMAX:
	  *(SANE_Word *) v = s->xmax;
	  break;
	case CS9K_OPTION_YMIN:
	  *(SANE_Word *) v = s->ymin;
	  break;
	case CS9K_OPTION_YMAX:
	  *(SANE_Word *) v = s->ymax;
	  break;
	case CS9K_OPTION_SET_FRAME_USER:
	  *(SANE_Bool *) v = s->sf_user;
	  break;
	case CS9K_OPTION_SET_FRAME_ORIGIN_X:
	  *(SANE_Word *) v = s->sf_origin_x;
	  break;
	case CS9K_OPTION_SET_FRAME_ORIGIN_Y:
	  *(SANE_Word *) v = s->sf_origin_y;
	  break;
	case CS9K_OPTION_SET_FRAME_STRIDE_X:
	  *(SANE_Word *) v = s->sf_stride_x;
	  break;
	case CS9K_OPTION_SET_FRAME_STRIDE_Y:
	  *(SANE_Word *) v = s->sf_stride_y;
	  break;
	case CS9K_OPTION_SET_FRAME_WIDTH:
	  *(SANE_Word *) v = s->sf_width;
	  break;
	case CS9K_OPTION_SET_FRAME_HEIGHT:
	  *(SANE_Word *) v = s->sf_height;
	  break;
	case CS9K_OPTION_SET_FRAME_PRESET:
	  strcpy (v, cs9k_frame_preset_list[0]);
	  break;
	case CS9K_OPTION_FOCUS_ON_CENTRE:
	  *(SANE_Word *) v = s->focus_on_centre;
	  break;
	case CS9K_OPTION_FOCUS:
	  *(SANE_Word *) v = s->focusz;
	  break;
	case CS9K_OPTION_AUTOFOCUS:
	  *(SANE_Word *) v = s->autofocus;
	  break;
	case CS9K_OPTION_FOCUSX:
	  *(SANE_Word *) v = s->focusx;
	  break;
	case CS9K_OPTION_FOCUSY:
	  *(SANE_Word *) v = s->focusy;
	  break;
	case CS9K_OPTION_SCAN_AE:
	  *(SANE_Word *) v = s->ae;
	  break;
	case CS9K_OPTION_SCAN_AE_WB:
	  *(SANE_Word *) v = s->aewb;
	  break;
	default:
	  DBG (4, "%s: Unknown option (bug?).\n", __func__);
	  return SANE_STATUS_INVAL;
	}
      break;

    case SANE_ACTION_SET_VALUE:
      if (s->scanning)
	return SANE_STATUS_INVAL;

      cs9k_option_set_value (s, n, v);
      if (i)
	{
	  flags |= s->option_flags;
	  s->option_flags = 0;
	}
      break;

    default:
      DBG (1, "BUG: sane_control_option(): Unknown action number.\n");
      return SANE_STATUS_INVAL;
      break;
    }

  if (i)
    {
      DBG (10, "%s: Reporting val = 0x%x, flags%s%s%s\n", __func__,
	   (v != NULL) ? *(int *) v : -1,
	   (flags & SANE_INFO_INEXACT ? " SANE_INFO_INEXACT" : ""),
	   (flags & SANE_INFO_RELOAD_OPTIONS ? " SANE_INFO_RELOAD_OPTIONS" :
	    ""),
	   (flags & SANE_INFO_RELOAD_PARAMS ? " SANE_INFO_RELOAD_PARAMS" :
	    ""));
      *i = flags;
    }

  else if (flags)
    DBG (4,
	 "  sane_control_option() front end ignoring flags %08x, action %d, control %d\n",
	 flags, a, n);

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_get_parameters (SANE_Handle h, SANE_Parameters * p)
{
  cs9k_t *s = (cs9k_t *) h;
  SANE_Status status;

  DBG (10, "%s: %s\n", __func__,
       s->scanning ? "scanning" : "calc parameters");

  if (!s->scanning)
    {				/* only recalculate when not scanning */
      status = cs9k_convert_options (s);
      if (status != SANE_STATUS_GOOD)
	return status;
    }

  if (0 && s->preview)
    p->format = SANE_FRAME_RGB;
  else
    switch (s->infrared)
      {
      case CS9K_INFRARED_OFF:
	p->format = SANE_FRAME_RGB;
	break;
      case CS9K_INFRARED_4PLANE:
	p->format = SANE_FRAME_RGBI;
	break;
      case CS9K_INFRARED_IN:
	p->format = SANE_FRAME_RGB;
	break;
      case CS9K_INFRARED_OUT:
	p->format = SANE_FRAME_GRAY;
	break;
      case CS9K_INFRARED_ONLY:
	p->format = SANE_FRAME_GRAY;
	break;
      default:
	DBG (4, "Unknown infrared state %d.\n", s->infrared_fmt);
	p->format = SANE_FRAME_RGB;
	break;
      }
  p->bytes_per_line = s->n_colors_out * s->logical_width * s->bytes_per_pixel;

  p->last_frame = SANE_TRUE;
  p->lines = s->logical_height;
  p->depth = 8 * s->bytes_per_pixel;
  p->pixels_per_line = s->logical_width;

  s->xfer_bytes_total =
    s->bytes_per_pixel * s->n_colors_out * s->logical_width *
    s->logical_height;

  DBG (10, "%s:===== preview: %d, thumbnail: %d, infrared_fmt: %d =====\n",
       __func__, s->preview, s->thumbnail, s->infrared);
  DBG (10, "%s:           format: %s\n", __func__,
       (p->format == SANE_FRAME_GRAY) ? "SANE_FRAME_GRAY" : (p->format ==
							     SANE_FRAME_RGB) ?
       "SANE_FRAME_RGB" : (p->format ==
			   SANE_FRAME_RGBI) ? "SANE_FRAME_RGBI" : "");
  DBG (10, "%s:       last_frame: %d\n", __func__, p->last_frame);
  DBG (10, "%s:   bytes_per_line: %d\n", __func__, p->bytes_per_line);
  DBG (10, "%s:  pixels_per_line: %d\n", __func__, p->pixels_per_line);
  DBG (10, "%s:            lines: %d\n", __func__, p->lines);
  DBG (10, "%s:            depth: %d\n", __func__, p->depth);

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_start (SANE_Handle h)
{
  cs9k_t *s = (cs9k_t *) h;
  SANE_Status status;

  DBG (10, "%s\n", __func__);

  if (s->scanning)
    return SANE_STATUS_INVAL;

  if (s->c_frame > s->n_frames)
    {
      s->c_frame = s->i_frame;
      DBG (4, "%s: no more frames\n", __func__);
      if (!s->preview || !s->thumbnail)
	return SANE_STATUS_NO_DOCS;
    }

  if (s->n_frames > 1)
    {
      DBG (4, "%s: scanning frame at position %d\n", __func__, s->c_frame);
    }

  status = cs9k_convert_options (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  s->line_buf_i = 0;
  s->xfer_position = 0;

  s->scanning = SANE_TRUE;

  if (s->infrared != CS9K_INFRARED_OUT
      && !(s->preview && s->thumbnail && s->thumbnail_done))
    {
      /* load if appropriate */
      if (s->autoload)
	{
	  status = cs9k_load (s);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	}

      /* check for documents */
      status = cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
      if (status != SANE_STATUS_GOOD)
	return status;
      if (s->status & CS9K_STATUS_NO_DOCS)
	{
	  DBG (4, "%s: no documents.\n", __func__);
	  return SANE_STATUS_NO_DOCS;
	}

      if (s->autofocus)
	{
	  status = cs9k_autofocus (s);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	}

      if (s->aewb)
	{
	  status = cs9k_autoexposure (s, 1);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	}
      else if (s->ae)
	{
	  status = cs9k_autoexposure (s, 0);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	}
    }
  return cs9k_scan (s, CS9K_SCAN_NORMAL);
}

SANE_Status
sane_read (SANE_Handle h, SANE_Byte * buf, SANE_Int maxlen, SANE_Int * len)
{
  cs9k_t *s = (cs9k_t *) h;
  SANE_Status status;
  ssize_t xfer_len_in, xfer_len_line, xfer_len_out;
  SANE_Int index;
  int color;
  u_int8_t *s8 = NULL;
  u_int16_t *s16 = NULL;
  SANE_Byte *line_buf_new;
  int j;
  SANE_Bool hist_y_inrange;
  SANE_Int hist_x_lowrange, hist_x_highrange;
  SANE_Int src_pel, dst_pel;

  DBG (32, "%s, maxlen = %i.\n", __func__, maxlen);

  if (!s->scanning)
    {
      *len = 0;
      return SANE_STATUS_CANCELLED;
    }

  if (s->xfer_position == 0)
    {
      /* Haven't read any data yet; initialize whatever's needed */
      DBG (10, "%s: Start of new image %dx%d, %d Bpp, image %d bytes\n",
	   __func__, s->logical_width, s->logical_height, s->bytes_per_pixel,
	   s->xfer_bytes_total);

      /* Grow thumbnail buffer if necessary */
      if (s->preview && s->thumbnail
	  && (size_t) s->thumbnail_buf_size < s->xfer_bytes_total)
	{
	  s->thumbnail_ibuf =
	    cs9k_xrealloc (s->thumbnail_ibuf, s->xfer_bytes_total * 2);
	  if (!s->thumbnail_ibuf)
	    return SANE_STATUS_NO_MEM;
	  s->thumbnail_obuf = s->thumbnail_ibuf + s->xfer_bytes_total;
	  DBG (4, "Thumbnail buffers at %p and %p (image is %d bytes)\n",
	       s->thumbnail_ibuf, s->thumbnail_obuf, s->xfer_bytes_total);
	  s->thumbnail_buf_size = s->xfer_bytes_total;
	}

      /* Grow histogram buffer if necessary */
      /* 1/HIST_STEPS stop intervals */
      s->histogram_size = s->real_depth * HIST_STEPS;
      if (s->histogram_size > s->histogram_alloc)
	{
	  s->histogram[0] =
	    cs9k_xrealloc (s->histogram[0],
			   4 * s->histogram_size * sizeof (SANE_Word));

	  if (s->histogram[0])
	    {
	      for (index = 1; index < 4; index++)
		s->histogram[index] =
		  s->histogram[0] + index * s->histogram_size;
	      s->histogram_alloc = s->histogram_size;
	      DBG (12,
		   "%s: Allocated %d entry histograms at %p, %p, %p, %p\n",
		   __func__, s->histogram_alloc, (void *) s->histogram[0],
		   (void *) s->histogram[1], (void *) s->histogram[2],
		   (void *) s->histogram[3]);
	    }
	  else
	    {
	      for (index = 1; index < 4; index++)
		s->histogram[index] = NULL;
	      s->histogram_size = 0;
	      s->histogram_alloc = 0;
	    }
	}

      /* Initialize histogram, real_data coming */
      if ((s->infrared != CS9K_INFRARED_OUT)
	  && !(s->preview && s->thumbnail && s->thumbnail_done)
	  && s->histogram_size)
	memset (s->histogram[0], 0,
		4 * s->histogram_size * sizeof (SANE_Int));

      if (s->preview && s->thumbnail && !s->thumbnail_done)
	/* Starting new thumbnail scan, clear frame status */
	for (j = 0; j < s->n_frames; j++)
	  s->frame_bounds[j].flags = 0;

      if (s->preview && s->thumbnail && s->thumbnail_done)
	/* Re-using thumbnail data.  Re-draw markup to reflect */
	/*   changes in boundaries and frame status            */
	cs9k_draw_bounds (s);

      /* Grow infrared frame buffer if necessary */
      if (s->infrared == CS9K_INFRARED_IN
	  && (s->infrared_buf_size <
	      s->logical_width * s->logical_height * s->bytes_per_pixel))
	{
	  s->infrared_buf =
	    cs9k_xrealloc (s->infrared_buf,
			   s->logical_width * s->logical_height *
			   s->bytes_per_pixel);
	  if (!s->infrared_buf)
	    return SANE_STATUS_NO_MEM;
	  s->infrared_buf_size =
	    s->logical_width * s->logical_height * s->bytes_per_pixel;
	  DBG (4, "Infrared buffer now %d bytes.\n", s->infrared_buf_size);
	}
    }				/* End init at start of image */

  /* transfer leftovers from buffer */
  if (s->line_buf_i > 0)
    {
      xfer_len_out = s->line_buf_n - s->line_buf_i;
      if (xfer_len_out > maxlen)
	xfer_len_out = maxlen;

      memcpy (buf, &(s->line_buf[s->line_buf_i]), xfer_len_out);

      s->line_buf_i += xfer_len_out;
      if (s->line_buf_i >= s->line_buf_n)
	s->line_buf_i = 0;

      *len = xfer_len_out;
      return SANE_STATUS_GOOD;
    }

  xfer_len_line = s->n_colors_out * s->logical_width * s->bytes_per_pixel;
  xfer_len_in =
    s->n_colors_in * (s->logical_width - s->thumb_padx) * s->bytes_per_pixel +
    s->n_colors_in * s->odd_padding;

  DBG (30, "%s: begin: Input %d bytes/line, output %d bytes/line\n",
       __func__, xfer_len_in, xfer_len_line);

  if ((xfer_len_in & 0x3f))
    {
      int d = ((xfer_len_in / 512) * 512) + 512;
      s->block_padding = d - xfer_len_in;
    }

  DBG (30, "%s: block_padding = %d, odd_padding = %d\n",
       __func__, s->block_padding, s->odd_padding);

  DBG (30,
       "%s: colors = %d:%d, logical_width = %d, bytes_per_pixel = %d\n",
       __func__, s->n_colors_in, s->n_colors_out, s->logical_width,
       s->bytes_per_pixel);


  /* Do not change the behaviour of older models, pad to 512 */
  if (s->caps & CS9K_CAPS_PAD_BLOCK)
    {
      xfer_len_in += s->block_padding;
      if (xfer_len_in & 0x3f)
	DBG (1, "BUG: %s, not a multiple of 64. (0x%06lx)\n",
	     __func__, (long) xfer_len_in);
    }

  if ((s->xfer_position + xfer_len_line) > s->xfer_bytes_total)
    xfer_len_line = s->xfer_bytes_total - s->xfer_position;	/* just in case */

  if (s->preview && s->thumbnail && s->thumbnail_done
      && (size_t) s->thumb_pos_in != s->xfer_bytes_total)
    {
      DBG (1,
	   "%s: Thumbnail invalid, thumb_pos_in = %d, xfer_bytes_total = %d\n",
	   __func__, s->thumb_pos_in, s->xfer_bytes_total);
      s->thumbnail_done = 0;
      return SANE_STATUS_INVAL;
    }

  DBG (30, "%s: final: Input %d bytes/line, output %d bytes/line\n",
       __func__, xfer_len_in, xfer_len_line);


  /* ============================================================ */
  /* no more data - do any required post-processing */
  if (xfer_len_line == 0)
    {
      *len = 0;

      DBG (4, "%s: End of image detected.\n", __func__);

      /* Advance frame if appropriate */
      if ((s->n_colors_out < s->n_colors_in)
	  && (s->infrared == CS9K_INFRARED_IN))
	{
	  if (!s->preview)
	    {
	      s->infrared = CS9K_INFRARED_OUT;
	      DBG (8, "Infrared out mode\n");
	      s->option_flags |= SANE_INFO_RELOAD_OPTIONS;
	    }
	}
      else if (s->infrared == CS9K_INFRARED_OUT)
	{
	  s->infrared = CS9K_INFRARED_IN;
	  DBG (8, "Infrared in mode\n");
	  if (s->source)
	    {
	      s->c_frame++;
	      while (((s->frame_bounds[s->c_frame - 1].
		       flags & CS9K_FBF_PRESENT) == 0)
		     && (s->c_frame <= s->n_frames))
		s->c_frame++;
	    }
	  s->option_flags |= SANE_INFO_RELOAD_OPTIONS;
	}
      else if (s->source && !s->preview)
	{
	  s->c_frame++;
	  while (((s->frame_bounds[s->c_frame - 1].
		   flags & CS9K_FBF_PRESENT) == 0)
		 && (s->c_frame <= s->n_frames))
	    s->c_frame++;
	    s->option_flags |= SANE_INFO_RELOAD_OPTIONS;
	}

      /* Draw markup on thumbnail preview and evaluate exposure per frame */
      /* or evaluate exposure per frame */
      if (s->n_colors_in > 0)
	{
	  if (s->preview && s->thumbnail)
	    {
	      if (!s->thumbnail_done)
		{
		  if ((size_t) s->thumb_pos_in == s->xfer_bytes_total)
		    {
		      DBG (12, "%s: thumbnail done.\n", __func__);
		      s->thumbnail_done = 1;
		      cs9k_draw_bounds (s);
		    }
		  else
		    {
		      DBG (1,
			   "%s: thumbnail done, but transferred %d of %d bytes!\n",
			   __func__, s->thumb_pos_in, s->xfer_bytes_total);
		    }
		}
	    }
	  else
	    {
	      /* Generate histogram for exposure advice */
	      cs9k_analyze_histogram (s->histogram, s->n_colors_out,
				      s->histogram_size,
				      SANE_UNFIX (s->adj_exp_tgt),
				      &s->adj_exp_type, &s->adj_exp);
	      cs9k_set_adjust_message (s, CS9K_OPTION_ADJUST_EXPOSURE,
				       s->adj_exp, s->adj_exp_type);

	      if (s->n_colors_in > s->n_colors_out)
		{
		  cs9k_analyze_histogram (s->histogram + 3, 1,
					  s->histogram_size,
					  SANE_UNFIX (s->adj_exp_tgt_ir),
					  &s->adj_exp_type_ir,
					  &s->adj_exp_ir);
		  cs9k_set_adjust_message (s, CS9K_OPTION_ADJUST_EXPOSURE_IR,
					   s->adj_exp_ir, s->adj_exp_type_ir);
		}
	    }
	}
      s->infrared_index = 0;
      s->thumb_pos_out = 0;
      s->scanning = SANE_FALSE;
      return SANE_STATUS_EOF;
    }

  /* Done post-processing */
  /* ============================================================ */

  /* Adjust buffer to line length if needed */
  if (xfer_len_line != s->line_buf_n)
    {
      if (xfer_len_line > s->line_buf_size)
	{
	  line_buf_new =
	    (SANE_Byte *) cs9k_xrealloc (s->line_buf,
					 xfer_len_line * sizeof (SANE_Byte));
	  if (!line_buf_new)
	    {
	      *len = 0;
	      return SANE_STATUS_NO_MEM;
	    }
	  DBG (4, "Reallocated line buffer %d => %d bytes.\n",
	       s->line_buf_size, xfer_len_line);
	  s->line_buf = line_buf_new;
	  s->line_buf_size = xfer_len_line;
	}
      s->line_buf_n = xfer_len_line;
    }

  if (s->preview && s->thumbnail && s->thumbnail_done)
    {
      /* Return existing thumbnail data */
      memcpy (s->line_buf, &s->thumbnail_obuf[s->thumb_pos_out],
	      xfer_len_line);
      s->thumb_pos_out += xfer_len_line;
    }

  else if (s->infrared == CS9K_INFRARED_OUT)
    {
      /* Return IR data from earlier scan */
      memcpy (s->line_buf, s->infrared_buf + s->infrared_index,
	      s->logical_width * s->bytes_per_pixel);
      DBG (40,
	   "Copied %d bytes from infrared buffer at index %d; xfer_len_line = %d\n",
	   s->logical_width * s->bytes_per_pixel, s->infrared_index,
	   xfer_len_line);
      s->infrared_index += s->logical_width * s->bytes_per_pixel;
    }

  else if ((s->xfer_bytes_total - s->xfer_position) <=
	   (unsigned long) (s->thumb_pady * xfer_len_line))
    {
      /* Padding to match thumbnail to frame aspect ratio */
      memset (s->line_buf, 0x80, xfer_len_line);
    }

  else
    {
      if (s->preview)
	{
	  hist_y_inrange =
	    (s->xfer_position > s->xfer_bytes_total / HIST_FRAME_MARGIN)
	    && (s->xfer_position <
		s->xfer_bytes_total * (HIST_FRAME_MARGIN -
				       1) / HIST_FRAME_MARGIN);
	  hist_x_lowrange = s->logical_width / HIST_FRAME_MARGIN;
	  hist_x_highrange = s->logical_width * 9 / HIST_FRAME_MARGIN;
	}
      else
	{
	  hist_y_inrange = SANE_TRUE;
	  hist_x_lowrange = 0;
	  hist_x_highrange = s->logical_width - 1;
	}

      /* Fetch a line of image data from device */
      cs9k_scanner_ready (s, CS9K_STATUS_READY);
      cs9k_init_buffer (s);
      cs9k_parse_cmd (s, "28 00 00 00 00 00");
      cs9k_pack_byte (s, (xfer_len_in >> 16) & 0xff);
      cs9k_pack_byte (s, (xfer_len_in >> 8) & 0xff);
      cs9k_pack_byte (s, xfer_len_in & 0xff);
      cs9k_parse_cmd (s, "00");
      s->n_recv = xfer_len_in;

      status = cs9k_issue_cmd (s);
      if (status != SANE_STATUS_GOOD)
	{
	  *len = 0;
	  return status;
	}

      for (index = 0; index < s->logical_width; index++)
	{
	  for (color = 0; color < s->n_colors_out; color++)
	    {
	      SANE_Int srccolor = color + (s->infrared == CS9K_INFRARED_ONLY ?
			      s->n_colors_in - s->n_colors_out : 0);

	      dst_pel = s->bytes_per_pixel * (s->n_colors_out * index + color);
	      src_pel = s->bytes_per_pixel *
		        (srccolor * (s->logical_width - s->thumb_padx) +
			 (srccolor + 1) * s->odd_padding + index);

	      switch (s->bytes_per_pixel)
		{
		case 1:
		  {
		    if (s->preview && s->thumbnail)
		      s8 = (u_int8_t *) &
			   (s->thumbnail_ibuf[s->thumb_pos_in + dst_pel]);
		    else
		      s8 = (u_int8_t *) & (s->line_buf[dst_pel]);

		    if (index < (s->logical_width - s->thumb_padx))
		      {
			*s8 = s->recv_buf[src_pel];
			if (s->histogram_size && hist_y_inrange &&
			    (index >= hist_x_lowrange) &&
			    (index <= hist_x_highrange))
			  s->histogram[srccolor]
				      [(int) (log2 (*s8) * HIST_STEPS)]++;
		      }
		    else
		      *s8 = 0x2e;
		  }
		  break;

		case 2:
		  {
		    if (s->preview && s->thumbnail)
		      s16 = (u_int16_t *) &
			    (s->thumbnail_ibuf[s->thumb_pos_in + dst_pel]);
		    else
		      s16 = (u_int16_t *) & (s->line_buf[dst_pel]);

		    if (index < (s->logical_width - s->thumb_padx))
		      {
			*s16 = (s->recv_buf[src_pel] << 8) + s->recv_buf[src_pel + 1];
			if (s->histogram_size && hist_y_inrange &&
			    (index >= hist_x_lowrange) &&
			    (index <= hist_x_highrange))
			  s->histogram[srccolor][(int)
					      (log2 (*s16) * HIST_STEPS)]++;
			*s16 <<= s->shift_bits;
		      }
		    else
		      *s16 = 0x2e2e;
		  }
		  break;

		default:
		  DBG (1,
		       "BUG: sane_read(): Unknown number of bytes per pixel.\n");
		  *len = 0;
		  return SANE_STATUS_INVAL;
		  break;
		}
	    }

	  if ((s->n_colors_in > s->n_colors_out) &&
	      (s->infrared == CS9K_INFRARED_IN))
	    {
	      dst_pel = s->infrared_index + s->bytes_per_pixel * index;
	      src_pel = s->bytes_per_pixel *
		        (s->n_colors_out * (s->logical_width - s->thumb_padx) +
			 (s->n_colors_out + 1) * s->odd_padding + index);

	      switch (s->bytes_per_pixel)
		{
		case 1:
		  {
		    s8 = (u_int8_t *) & (s->infrared_buf[dst_pel]);
		    *s8 = s->recv_buf[src_pel];
		    if (s->histogram_size)
		      s->histogram[3][(int) (log2 (*s8) * HIST_STEPS)]++;
		  }
		  break;

		case 2:
		  {
		    s16 = (u_int16_t *) & (s->infrared_buf[dst_pel]);
		    *s16 = ((s->recv_buf[src_pel] << 8) +
			     s->recv_buf[src_pel + 1]);
		    if (s->histogram_size)
		      s->histogram[3][(int) (log2 (*s16) * HIST_STEPS)]++;
		    *s16 <<= s->shift_bits;
		  }
		  break;

		default:
		  DBG (1,
		       "BUG: sane_read(): Unknown number of bytes per pixel.\n");
		  *len = 0;
		  return SANE_STATUS_INVAL;
		  break;
		}
	    }
	}

      if ((s->n_colors_in > s->n_colors_out)
	  && (s->infrared == CS9K_INFRARED_IN))
	{
	  DBG (80,
	       "Copied %d bytes to infrared buffer + %d; xfer_len_line = %d\n",
	       s->logical_width * s->bytes_per_pixel, s->infrared_index,
	       xfer_len_line);
	  s->infrared_index += s->logical_width * s->bytes_per_pixel;
	}

      if (s->preview && s->thumbnail)
	{
	  s->thumb_pos_in += xfer_len_line;
	  /* draw frames on the fly */
	  cs9k_draw_bounds (s);
	  DBG (10, "Copying %d bytes %p -> %p\n", xfer_len_line,
	       &s->thumbnail_obuf[s->thumb_pos_out], s->line_buf);
	  memcpy (s->line_buf, &s->thumbnail_obuf[s->thumb_pos_out],
		  xfer_len_line);
	  s->thumb_pos_out += xfer_len_line;
	}
    }

  s->xfer_position += xfer_len_line;

  xfer_len_out = xfer_len_line;
  if (xfer_len_out > maxlen)
    xfer_len_out = maxlen;

  memcpy (buf, s->line_buf, xfer_len_out);
  if (xfer_len_out < xfer_len_line)
    s->line_buf_i = xfer_len_out;	/* data left in the line buffer, read out next time */

  *len = xfer_len_out;
  return SANE_STATUS_GOOD;
}

void
sane_cancel (SANE_Handle h)
{
  cs9k_t *s = (cs9k_t *) h;

  DBG (10, "%s, scanning = %d.\n", __func__, s->scanning);

  if (s->scanning)
    {
      cs9k_init_buffer (s);
      cs9k_parse_cmd (s, "c0 00 00 00 00 00");
      cs9k_issue_cmd (s);
    }

  s->scanning = SANE_FALSE;
}

SANE_Status
sane_set_io_mode (SANE_Handle h, SANE_Bool m)
{
  cs9k_t *s = (cs9k_t *) h;

  DBG (10, "%s\n", __func__);

  if (!s->scanning)
    return SANE_STATUS_INVAL;
  if (m == SANE_FALSE)
    return SANE_STATUS_GOOD;
  else
    return SANE_STATUS_UNSUPPORTED;
}

SANE_Status
sane_get_select_fd (SANE_Handle h, SANE_Int * fd)
{
  cs9k_t *s = (cs9k_t *) h;

  DBG (10, "%s\n", __func__);

  fd = fd;			/* to shut up compiler */
  s = s;			/* to shut up compiler */

  return SANE_STATUS_UNSUPPORTED;
}


/* ========================================================================= */
/* private functions */

static void
cs9k_trim (char *s)
{
  int i, l = strlen (s);

  for (i = l - 1; i > 0; i--)
    {
      if (s[i] == ' ')
	s[i] = '\0';
      else
	break;
    }
}

static SANE_Status
cs9k_open (const char *device, cs9k_interface_t interface, cs9k_t ** sp)
{
  SANE_Status status;
  cs9k_t *s;
  char *prefix = NULL, *line;
  int i;
  int alloc_failed = 0;
  SANE_Device **device_list_new;

  DBG (6, "%s, device = %s, interface = %i\n", __func__, device, interface);

  if (!strncmp (device, "auto", 5))
    {
      try_interface = CS9K_INTERFACE_SCSI;
      sanei_config_attach_matching_devices ("scsi Nikon *", cs9k_attach);
      try_interface = CS9K_INTERFACE_USB;
      sanei_usb_attach_matching_devices ("usb 0x04b0 0x4000", cs9k_attach);
      sanei_usb_attach_matching_devices ("usb 0x04b0 0x4001", cs9k_attach);
      sanei_usb_attach_matching_devices ("usb 0x04b0 0x4002", cs9k_attach);
      return SANE_STATUS_GOOD;
    }

  if ((s = (cs9k_t *) cs9k_xmalloc (sizeof (cs9k_t))) == NULL)
    return SANE_STATUS_NO_MEM;
  memset (s, 0, sizeof (cs9k_t));

  s->send_buf = s->recv_buf = NULL;
  s->send_buf_size = s->recv_buf_size = 0;
  s->depth_list = (SANE_Word *) cs9k_xmalloc (9 * sizeof (SANE_Word));
  if (!s->depth_list)
    {
      DBG (4, "Error: could not allocate memory for resolution list.\n");
      cs9k_xfree (s);
      return SANE_STATUS_NO_MEM;
    }

  switch (interface)
    {
    case CS9K_INTERFACE_UNKNOWN:
      for (i = 0; i < 2; i++)
	{
	  switch (i)
	    {
	    case 1:
	      prefix = "usb:";
	      try_interface = CS9K_INTERFACE_USB;
	      break;
	    default:
	      prefix = "scsi:";
	      try_interface = CS9K_INTERFACE_SCSI;
	      break;
	    }
	  if (!strncmp (device, prefix, strlen (prefix)))
	    {
	      const void *p = device + strlen (prefix);
	      cs9k_xfree (s->depth_list);
	      cs9k_xfree (s);
	      return cs9k_open (p, try_interface, sp);
	    }
	}
      cs9k_xfree (s->depth_list);
      cs9k_xfree (s);
      return SANE_STATUS_INVAL;
      break;
    case CS9K_INTERFACE_SCSI:
      s->interface = CS9K_INTERFACE_SCSI;
      DBG (6,
	   "%s, trying to open %s, assuming SCSI or SBP2 interface\n",
	   __func__, device);
      status = sanei_scsi_open (device, &s->fd, cs9k_scsi_sense_handler, s);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (6, " ...failed: %s.\n", sane_strstatus (status));
	  cs9k_xfree (s->depth_list);
	  cs9k_xfree (s);
	  return status;
	}
      break;
    case CS9K_INTERFACE_USB:
      s->interface = CS9K_INTERFACE_USB;
      DBG (6, "%s, trying to open %s, assuming USB interface\n",
	   __func__, device);
      status = sanei_usb_open (device, &s->fd);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (6, " ...failed: %s.\n", sane_strstatus (status));
	  cs9k_xfree (s->depth_list);
	  cs9k_xfree (s);
	  return status;
	}
      break;
    }

  open_devices++;
  DBG (6, "%s, trying to identify device.\n", __func__);

  /* identify scanner */
  status = cs9k_page_inquiry (s, -1);
  if (status != SANE_STATUS_GOOD)
    {
      cs9k_close (s);
      return status;
    }

  strncpy (s->vendor_string, (char *) s->recv_buf + 8, 8);
  s->vendor_string[8] = '\0';
  strncpy (s->product_string, (char *) s->recv_buf + 16, 16);
  s->product_string[16] = '\0';
  strncpy (s->revision_string, (char *) s->recv_buf + 32, 4);
  s->revision_string[4] = '\0';

  DBG (10,
       "%s, vendor = '%s', product = '%s', revision = '%s'.\n",
       __func__, s->vendor_string, s->product_string, s->revision_string);

  s->caps = CS9K_CAPS_PAD_WORD;	/* Assume word aligned data */
  s->caps |= CS9K_CAPS_INDEP_XY;
  if (s->recv_buf[1] & 0x80)	/* RMB - Removable Medium Bit */
    s->caps |= CS9K_CAPS_HOLDER;

  if (!strncmp (s->product_string, "COOLSCANIII     ", 16))
    {
      s->type = CS9K_TYPE_LS30;
      s->caps &= ~CS9K_CAPS_PAD_WORD;
    }
  else if (!strncmp (s->product_string, "LS-40 ED        ", 16))
    {
      s->type = CS9K_TYPE_LS40;
      s->caps |= CS9K_CAPS_CTL_VS10;
    }
  else if (!strncmp (s->product_string, "LS-50 ED        ", 16))
    {
      s->type = CS9K_TYPE_LS50;
      s->caps |= CS9K_CAPS_PAD_BLOCK;
      s->caps |= CS9K_CAPS_CTL_VS10;
    }
  else if (!strncmp (s->product_string, "LS-2000         ", 16))
    {
      s->type = CS9K_TYPE_LS2000;
      s->caps &= ~CS9K_CAPS_PAD_WORD;
    }
  else if (!strncmp (s->product_string, "LS-4000 ED      ", 16))
    {
      s->type = CS9K_TYPE_LS4000;
      s->caps |= CS9K_CAPS_CTL_VS10;
    }
  else if (!strncmp (s->product_string, "LS-5000 ED      ", 16))
    {
      s->type = CS9K_TYPE_LS5000;
      s->caps |= CS9K_CAPS_PAD_BLOCK;
      s->caps |= CS9K_CAPS_CTL_VS10;
    }
  else if (!strncmp (s->product_string, "LS-8000 ED      ", 16))
    {
      s->type = CS9K_TYPE_LS8000;
    }
  else if (!strncmp (s->product_string, "LS-9000 ED      ", 16))
    {
      s->type = CS9K_TYPE_LS9000;
      s->caps |= CS9K_CAPS_NEGATIVE;
      s->caps &= ~CS9K_CAPS_INDEP_XY;	/* Can't find capability bit */
    }
  else if (!strncmp (s->vendor_string, "Nikon", 5)
	   && !strncmp (s->product_string, "LS", 2))
    {
      s->type = CS9K_TYPE_LS_X;
    }

  if (s->type != CS9K_TYPE_UNKOWN)
    DBG (10,
	 "%s, device identified as coolscan9k type #%i.\n", __func__,
	 s->type);
  else
    {
      DBG (10, "%s, device not identified.\n", __func__);
      cs9k_close (s);
      return SANE_STATUS_UNSUPPORTED;
    }

  cs9k_trim (s->vendor_string);
  cs9k_trim (s->product_string);
  cs9k_trim (s->revision_string);

  if (sp)
    *sp = s;
  else
    {
      device_list_new =
	(SANE_Device **) cs9k_xrealloc (device_list,
					(n_device_list +
					 2) * sizeof (SANE_Device *));
      if (!device_list_new)
	return SANE_STATUS_NO_MEM;
      device_list = device_list_new;
      device_list[n_device_list] =
	(SANE_Device *) cs9k_xmalloc (sizeof (SANE_Device));
      if (!device_list[n_device_list])
	return SANE_STATUS_NO_MEM;
      switch (interface)
	{
	case CS9K_INTERFACE_UNKNOWN:
	  DBG (1, "BUG: cs9k_open(): unknown interface.\n");
	  cs9k_close (s);
	  return SANE_STATUS_UNSUPPORTED;
	  break;
	case CS9K_INTERFACE_SCSI:
	  prefix = "scsi:";
	  break;
	case CS9K_INTERFACE_USB:
	  prefix = "usb:";
	  break;
	}

      line = (char *) cs9k_xmalloc (strlen (device) + strlen (prefix) + 1);
      if (!line)
	alloc_failed = 1;
      else
	{
	  strcpy (line, prefix);
	  strcat (line, device);
	  device_list[n_device_list]->name = line;
	}

      line = (char *) cs9k_xmalloc (strlen (s->vendor_string) + 1);
      if (!line)
	alloc_failed = 1;
      else
	{
	  strcpy (line, s->vendor_string);
	  device_list[n_device_list]->vendor = line;
	}

      line = (char *) cs9k_xmalloc (strlen (s->product_string) + 1);
      if (!line)
	alloc_failed = 1;
      else
	{
	  strcpy (line, s->product_string);
	  device_list[n_device_list]->model = line;
	}

      device_list[n_device_list]->type = "film scanner";

      if (alloc_failed)
	{
	  cs9k_xfree (device_list[n_device_list]->name);
	  cs9k_xfree (device_list[n_device_list]->vendor);
	  cs9k_xfree (device_list[n_device_list]->model);
	  cs9k_xfree (device_list[n_device_list]);
	}
      else
	n_device_list++;
      device_list[n_device_list] = NULL;

      cs9k_close (s);
    }

  return SANE_STATUS_GOOD;
}

void
cs9k_close (cs9k_t * s)
{
  cs9k_xfree (s->lut_r);
  cs9k_xfree (s->lut_g);
  cs9k_xfree (s->lut_b);
  cs9k_xfree (s->lut_neutral);
  cs9k_xfree (s->line_buf);
  cs9k_xfree (s->depth_list);
  cs9k_xfree (s->histogram[0]);
  cs9k_xfree (s->thumbnail_ibuf);
  cs9k_xfree (s->frame_bounds);
  cs9k_xfree (s->resx_list);
  cs9k_xfree (s->resy_list);
  cs9k_xfree (s->res_thumb_list);

  switch (s->interface)
    {
    case CS9K_INTERFACE_UNKNOWN:
      DBG (0, "BUG: %s: Unknown interface number.\n", __func__);
      break;
    case CS9K_INTERFACE_SCSI:
      sanei_scsi_close (s->fd);
      open_devices--;
      break;
    case CS9K_INTERFACE_USB:
      sanei_usb_close (s->fd);
      open_devices--;
      break;
    }

  cs9k_xfree (s->send_buf);
  cs9k_xfree (s->recv_buf);
  cs9k_xfree (s);
}

static SANE_Status
cs9k_attach (const char *dev)
{
  SANE_Status status;

  if (try_interface == CS9K_INTERFACE_UNKNOWN)
    return SANE_STATUS_UNSUPPORTED;

  status = cs9k_open (dev, try_interface, NULL);
  return status;
}

static SANE_Status
cs9k_scsi_sense_handler (int fd, u_char * sense_buffer, void *arg)
{
  cs9k_t *s = (cs9k_t *) arg;

  fd = fd;			/* to shut up compiler */

  s->sense_key = sense_buffer[2] & 0x0f;
  s->sense_asc = sense_buffer[12];
  s->sense_ascq = sense_buffer[13];
  s->sense_info = cs9k_field2val (sense_buffer + 3, 4);
  s->sense_sks = cs9k_field2val (sense_buffer + 15, 3);

  return cs9k_parse_sense_data (s);
}

static SANE_Status
cs9k_parse_sense_data (cs9k_t * s)
{
  SANE_Status status = SANE_STATUS_GOOD;

  s->sense_code =
    (s->sense_key << 24) + (s->sense_asc << 16) + (s->sense_ascq << 8);

  if (s->sense_key)
    {
      DBG (14, "sense code: %02lx-%02lx-%02lx-%02lx\n", s->sense_key,
	   s->sense_asc, s->sense_ascq, s->sense_info);
      cs9k_decode_sense (s->sense_key, s->sense_asc, s->sense_ascq,
			 s->sense_info, s->sense_sks);
    }

  switch (s->sense_key)
    {
    case 0x00:
      s->status = CS9K_STATUS_READY;
      break;

    case 0x02:
      switch (s->sense_asc)
	{
	case 0x04:
	  DBG (15, " processing\n");
	  s->status = CS9K_STATUS_PROCESSING;
	  break;
	case 0x3a:
	  DBG (15, " no docs\n");
	  cs9k_seq_begin (s);
	  s->status = CS9K_STATUS_NO_DOCS;
	  break;
	default:
	  DBG (15, " default\n");
	  s->status = CS9K_STATUS_ERROR;
	  status = SANE_STATUS_IO_ERROR;
	  break;
	}
      break;

    case 0x06:
      cs9k_seq_begin (s);
      s->medium_changed = 1;
      DBG (4, "Detected medium change\n");
      break;

    case 0x09:
      if ((s->sense_code == 0x09800600) || (s->sense_code == 0x09800601))
	DBG (4, "Re-issuing command\n");
      s->status = CS9K_STATUS_REISSUE;
      break;

    default:
      s->status = CS9K_STATUS_ERROR;
      status = SANE_STATUS_IO_ERROR;
      break;
    }

  return status;
}

static void
cs9k_init_buffer (cs9k_t * s)
{
  s->n_cmd = 0;
  s->n_send = 0;
  s->n_recv = 0;
}

static SANE_Status
cs9k_pack_byte (cs9k_t * s, SANE_Byte byte)
{
  while (s->send_buf_size <= s->n_send)
    {
      s->send_buf_size += 16;
      s->send_buf =
	(SANE_Byte *) cs9k_xrealloc (s->send_buf, s->send_buf_size);
      if (!s->send_buf)
	return SANE_STATUS_NO_MEM;
    }

  s->send_buf[s->n_send++] = byte;

  return SANE_STATUS_GOOD;
}

static void
cs9k_pack_3byte (cs9k_t * s, SANE_Word word)
{
  cs9k_pack_byte (s, (word >> 16) & 0xff);
  cs9k_pack_byte (s, (word >> 8) & 0xff);
  cs9k_pack_byte (s, word & 0xff);
}

static void
cs9k_pack_long (cs9k_t * s, unsigned long val)
{
  cs9k_pack_byte (s, (val >> 24) & 0xff);
  cs9k_pack_byte (s, (val >> 16) & 0xff);
  cs9k_pack_byte (s, (val >> 8) & 0xff);
  cs9k_pack_byte (s, val & 0xff);
}

static void
cs9k_pack_word (cs9k_t * s, unsigned long val)
{
  cs9k_pack_byte (s, (val >> 8) & 0xff);
  cs9k_pack_byte (s, val & 0xff);
}

static SANE_Status
cs9k_parse_cmd (cs9k_t * s, char *text)
{
  size_t i, j;
  char c, h;
  SANE_Status status;

  for (i = 0; i < strlen (text); i += 2)
    if (text[i] == ' ')
      i--;			/* a bit dirty... advance by -1+2=1 */
    else
      {
	if ((!isxdigit (text[i])) || (!isxdigit (text[i + 1])))
	  DBG (1, "BUG: cs9k_parse_cmd(): Parser got invalid character.\n");
	c = 0;
	for (j = 0; j < 2; j++)
	  {
	    h = tolower (text[i + j]);
	    if ((h >= 'a') && (h <= 'f'))
	      c += 10 + h - 'a';
	    else
	      c += h - '0';
	    if (j == 0)
	      c <<= 4;
	  }
	status = cs9k_pack_byte (s, c);
	if (status != SANE_STATUS_GOOD)
	  return status;
      }

  return SANE_STATUS_GOOD;
}

static SANE_Status
cs9k_grow_send_buffer (cs9k_t * s)
{
  if (s->n_send > s->send_buf_size)
    {
      s->send_buf_size = s->n_send;
      s->send_buf =
	(SANE_Byte *) cs9k_xrealloc (s->send_buf, s->send_buf_size);
      if (!s->send_buf)
	return SANE_STATUS_NO_MEM;
    }

  return SANE_STATUS_GOOD;
}

static SANE_Status
cs9k_issue_cmd (cs9k_t * s)
{
  SANE_Status status = SANE_STATUS_INVAL;
  size_t n_data, n_status;
  static SANE_Byte status_buf[8];
  int status_only = 0;
  int i;

  DBG (20,
       "cs9k_issue_cmd(): opcode = 0x%02x, n_send = %lu, n_recv = %lu.\n",
       s->send_buf[0], (unsigned long) s->n_send, (unsigned long) s->n_recv);
  DBG (80, "Send: ");
  for (i = 0; i < (int) s->n_send; i++)
    DBG (80, "%02x ", s->send_buf[i]);
  DBG (80, "\n");

  s->status = CS9K_STATUS_READY;

  if (!s->n_cmd)
    switch (s->send_buf[0])
      {
      case 0x00:
      case 0x12:
      case 0x15:
      case 0x16:
      case 0x17:
      case 0x1a:
      case 0x1b:
      case 0x1c:
      case 0x1d:
      case 0xc0:
      case 0xc1:
	s->n_cmd = 6;
	break;
      case 0x24:
      case 0x25:
      case 0x28:
      case 0x2a:
      case 0xe0:
      case 0xe1:
	s->n_cmd = 10;
	break;
      default:
	DBG (1,
	     "BUG: cs9k_issue_cmd(): Unknown command opcode 0x%02x.\n",
	     s->send_buf[0]);
	break;
      }

  if (s->n_send < s->n_cmd)
    {
      DBG (1,
	   "BUG: cs9k_issue_cmd(): Negative number of data out bytes requested.\n");
      return SANE_STATUS_INVAL;
    }

  n_data = s->n_send - s->n_cmd;
  if (s->n_recv > 0)
    {
      if (n_data > 0)
	{
	  DBG (1,
	       "BUG: cs9k_issue_cmd(): Both data in and data out requested.\n");
	  return SANE_STATUS_INVAL;
	}
      else
	{
	  n_data = s->n_recv;
	}
    }

  s->recv_buf = (SANE_Byte *) cs9k_xrealloc (s->recv_buf, s->n_recv);
  if (!s->recv_buf)
    return SANE_STATUS_NO_MEM;

  switch (s->interface)
    {
    case CS9K_INTERFACE_UNKNOWN:
      DBG (1,
	   "BUG: cs9k_issue_cmd(): Unknown or uninitialized interface number.\n");
      break;

    case CS9K_INTERFACE_SCSI:
      status =
	sanei_scsi_cmd2 (s->fd, s->send_buf, s->n_cmd,
			 s->send_buf + s->n_cmd, s->n_send - s->n_cmd,
			 s->recv_buf, &s->n_recv);
      /* status = SANE_STATUS_GOOD; */
      break;

    case CS9K_INTERFACE_USB:
      status = sanei_usb_write_bulk (s->fd, s->send_buf, &s->n_cmd);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (1, "Error: cs9k_issue_cmd(): Could not write command.\n");
	  return SANE_STATUS_IO_ERROR;
	}

      switch (cs9k_phase_check (s))
	{
	case CS9K_PHASE_OUT:
	  if (s->n_send - s->n_cmd < n_data || !n_data)
	    {
	      DBG (4,
		   "Error: cs9k_issue_cmd(): Unexpected data out phase.\n");
	      return SANE_STATUS_IO_ERROR;
	    }
	  status = sanei_usb_write_bulk (s->fd,
					 s->send_buf + s->n_cmd, &n_data);
	  break;

	case CS9K_PHASE_IN:
	  if (s->n_recv < n_data || !n_data)
	    {
	      DBG (4, "Error: cs9k_issue_cmd(): Unexpected data in phase.\n");
	      return SANE_STATUS_IO_ERROR;
	    }
	  status = sanei_usb_read_bulk (s->fd, s->recv_buf, &n_data);
	  s->n_recv = n_data;
	  break;

	case CS9K_PHASE_NONE:
	  DBG (4, "%s: No command received!\n", __func__);
	  return SANE_STATUS_IO_ERROR;

	default:
	  if (n_data)
	    {
	      DBG (4,
		   "%s: Unexpected non-data phase, but n_data != 0 (%lu).\n",
		   __func__, (u_long) n_data);
	      status_only = 1;
	    }
	  break;
	}

      n_status = 8;
      status = sanei_usb_read_bulk (s->fd, status_buf, &n_status);
      if (n_status != 8)
	{
	  DBG (4,
	       "Error: cs9k_issue_cmd(): Failed to read 8 status bytes from USB.\n");
	  return SANE_STATUS_IO_ERROR;
	}

      s->sense_key = status_buf[1] & 0x0f;
      s->sense_asc = status_buf[2] & 0xff;
      s->sense_ascq = status_buf[3] & 0xff;
      s->sense_info = status_buf[4] & 0xff;
      status = cs9k_parse_sense_data (s);
      break;
    }

  if (!status)
    {
      DBG (80, "Recv %d: ", (int) s->n_recv);
      for (i = 0; i < (int) s->n_recv && i < 512; i++)
	DBG (80, "%02x ", s->recv_buf[i]);
      DBG (80, "\n");
    }
  if (status_only)
    return SANE_STATUS_IO_ERROR;
  else
    return status;
}

static cs9k_phase_t
cs9k_phase_check (cs9k_t * s)
{
  static SANE_Byte phase_send_buf[1] = { 0xd0 }, phase_recv_buf[1];
  SANE_Status status = 0;
  size_t n = 1;

  status = sanei_usb_write_bulk (s->fd, phase_send_buf, &n);
  status |= sanei_usb_read_bulk (s->fd, phase_recv_buf, &n);

  DBG (40, "%s: returned phase = 0x%02x.\n", __func__, phase_recv_buf[0]);

  if (status != SANE_STATUS_GOOD)
    return -1;
  else
    return phase_recv_buf[0];
}

static SANE_Status
cs9k_scanner_ready (cs9k_t * s, int flags)
{
  SANE_Status status = SANE_STATUS_GOOD;
  int i = -1;
  unsigned long count = 0;
  int retry = 5;

  do
    {
      if (i >= 0)		/* dirty !!! */
	usleep (1000000);
      /* test unit ready */
      cs9k_init_buffer (s);
      for (i = 0; i < 6; i++)
	cs9k_pack_byte (s, 0x00);

      status = cs9k_issue_cmd (s);
      if (status != SANE_STATUS_GOOD)
	if (--retry < 0)
	  return status;

      if (++count > 120)
	{			/* 120s timeout */
	  DBG (4, "Error: %s: Timeout expired.\n", __func__);
	  status = SANE_STATUS_IO_ERROR;
	  break;
	}
    }
  while (s->status & ~flags);	/* until all relevant bits are 0 */

  return status;
}

SANE_Word
cs9k_option_set_value (cs9k_t * s, cs9k_option_t n, void *v)
{
  SANE_Word flags = 0;
  cs9k_pixel_t pixel;
  SANE_Word val = 0;
  SANE_Option_Descriptor o = s->option_list[n];

  switch (o.type)
    {
    case SANE_TYPE_BOOL:
      if ((*(SANE_Word *) v != SANE_TRUE) && (*(SANE_Word *) v != SANE_FALSE))
	return SANE_STATUS_INVAL;
      val = *(SANE_Word *) v;
      DBG (12, "%s: set %2d (%s) to %s\n", __func__, n, o.title,
	   val ? "SANE_TRUE" : "SANE_FALSE");
      break;

    case SANE_TYPE_INT:
    case SANE_TYPE_FIXED:
      switch (o.constraint_type)
	{
	case SANE_CONSTRAINT_RANGE:
	  flags |= cs9k_adjustWord2optrange (s, (SANE_Word *) v, n);
	  break;
	case SANE_CONSTRAINT_WORD_LIST:
	  flags |=
	    cs9k_adjustWord2list ((SANE_Word *) v, o.constraint.word_list,
				  "option_set_value INT/FIXED");
	  break;
	default:
	  DBG (4, "Unknown constraint type %d.\n", o.constraint_type);
	  break;
	}
      val = *(SANE_Word *) v;
      if (o.type == SANE_TYPE_INT)
        DBG (12, "%s: set %2d (%s) to 0x%x\n", __func__, n, o.title, val);
      else
        DBG (12, "%s: set %2d (%s) to %.3f\n", __func__, n, o.title, SANE_UNFIX(val));
      break;

    case SANE_TYPE_STRING:
      if (o.constraint_type == SANE_CONSTRAINT_STRING_LIST)
	flags |=
	  cs9k_convertString2val ((SANE_String) v, o.constraint.string_list,
				  &val);
      else
	DBG (1,
	     "%s: Unrecognized constraint (%d) for string control %d (%s)\n",
	     __func__, o.constraint_type, n, o.title);
      DBG (12, "%s: set %2d (%s) to %s\n", __func__, n, o.title, (char *) v);
      break;

    case SANE_TYPE_BUTTON:
      break;

    case SANE_TYPE_GROUP:
      DBG (1, "%s: Attempt to write group %d (%s)\n", __func__, n, o.title);
      return SANE_STATUS_INVAL;

    default:
      DBG (1, "%s: Unrecognized constraint type %d.\n", __func__, o.type);
      break;
    }

  switch (n)
    {
    case CS9K_OPTION_NUM:
      return SANE_STATUS_INVAL;
      break;

    case CS9K_OPTION_NEGATIVE:
      if ((s->negative && !val) || (!s->negative && val))
	{
	  s->negative = val;
	  flags |= SANE_INFO_RELOAD_OPTIONS;	/* Exposure values change */
	}
      break;

    case CS9K_OPTION_INFRARED:
      if ((val == SANE_TRUE && s->infrared == CS9K_INFRARED_OFF)
	  || (val == SANE_FALSE && s->infrared != CS9K_INFRARED_OFF))
	{
	  if (val)
	    {
	      switch (s->infrared_fmt)
		{
		case 0:	/* 2 images, RGB + GRAY */
		  s->infrared = CS9K_INFRARED_IN;
		  break;
		case 1:	/* RGBI */
		  s->infrared = CS9K_INFRARED_4PLANE;
		  break;
		case 2:	/* RGBI */
		  s->infrared = CS9K_INFRARED_ONLY;
		  break;
		default:
		  DBG (1,
		       "%s: ERROR - Unknown IR format %d; disabling infrared.\n",
		       __func__, s->infrared_fmt);
		  s->infrared = CS9K_INFRARED_OFF;
		  break;
		}
	    }
	  else
	    s->infrared = CS9K_INFRARED_OFF;
	  cs9k_seq_begin (s);
	  flags |=
	    cs9k_set_visible (s, CS9K_OPTION_INFRARED_FMT,
			      s->infrared != CS9K_INFRARED_OFF);
	  flags |=
	    cs9k_set_visible (s, CS9K_OPTION_EXPOSURE_I,
			      s->infrared != CS9K_INFRARED_OFF);
	  flags |=
	    cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_TARGET_IR,
			      s->infrared != CS9K_INFRARED_OFF);
	  flags |=
	    cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_TEXT_IR,
			      s->infrared != CS9K_INFRARED_OFF);
	  flags |=
	    cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_IR,
			      s->adj_exp_ir > 0.02 &&
			      s->infrared != CS9K_INFRARED_OFF);
	  DBG (10, "%s: Toggle infrared to type %d.\n", __func__,
	       s->infrared);
	}
      break;

    case CS9K_OPTION_INFRARED_FMT:
      if (val != s->infrared_fmt)
	{
	  s->infrared_fmt = val;
	  switch (s->infrared_fmt)
	    {
	    case 0:		/* 2 images, RGB + GRAY */
	      s->infrared = CS9K_INFRARED_IN;
	      break;
	    case 1:		/* RGBI */
	      s->infrared = CS9K_INFRARED_4PLANE;
	      break;
	    case 2:	/* RGBI */
	      s->infrared = CS9K_INFRARED_ONLY;
	      break;
	    default:
	      DBG (1,
		   "%s: ERROR - Unknown IR format %d; disabling infrared.\n",
		   __func__, s->infrared_fmt);
	      s->infrared = CS9K_INFRARED_OFF;
	      break;
	    }
	  s->infrared_index = 0;
	  flags |= SANE_INFO_RELOAD_PARAMS;
	}
      break;

    case CS9K_OPTION_DEPTH:
      if (s->depth != val)
	{
	  s->depth = val;
	  flags |= SANE_INFO_RELOAD_PARAMS;
	}
      break;

    case CS9K_OPTION_PREVIEW:
      s->preview = val;
      break;

    case CS9K_OPTION_AUTOLOAD:
      if (s->exe_dx & CS9K_EXE_D1_LOAD)
	s->autoload = val;
      else
	s->autoload = SANE_FALSE;
      break;

    case CS9K_OPTION_EXPOSURE:
      if (s->negative)
	{
	  if (s->exposure != s->ranges[CS9K_OPTION_EXPOSURE].max - val)
	    {
	      s->exposure = s->ranges[CS9K_OPTION_EXPOSURE].max - val;
	      cs9k_seq_begin (s);
	    }
	}
      else
	{
	  if (s->exposure != val)
	    {
	      s->exposure = val;
	      cs9k_seq_begin (s);
	    }
	}
      break;
    case CS9K_OPTION_EXPOSURE_R:
      if (s->negative)
	{
	  if (s->exposure_r != -val)
	    {
	      s->exposure_r = -val;
	      cs9k_seq_begin (s);
	    }
	}
      else
	{
	  if (s->exposure_r != val)
	    {
	      s->exposure_r = val;
	      cs9k_seq_begin (s);
	    }
	}
      break;
    case CS9K_OPTION_EXPOSURE_G:
      if (s->negative)
	{
	  if (s->exposure_g != -val)
	    {
	      s->exposure_g = -val;
	      cs9k_seq_begin (s);
	    }
	}
      else
	{
	  if (s->exposure_g != val)
	    {
	      s->exposure_g = val;
	      cs9k_seq_begin (s);
	    }
	}
      break;
    case CS9K_OPTION_EXPOSURE_B:
      if (s->negative)
	{
	  if (s->exposure_b != -val)
	    {
	      s->exposure_b = -val;
	      cs9k_seq_begin (s);
	    }
	}
      else
	{
	  if (s->exposure_b != val)
	    {
	      s->exposure_b = val;
	      cs9k_seq_begin (s);
	    }
	}
      break;
    case CS9K_OPTION_EXPOSURE_I:
      if (s->negative)
	{
	  if (s->exposure_i != s->ranges[n].max - val)
	    {
	      s->exposure_i = s->ranges[n].max - val;
	      cs9k_seq_begin (s);
	    }
	}
      else
	{
	  if (s->exposure_i != val)
	    {
	      s->exposure_i = val;
	      cs9k_seq_begin (s);
	    }
	}
      break;

    case CS9K_OPTION_EXPOSURE_PER_FRAME:
      s->exposure_per_frame = val;
      break;
    case CS9K_OPTION_ADJUST_EXPOSURE_TARGET:
      s->adj_exp_tgt = val;
      break;
    case CS9K_OPTION_ADJUST_EXPOSURE_TARGET_IR:
      s->adj_exp_tgt_ir = val;
      break;

    case CS9K_OPTION_ADJUST_EXPOSURE_TEXT:
      return SANE_STATUS_INVAL;
    case CS9K_OPTION_ADJUST_EXPOSURE:
      if (s->adj_exp != 0)
	{
	  DBG (12,
	       "%s: User requested adjust exposure by %.2f; was exposure = %.2f\n",
	       __func__, SANE_UNFIX (s->adj_exp), SANE_UNFIX (s->exposure));
	  s->exposure += s->adj_exp;
	  cs9k_adjustWord2optrange (s, &s->exposure, CS9K_OPTION_EXPOSURE);
	  s->adj_exp = SANE_FIX (0.0);
	  DBG (12,
	       "%s: User requested adjust exposure by %.2f; now exposure = %.2f\n",
	       __func__, SANE_UNFIX (s->adj_exp), SANE_UNFIX (s->exposure));
	  cs9k_set_adjust_message (s, CS9K_OPTION_ADJUST_EXPOSURE, s->adj_exp,
				   s->adj_exp_type);
	  flags |= SANE_INFO_RELOAD_OPTIONS;
	  cs9k_seq_begin (s);
	}
      break;
    case CS9K_OPTION_ADJUST_EXPOSURE_TEXT_IR:
      return SANE_STATUS_INVAL;
    case CS9K_OPTION_ADJUST_EXPOSURE_IR:
      if (s->adj_exp_ir != 0)
	{
	  s->exposure_i += s->adj_exp_ir;
	  cs9k_adjustWord2optrange (s, &s->exposure_i,
				    CS9K_OPTION_EXPOSURE_I);
	  s->adj_exp_ir = SANE_FIX (0.0);
	  cs9k_seq_begin (s);
	}
      flags |=
	cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_IR, SANE_FALSE);
      break;

    case CS9K_OPTION_LUT_R:
      if (!(s->lut_r))
	return SANE_STATUS_INVAL;
      for (pixel = 0; pixel < s->n_lut; pixel++)
	s->lut_r[pixel] = ((SANE_Word *) v)[pixel];
      break;
    case CS9K_OPTION_LUT_G:
      if (!(s->lut_g))
	return SANE_STATUS_INVAL;
      for (pixel = 0; pixel < s->n_lut; pixel++)
	s->lut_g[pixel] = ((SANE_Word *) v)[pixel];
      break;
    case CS9K_OPTION_LUT_B:
      if (!(s->lut_b))
	return SANE_STATUS_INVAL;
      for (pixel = 0; pixel < s->n_lut; pixel++)
	s->lut_b[pixel] = ((SANE_Word *) v)[pixel];
      break;

    case CS9K_OPTION_REFRESH:
      if (val > 0)
	{
	  if (val > 1)
	    {
	      cs9k_seq_begin (s);
	      cs9k_full_inquiry (s, &flags);
	    }
	  flags |= SANE_INFO_RELOAD_OPTIONS | SANE_INFO_INEXACT;
	}
      break;
    case CS9K_OPTION_LOAD:
      cs9k_load (s);
      break;
    case CS9K_OPTION_EJECT:
      cs9k_eject (s);
      break;
    case CS9K_OPTION_RESET:
      cs9k_reset (s);
      cs9k_seq_begin (s);
      cs9k_full_inquiry (s, &flags);
      break;

    case CS9K_OPTION_SOURCE:
      if (s->source != val)
	{
	  s->source = val;
	  s->c_frame = s->i_frame;	/* Reset current frame pointer */
	  flags |= cs9k_set_visible (s, CS9K_OPTION_FRAME, s->n_frames > 1
				     && (!s->thumbnail || s->source));
	  flags |= cs9k_set_visible (s, CS9K_OPTION_EXPOSURE_PER_FRAME, s->n_frames > 1
				     && s->thumbnail && s->source);
	}
      break;

    case CS9K_OPTION_HOLDER:
      return SANE_STATUS_UNSUPPORTED;

    case CS9K_OPTION_NEXTSCAN:
      return SANE_STATUS_UNSUPPORTED;

    case CS9K_OPTION_SCAN_THUMBNAILS:
      if (s->thumbnail != val)
	{
	  s->thumbnail = val;
	  s->thumbnail_done = 0;
	  s->thumb_pos_in = 0;
	  flags |= cs9k_set_visible (s, CS9K_OPTION_FRAME, s->n_frames > 1
				     && (!s->thumbnail || s->source));
	  flags |= cs9k_set_visible (s, CS9K_OPTION_EXPOSURE_PER_FRAME,
				     s->n_frames > 1 &&
				     s->thumbnail && s->source);
	}
      break;

    case CS9K_OPTION_FRAME:
      if (s->i_frame != val)
	{
	  s->i_frame = val;
	  s->c_frame = s->i_frame;
	  flags |=
	    cs9k_set_range (s, CS9K_OPTION_XMIN, 0,
			    s->frame_bounds[s->i_frame - 1].width - 2, 1);
	  flags |=
	    cs9k_set_range (s, CS9K_OPTION_XMAX, 1,
			    s->frame_bounds[s->i_frame - 1].width - 1, 1);
	  flags |=
	    cs9k_set_range (s, CS9K_OPTION_YMIN, 0,
			    s->frame_bounds[s->i_frame - 1].height - 2, 1);
	  flags |=
	    cs9k_set_range (s, CS9K_OPTION_YMAX, 1,
			    s->frame_bounds[s->i_frame - 1].height - 1, 1);
	  flags |= SANE_INFO_RELOAD_OPTIONS;
	}
      break;

    case CS9K_OPTION_SUBFRAME_X:
      s->subframe_x = SANE_UNFIX (val);
      break;
    case CS9K_OPTION_SUBFRAME_Y:
      s->subframe_y = SANE_UNFIX (val);
      break;

    case CS9K_OPTION_RES:
      if (s->res != val)
	{
	  s->res = val;
	  flags |= SANE_INFO_RELOAD_PARAMS;
	}
      break;
    case CS9K_OPTION_RESX:
      if (s->resx != val)
	{
	  s->resx = val;
	  flags |= SANE_INFO_RELOAD_PARAMS;
	}
      break;
    case CS9K_OPTION_RESY:
      if (s->resy != val)
	{
	  s->resy = val;
	  flags |= SANE_INFO_RELOAD_PARAMS;
	}
      break;

    case CS9K_OPTION_RES_INDEPENDENT:
      if (s->res_independent != val)
	{
	  if (s->caps & CS9K_CAPS_INDEP_XY)
	    s->res_independent = val;
	  else if (val)
	    return SANE_STATUS_INVAL;

	  /* For future SANE versions
	     flags |= cs9k_set_visible(s, CS9K_OPTION_RES, !s->res_independent);
	     flags |= cs9k_set_visible(s, CS9K_OPTION_RESX, s->res_independent);
	   */
	  flags |= cs9k_set_visible (s, CS9K_OPTION_RESY, s->res_independent);
	}
      break;

    case CS9K_OPTION_PREVIEW_RESOLUTION:
      s->res_preview = val;
      break;

    case CS9K_OPTION_THUMB_RESOLUTION:
      s->res_thumb = val;
      break;

    case CS9K_OPTION_XMIN:
      s->xmin = val;
      break;
    case CS9K_OPTION_XMAX:
      s->xmax = val;
      break;
    case CS9K_OPTION_YMIN:
      s->ymin = val;
      break;
    case CS9K_OPTION_YMAX:
      s->ymax = val;
      break;

    case CS9K_OPTION_SET_FRAME_USER:
      if (s->sf_user != val)
	s->sf_user = val;
      flags |=
	cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_ORIGIN_X, s->sf_user);
      flags |=
	cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_ORIGIN_Y, s->sf_user);
      flags |=
	cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_STRIDE_X, s->sf_user);
      flags |=
	cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_STRIDE_Y, s->sf_user);
      flags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_WIDTH, s->sf_user);
      flags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_HEIGHT, s->sf_user);
      flags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_PRESET, s->sf_user);
      flags |= cs9k_set_visible (s, CS9K_OPTION_SUBFRAME_X, !s->sf_user);
      flags |= cs9k_set_visible (s, CS9K_OPTION_SUBFRAME_Y, !s->sf_user);

      if (s->sf_user)
	cs9k_set_frame_defs (s, &flags);
      else
	cs9k_read_frame_defs (s, &flags);
      break;
    case CS9K_OPTION_SET_FRAME_ORIGIN_X:
      if (s->sf_origin_x != val)
	{
	  s->sf_origin_x = val;
	  cs9k_set_frame_defs (s, &flags);
	  s->thumbnail_done = SANE_FALSE;
	  s->thumb_pos_in = 0;
	}
      break;
    case CS9K_OPTION_SET_FRAME_ORIGIN_Y:
      if (s->sf_origin_y != val)
	{
	  s->sf_origin_y = val;
	  cs9k_set_frame_defs (s, &flags);
	  s->thumbnail_done = SANE_FALSE;
	  s->thumb_pos_in = 0;
	}
      break;
    case CS9K_OPTION_SET_FRAME_STRIDE_X:
      if (s->sf_stride_x != val)
	{
	  s->sf_stride_x = val;
	  cs9k_set_frame_defs (s, &flags);
	}
      break;
    case CS9K_OPTION_SET_FRAME_STRIDE_Y:
      if (s->sf_stride_y != val)
	{
	  s->sf_stride_y = val;
	  cs9k_set_frame_defs (s, &flags);
	}
      break;
    case CS9K_OPTION_SET_FRAME_WIDTH:
      if (s->sf_width != val)
	{
	  s->sf_width = val;
	  cs9k_set_frame_defs (s, &flags);
	}
      break;
    case CS9K_OPTION_SET_FRAME_HEIGHT:
      if (s->sf_height != val)
	{
	  s->sf_height = val;
	  cs9k_set_frame_defs (s, &flags);
	}
      break;

    case CS9K_OPTION_SET_FRAME_PRESET:
      if (val)
	{
	  s->sf_origin_x = cs9k_frame_presets[val].sf_origin_x;
	  s->sf_origin_y = cs9k_frame_presets[val].sf_origin_y;
	  s->sf_stride_x = cs9k_frame_presets[val].sf_stride_x;
	  s->sf_stride_y = cs9k_frame_presets[val].sf_stride_y;
	  s->sf_width = cs9k_frame_presets[val].sf_width;
	  s->sf_height = cs9k_frame_presets[val].sf_height;
	  cs9k_set_frame_defs (s, &flags);
	  flags |= SANE_INFO_RELOAD_OPTIONS;
	}
      break;

    case CS9K_OPTION_FOCUS_ON_CENTRE:
      if (s->focus_on_centre != val)
	{
	  s->focus_on_centre = val;
	  flags |=
	    cs9k_set_visible (s, CS9K_OPTION_FOCUSX, !s->focus_on_centre);
	  flags |=
	    cs9k_set_visible (s, CS9K_OPTION_FOCUSY, !s->focus_on_centre);
	}
      break;
    case CS9K_OPTION_FOCUS:
      if (s->focusz != val)
	s->focusz = val;
      break;
    case CS9K_OPTION_AUTOFOCUS:
      if (s->autofocus != val)
	{
	  if (s->exe_ax & CS9K_EXE_A0_AF)
	    s->autofocus = val;
	  else
	    {
	      DBG (1, "Autofocus unsupported.\n");
	      return SANE_STATUS_INVAL;
	    }
	}
      break;
    case CS9K_OPTION_FOCUSX:
      s->focusx = val;
      break;
    case CS9K_OPTION_FOCUSY:
      s->focusy = val;
      break;

    case CS9K_OPTION_SCAN_AE:
      if (s->ae != val)
	{
	  if (s->scankinds & CS9K_SCANKIND_AESCAN)
	    s->ae = val;
	  else
	    {
	      DBG (1, "Autoexposure unsupported.\n");
	      return SANE_STATUS_INVAL;
	    }
	}
      break;
    case CS9K_OPTION_SCAN_AE_WB:
      if (s->aewb != val)
	{
	  if (s->scankinds & CS9K_SCANKIND_AEWBSCAN)
	    s->aewb = val;
	  else
	    {
	      DBG (1, "Autoexposure with white balance unsupported.\n");
	      return SANE_STATUS_INVAL;
	    }
	}
      break;

    default:
      DBG (4, "%s: ERROR: Unknown option number %d (bug?).\n", __func__, n);
      return SANE_STATUS_INVAL;
      break;
    }

  s->option_flags = flags;
  return 0;
}

static SANE_Status
cs9k_page_inquiry (cs9k_t * s, int page)
{
  SANE_Status status;

  size_t n;

  if (page >= 0)
    {

      cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
      cs9k_init_buffer (s);
      cs9k_parse_cmd (s, "12 01");
      cs9k_pack_byte (s, page);
      cs9k_parse_cmd (s, "00 04 00");
      s->n_recv = 4;
      status = cs9k_issue_cmd (s);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (4,
	       "Error: cs9k_page_inquiry(): Inquiry of page size failed: %s.\n",
	       sane_strstatus (status));
	  return status;
	}
      n = s->recv_buf[3] + 4;

    }
  else
    n = 36;

  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);
  if (page >= 0)
    {
      cs9k_parse_cmd (s, "12 01");
      cs9k_pack_byte (s, page);
      cs9k_parse_cmd (s, "00");
    }
  else
    cs9k_parse_cmd (s, "12 00 00 00");
  cs9k_pack_byte (s, n);
  cs9k_parse_cmd (s, "00");
  s->n_recv = n;

  status = cs9k_issue_cmd (s);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (4, "Error: %s: inquiry of page failed: %s.\n",
	   __func__, sane_strstatus (status));
      return status;
    }
  if (s->n_recv == 0)
    {
      DBG (4, "Error: %s: inquiry of page 0x%02x returned no data.\n",
	   __func__, page);
      return SANE_STATUS_INVAL;
    }
  if ((page >= 0) && (s->recv_buf[1] != page))
    {
      DBG (4,
	   "Error: %s: inquiry of page 0x%02x returned page 0x%02x.\n",
	   __func__, page, s->recv_buf[1]);
      cs9k_reset (s);
      return SANE_STATUS_INVAL;
    }

  return SANE_STATUS_GOOD;
}

/* Set up user-defined frame boundaries */
static SANE_Word
cs9k_set_frame_defs (cs9k_t * s, SANE_Word * flags)
{
  SANE_Int x, y, max_frames, nframes, tflags = 0;

  DBG (4, "Entering %s\n", __func__);

  if (s->sf_stride_x <= 0)
    nframes = 0;
  else
    {
      max_frames = ((s->swax_max - s->sf_origin_x) / s->sf_stride_x + 1) *
	((s->sway_max - s->sf_origin_y) / s->sf_stride_y + 1);
      max_frames = (max_frames + 15) & ~0xf;
      if (max_frames > s->frame_bounds_size)
	{
	  s->frame_bounds =
	    (cs9k_frame_bounds_t *) cs9k_xrealloc (s->frame_bounds,
						   max_frames *
						   sizeof
						   (cs9k_frame_bounds_t));
	  if (!s->frame_bounds)
	    {
	      DBG (4,
		   "Couldn't allocate space for %d frames xy locations.\n",
		   s->n_frames);
	      return SANE_STATUS_NO_MEM;
	    }
	  s->frame_bounds_size = max_frames;
	  DBG (4, "%s: Allocated space for %d frames\n", __func__,
	       s->frame_bounds_size);
	}

      nframes = 0;
      for (x = s->sf_origin_x; (x + s->sf_width * 3 / 4) < s->swax_max;
	   x += s->sf_stride_x)
	for (y = s->sf_origin_y; (y + s->sf_height * 3 / 4) < s->sway_max;
	     y += s->sf_stride_y)
	  {
	    s->frame_bounds[nframes].offset_x = x;
	    s->frame_bounds[nframes].offset_y = y;
	    s->frame_bounds[nframes].width = s->sf_width;
	    s->frame_bounds[nframes].height = s->sf_height;
	    s->frame_bounds[nframes].adj_exp = SANE_FIX (0.0);
	    s->frame_bounds[nframes].adj_exp_type = 0;
	    s->frame_bounds[nframes].flags = 0;
	    nframes++;
	  }
      DBG (4, "%s: Created %d frames\n", __func__, nframes);
    }

  if (nframes != s->n_frames)
    {
      DBG (4, "%s: Number of frames changed %d -> %d\n", __func__,
	   s->n_frames, nframes);

      if (nframes < 1)
	nframes = 1;

      s->n_frames = nframes;
      tflags |= cs9k_set_visible (s, CS9K_OPTION_FRAME, s->n_frames > 1
				  && (!s->thumbnail || s->source));
      tflags |= cs9k_set_range (s, CS9K_OPTION_FRAME, 1, s->n_frames, 1);
      cs9k_adjustWord2optrange (s, &s->n_frames, CS9K_OPTION_FRAME);
      cs9k_adjustWord2optrange (s, &s->i_frame, CS9K_OPTION_FRAME);
      tflags |= cs9k_set_visible (s, CS9K_OPTION_EXPOSURE_PER_FRAME,
		      s->n_frames > 1 && s->thumbnail && s->source);
      s->c_frame = s->i_frame;

    }
  if (flags)
    *flags = tflags;
  return 0;
}

/* Get frame descriptors from scanner */
static SANE_Status
cs9k_read_frame_defs (cs9k_t * s, SANE_Word * flags)
{
  SANE_Status status;
  SANE_Int i, max_frames, n_bounds, offset, descriptors, tflags = 0;

  DBG (4, "Entering %s\n", __func__);
  status = cs9k_page_inquiry (s, 0xc1);
  if (status != SANE_STATUS_GOOD)
    return status;

  cs9k_get_inquiry_value (s, 75, 1, &s->n_frames);

  if (s->n_frames < 1)
    s->n_frames = 1;

  max_frames = (s->n_frames + 15) & ~0xf;
  if (max_frames > s->frame_bounds_size)
    {
      s->frame_bounds =
	(cs9k_frame_bounds_t *) cs9k_xrealloc (s->frame_bounds,
					       max_frames *
					       sizeof (cs9k_frame_bounds_t));
      if (!s->frame_bounds)
	{
	  DBG (4,
	       "Couldn't allocate space for %d frames xy locations.\n",
	       s->n_frames);
	  return SANE_STATUS_NO_MEM;
	}
      s->frame_bounds_size = max_frames;
    }

  status = cs9k_page_inquiry (s, 0xc8);
  if (status)
    {
      DBG (4,
	   "%s: Failed to get page 0xc8, using defaults: %s\n",
	   __func__, sane_strstatus (status));

      offset = s->resy_max * 1.5;	/* Assume 35mm frames 1.5" pitch */
      for (i = 0; (SANE_Word) i < s->n_frames; i++)
	{
	  s->frame_bounds[i].offset_x = 0;
	  s->frame_bounds[i].offset_y = i * offset;
	  s->frame_bounds[i].width = s->frame_width;
	  s->frame_bounds[i].height = s->frame_height;
	  s->frame_bounds[i].adj_exp = SANE_FIX (0.0);
	  s->frame_bounds[i].adj_exp_type = 0;
	  s->frame_bounds[i].flags = 0;
	}
    }
  else
    {
      if ((cs9k_get_inquiry_value (s, 4, 1, &n_bounds) &
	   CS9K_INQ_BAD_OFFSET) == 0)
	if (n_bounds > s->n_frames)
	  {
	    DBG (1,
		 "There are %d frame descriptors but only %d frames.\n",
		 n_bounds, s->n_frames);
	    n_bounds = s->n_frames;
	  }
      descriptors = (s->recv_buf[3] - 1) / 16;
      DBG (10, "0xC8 has %d frame descriptors.\n", descriptors);
      for (i = 0; (SANE_Word) i < descriptors; i++)
	{
	  cs9k_get_inquiry_value (s, i * 16 + 5, 4,
				  &s->frame_bounds[i].offset_x);
	  cs9k_get_inquiry_value (s, i * 16 + 9, 4,
				  &s->frame_bounds[i].offset_y);
	  cs9k_get_inquiry_value (s, i * 16 + 13, 4,
				  &s->frame_bounds[i].width);
	  cs9k_get_inquiry_value (s, i * 16 + 17, 4,
				  &s->frame_bounds[i].height);
	  if (i == 0 && (s->frame_bounds[i].height == 0))
	    s->frame_bounds[i].height = s->frame_height;
	  s->frame_bounds[i].adj_exp = SANE_FIX (0.0);
	  s->frame_bounds[i].adj_exp_type = 0;
	  s->frame_bounds[i].flags = 0;
	}
      if (n_bounds > descriptors)
	{
	  status = cs9k_page_inquiry (s, 0xc9);
	  if (status)
	    return status;
	  n_bounds = s->recv_buf[3] / 16;
	  for (i = 0; (SANE_Word) i < n_bounds; i++)
	    {
	      cs9k_get_inquiry_value (s, i * 16 + 4, 4,
				      &s->frame_bounds[descriptors +
						       i].offset_x);
	      cs9k_get_inquiry_value (s, i * 16 + 8, 4,
				      &s->frame_bounds[descriptors +
						       i].offset_y);
	      cs9k_get_inquiry_value (s, i * 16 + 12, 4,
				      &s->frame_bounds[descriptors +
						       i].width);
	      cs9k_get_inquiry_value (s, i * 16 + 16, 4,
				      &s->frame_bounds[descriptors +
						       i].height);
	      s->frame_bounds[descriptors + i].adj_exp = SANE_FIX (0.0);
	      s->frame_bounds[descriptors + i].adj_exp_type = 0;
	      s->frame_bounds[descriptors + i].flags = 0;
	    }
	}
    }
  for (i = 0; (SANE_Word) i < s->n_frames; i++)
    {
      DBG (10, "Frame %2d/%2d at %5d, %d (%dx%d)\n", i + 1, s->n_frames,
	   s->frame_bounds[i].offset_x, s->frame_bounds[i].offset_y,
	   s->frame_bounds[i].width, s->frame_bounds[i].height);
    }
  tflags |= cs9k_set_visible (s, CS9K_OPTION_FRAME, s->n_frames > 1
			      && (!s->thumbnail || s->source));
  tflags |= cs9k_set_range (s, CS9K_OPTION_FRAME, 1, s->n_frames, 1);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_EXPOSURE_PER_FRAME,
			      s->n_frames > 1 && s->thumbnail && s->source);

  cs9k_adjustWord2optrange (s, &s->n_frames, CS9K_OPTION_FRAME);
  cs9k_adjustWord2optrange (s, &s->i_frame, CS9K_OPTION_FRAME);
  s->c_frame = s->i_frame;

  if (flags)
    *flags = tflags;

  DBG (10, "%s: Set up to scan frame %d of %d.\n", __func__, s->i_frame,
       s->n_frames);

  return SANE_STATUS_GOOD;
}

/* Extract device settings and capabilities */
static SANE_Status
cs9k_full_inquiry (cs9k_t * s, SANE_Int * flags)
{
  SANE_Status status;
  int pitch, pitch_max;
  size_t i;
  SANE_Int tflags = 0;
  SANE_Int dic_ctl, dic_len, alg_ctl, alg_len, offs, lut_scale, val_sz;
  SANE_Int holder_old, depth_mask, newval, junk;
  SANE_Int resx_optical, resx_min;
  SANE_Int resy_optical, resy_min;
  SANE_Int res_thumb_min, res_thumb_max;
  cs9k_pixel_t pixel;
  cs9k_inq_val_t parms_static = 0, parms_dynamic = 0;
  double exposure_c_range_max, exposure_range_max;

  DBG (4, "Entering %s\n", __func__);

  /*== Page 00 - page list; for debug only =================================*/
  status = cs9k_page_inquiry (s, 0x00);
  if (status != SANE_STATUS_GOOD)
    return status;
  for (i = 3; (unsigned) i < s->n_recv; i++)
    cs9k_get_inquiry_value (s, i, 1, &junk);

  /*== Page C1 =============================================================*/
  status = cs9k_page_inquiry (s, 0xc1);
  if (status != SANE_STATUS_GOOD)
    return status;

  cs9k_get_inquiry_value (s, 3, 1, &junk);	/* Page length */
  cs9k_get_inquiry_value (s, 4, 1, &junk);	/* SCSI data transfer function */
  cs9k_get_inquiry_value (s, 5, 2, &junk);	/* Window Descriptor block length */
  cs9k_get_inquiry_value (s, 7, 2, &junk);	/* Set parameter descriptor block length */
  cs9k_get_inquiry_value (s, 9, 2, &junk);	/* General SCSI buffer size */
  cs9k_get_inquiry_value (s, 11, 2, &junk);	/* Image buffer size (kB) */
  cs9k_get_inquiry_value (s, 13, 1, &junk);	/* Number of equipped unit */
  cs9k_get_inquiry_value (s, 14, 1, &junk);	/* Unit name ID */

  holder_old = s->holder_id;
  parms_dynamic |= cs9k_get_inquiry_value (s, 15, 1, &s->holder_id);
  parms_static |= cs9k_get_inquiry_value (s, 16, 1, &s->coord_base_info);
  cs9k_get_inquiry_value (s, 17, 1, &junk);
  parms_static |= cs9k_get_inquiry_value (s, 18, 2, &resx_optical);
  parms_static |= cs9k_get_inquiry_value (s, 20, 2, &s->resx_max);
  parms_static |= cs9k_get_inquiry_value (s, 22, 2, &resx_min);
  parms_dynamic |= cs9k_get_inquiry_value (s, 24, 4, &s->swax_max);
  parms_dynamic |= cs9k_get_inquiry_value (s, 28, 4, &s->swax_min);
  parms_dynamic |= cs9k_get_inquiry_value (s, 32, 4, &s->framex_pos);
  parms_dynamic |= cs9k_get_inquiry_value (s, 36, 4, &s->frame_width);

  parms_static |= cs9k_get_inquiry_value (s, 40, 2, &resy_optical);
  parms_static |= cs9k_get_inquiry_value (s, 42, 2, &s->resy_max);
  parms_static |= cs9k_get_inquiry_value (s, 44, 2, &resy_min);
  parms_dynamic |= cs9k_get_inquiry_value (s, 46, 4, &s->sway_max);
  parms_dynamic |= cs9k_get_inquiry_value (s, 50, 4, &s->sway_min);
  parms_dynamic |= cs9k_get_inquiry_value (s, 54, 4, &s->framey_pos);
  parms_dynamic |= cs9k_get_inquiry_value (s, 58, 4, &s->frame_height);
  cs9k_get_inquiry_value (s, 62, 4, &junk);
  cs9k_get_inquiry_value (s, 66, 4, &junk);

  parms_static |= cs9k_get_inquiry_value (s, 70, 2, &res_thumb_max);
  parms_static |= cs9k_get_inquiry_value (s, 72, 2, &res_thumb_min);

  cs9k_get_inquiry_value (s, 74, 1, &junk);
  cs9k_get_inquiry_value (s, 75, 1, &junk);

  parms_dynamic |= cs9k_get_inquiry_value (s, 76, 2, &s->focus_min);
  parms_dynamic |= cs9k_get_inquiry_value (s, 78, 2, &s->focus_max);

  cs9k_get_inquiry_value (s, 80, 2, &junk);

  parms_static |= cs9k_get_inquiry_value (s, 82, 1, &s->maxbits);
  if (s->type == CS9K_TYPE_LS30)	/* must be overridden, LS-30 claims to have 12 bits */
    s->maxbits = 10;

  if ((parms_static & CS9K_INQ_BAD_OFFSET)
      || (parms_dynamic & CS9K_INQ_BAD_OFFSET))
    {
      DBG (10,
	   "Page 0xc1 must be at least 83 bytes to get required parameters.\n");
      return SANE_STATUS_IO_ERROR;
    }

  parms_static |= cs9k_get_inquiry_value (s, 83, 2, &s->ccd_width);

  parms_static |= cs9k_get_inquiry_value (s, 85, 1, &s->linegapcount);

  cs9k_get_inquiry_value (s, 86, 1, &junk);
  cs9k_get_inquiry_value (s, 87, 2, &junk);
  cs9k_get_inquiry_value (s, 89, 2, &junk);

  s->unit_dpi = s->resx_max;

  /* Arbitrary - 1/4" frames min. */
  tflags |=
    cs9k_set_range (s, CS9K_OPTION_SET_FRAME_ORIGIN_X, s->swax_min,
		    s->ccd_width - s->unit_dpi / 4, 1);
  tflags |=
    cs9k_set_range (s, CS9K_OPTION_SET_FRAME_STRIDE_X, s->unit_dpi / 4,
		    s->swax_max + 1, 1);
  tflags |= cs9k_set_range (s, CS9K_OPTION_XMIN, 0, s->frame_width - 2, 1);
  tflags |= cs9k_set_range (s, CS9K_OPTION_XMAX, 1, s->frame_width - 1, 1);
  tflags |=
    cs9k_set_range (s, CS9K_OPTION_SET_FRAME_WIDTH, 0, s->frame_width, 1);
  tflags |= cs9k_set_range (s, CS9K_OPTION_FOCUSX, 0, s->frame_width - 1, 1);
  tflags |=
    cs9k_set_range (s, CS9K_OPTION_SET_FRAME_ORIGIN_Y, s->sway_min,
		    s->sway_max - s->unit_dpi / 4, 1);
  tflags |=
    cs9k_set_range (s, CS9K_OPTION_SET_FRAME_STRIDE_Y, s->unit_dpi / 4,
		    s->sway_max + 1, 1);
  tflags |= cs9k_set_range (s, CS9K_OPTION_YMIN, 0, s->frame_height - 2, 1);
  tflags |= cs9k_set_range (s, CS9K_OPTION_YMAX, 1, s->frame_height - 1, 1);
  tflags |=
    cs9k_set_range (s, CS9K_OPTION_SET_FRAME_HEIGHT, 0, s->frame_height, 1);
  tflags |= cs9k_set_range (s, CS9K_OPTION_FOCUSY, 0, s->frame_height - 1, 1);
  tflags |=
    cs9k_set_range (s, CS9K_OPTION_FOCUS, s->focus_min, s->focus_max, 1);

  if (parms_static & CS9K_INQ_BAD_OFFSET)
    DBG (10, "Error initializing static parameters from page 0xc1.\n");

  if (parms_dynamic & CS9K_INQ_BAD_OFFSET)
    DBG (10, "Error initializing dynamic parameters from page 0xc1.\n");

  if (s->init_done && (parms_static & CS9K_INQ_NEW_VALUE))
    DBG (1, "Some assumed static parameters changed on page 0xc1.\n");

  if (flags && (parms_dynamic & CS9K_INQ_NEW_VALUE))
    *flags |= SANE_INFO_RELOAD_OPTIONS;

  /*== Page HH =============================================================*/
  /*== Holder page is available only when holder is current, so must	  ==*/
  /*==	 fetch on demand.						  ==*/
  if ((s->caps & CS9K_CAPS_HOLDER) && (s->holder_id != holder_old))
    {
      status = cs9k_page_inquiry (s, s->holder_id);
      if (status)
	{
	  DBG (4, "Error: %s: Failed to get page: %s\n", __func__,
	       sane_strstatus (status));
	  return status;
	}
      strncpy (s->holder_string, (char *) (s->recv_buf + 5),
	       CS9K_HOLDER_NAME_MAX - 1);
      s->holder_string[CS9K_HOLDER_NAME_MAX - 1] = 0;
      DBG (10, "Holder is now '%s'\n", s->holder_string);

    }

  /*== Page D1 =============================================================*/

  parms_static = 0;
  parms_dynamic = 0;

  status = cs9k_page_inquiry (s, 0xd1);
  if (status)
    {
      DBG (4,
	   "%s: Failed to get page 0xd1, using defaults: %s\n",
	   __func__, sane_strstatus (status));
      depth_mask = 0;
      for (i = 0; i < 8; i++)
	if (cs9k_bit_depths[i] == 8 || cs9k_bit_depths[i] == s->maxbits)
	  depth_mask |= 1 << i;
    }
  else
    {
      parms_static |= cs9k_get_inquiry_value (s, 4, 1, &s->scankinds);
      parms_static |= cs9k_get_inquiry_value (s, 5, 1, &s->scanmodes);

      cs9k_get_inquiry_value (s, 6, 1, &junk);

      parms_static |= cs9k_get_inquiry_value (s, 7, 1, &s->scancolors);
      parms_static |= cs9k_get_inquiry_value (s, 8, 2, &junk);
      parms_static |= cs9k_get_inquiry_value (s, 9, 2, &junk);

      parms_static |= cs9k_get_inquiry_value (s, 10, 1, &depth_mask);

      cs9k_get_inquiry_value (s, 11, 1, &junk);

      parms_static |= cs9k_get_inquiry_value (s, 12, 1, &dic_ctl);	/* Digital Image Control Support */
      parms_static |= cs9k_get_inquiry_value (s, 13, 1, &dic_len);	/* Additional length for Digital Control Information */
      parms_static |= cs9k_get_inquiry_value (s, 14, 1, &alg_ctl);
      parms_static |= cs9k_get_inquiry_value (s, 15, 1, &alg_len);

      offs = 16;
      if (dic_len)
	for (i = 0; i < 8; i++)
	  if (dic_ctl & (1 << i))
	    DBG (1, "%s: Unknown digital control 0x%02x\n", __func__, 1 << i);

      offs = 16 + dic_len;
      parms_static |= cs9k_get_inquiry_value (s, offs, 1, &val_sz);
      for (i = 0; i < 8; i++)
	if (alg_ctl & (1 << i))
	  switch (i)
	    {
	    case 0:
	    case 1:
	    case 2:
	    case 3:
	    case 4:
	    case 5:
	      DBG (1,
		   "%s: Unknown analog control 0x%02x; hoping for zero length descriptor\n",
		   __func__, 1 << i);
	      break;
	    case 6:
	      parms_static |=
		cs9k_get_inquiry_value (s, offs + 1, val_sz,
					&s->exposure_min);
	      parms_static |=
		cs9k_get_inquiry_value (s, offs + 1 + val_sz, val_sz,
					&s->exposure_max);
	      offs += 2 * val_sz + 1;
	      break;
	    case 7:
	      DBG (1, "%s: Unknown analog control 0x%02x\n", __func__,
		   1 << i);
	      break;
	    }
      offs = 16 + dic_len + alg_len;
      cs9k_get_inquiry_value (s, offs + 0, 1, &junk);
      cs9k_get_inquiry_value (s, offs + 1, 1, &junk);
      cs9k_get_inquiry_value (s, offs + 2, 1, &junk);

      if (parms_static & CS9K_INQ_BAD_OFFSET)
	DBG (10, "Error initializing static parameters from page 0xd1.\n");

      if (parms_dynamic & CS9K_INQ_BAD_OFFSET)
	DBG (10, "Error initializing dynamic parameters from page 0xd1.\n");

      if (s->init_done && (parms_static & CS9K_INQ_NEW_VALUE))
	DBG (1, "Some assumed static parameters changed on page 0xd1.\n");

      if (flags && (parms_dynamic & CS9K_INQ_NEW_VALUE))
	*flags |= SANE_INFO_RELOAD_OPTIONS;
    }

  /*== Page E1 ===========================================================*/
  parms_static = 0;
  parms_dynamic = 0;

  status = cs9k_page_inquiry (s, 0xe1);
  if (status)
    {
      DBG (4,
	   "%s: Failed to get page 0xe1, using defaults: %s\n",
	   __func__, sane_strstatus (status));
      s->n_lut = s->n_lut_out = 1 << s->maxbits;
    }
  else
    {
      cs9k_get_inquiry_value (s, 4, 2, &s->host_coop);	/* Host cooperation function */
      cs9k_get_inquiry_value (s, 6, 3, (SANE_Word *)&s->sendread_cap1);	/* Send/Read supported information P. 1 */
      cs9k_get_inquiry_value (s, 9, 2, (SANE_Word *)&s->sendread_cap2);	/* Send/Read supported information P. 2 */
      cs9k_get_inquiry_value (s, 11, 1, &junk);
      parms_static |= cs9k_get_inquiry_value (s, 12, 1, &s->n_lut);	/* Number of bits on LUT input */
      parms_static |= cs9k_get_inquiry_value (s, 13, 1, &s->n_lut_out);
      if (s->n_lut)
	s->n_lut = 1 << s->n_lut;
      if (s->n_lut_out)
	s->n_lut_out = 1 << s->n_lut_out;
      cs9k_get_inquiry_value (s, 14, 1, &junk);
      cs9k_get_inquiry_value (s, 15, 1, &junk);
      cs9k_get_inquiry_value (s, 16, 1, &junk);
      cs9k_get_inquiry_value (s, 17, 1, &junk);
      cs9k_get_inquiry_value (s, 18, 1, &junk);
      cs9k_get_inquiry_value (s, 19, 1, &junk);
      parms_static |= cs9k_get_inquiry_value (s, 20, 2, (SANE_Word *)&s->exe_8x);	/* Command capabilities */
      parms_static |= cs9k_get_inquiry_value (s, 22, 2, (SANE_Word *)&s->exe_9x);
      parms_static |= cs9k_get_inquiry_value (s, 24, 2, (SANE_Word *)&s->exe_ax);
      parms_static |= cs9k_get_inquiry_value (s, 26, 2, (SANE_Word *)&s->exe_bx);
      parms_static |= cs9k_get_inquiry_value (s, 28, 2, (SANE_Word *)&s->exe_cx);
      parms_static |= cs9k_get_inquiry_value (s, 30, 2, (SANE_Word *)&s->exe_dx);
      cs9k_get_inquiry_value (s, 32, 2, &junk);
      cs9k_get_inquiry_value (s, 34, 2, &junk);
      cs9k_get_inquiry_value (s, 36, 1, &junk);
      cs9k_get_inquiry_value (s, 37, 1, &junk);
      cs9k_get_inquiry_value (s, 38, 1, &junk);
    }
  if (parms_static & CS9K_INQ_BAD_OFFSET)
    DBG (10, "Error initializing static parameters from page 0xe1.\n");

  if (parms_dynamic & CS9K_INQ_BAD_OFFSET)
    DBG (10, "Error initializing dynamic parameters from page 0xe1.\n");

  if (s->init_done && (parms_static & CS9K_INQ_NEW_VALUE))
    DBG (1, "Some assumed static parameters changed on page 0xf1.\n");

  if (flags && (parms_dynamic & CS9K_INQ_NEW_VALUE))
    *flags |= SANE_INFO_RELOAD_OPTIONS;

  /*== Page C8 - Additional Address page ===================================*/

  if (!s->sf_user)
    {
      status = cs9k_read_frame_defs (s, flags);
      if (status)
	return status;
    }

  /*== Page E3 - CCD measurement setting page ==============================*/

  status = cs9k_page_inquiry (s, 0xe3);
  if (status)
    {
      DBG (4,
	   "%s: Failed to get page 0xe3, using defaults: %s\n",
	   __func__, sane_strstatus (status));
    }
  else
    {
      cs9k_get_inquiry_value (s, 3, 1, &junk);
      cs9k_get_inquiry_value (s, 4, 2, &junk);
      cs9k_get_inquiry_value (s, 6, 2, &junk);
      cs9k_get_inquiry_value (s, 8, 1, &junk);
      cs9k_get_inquiry_value (s, 10, 1, &val_sz);
      if ((cs9k_get_inquiry_value (s, 4, 1, &val_sz) &
	   CS9K_INQ_BAD_OFFSET) == 0)
	for (i = 0; (SANE_Word) i < val_sz; i += 2)
	  cs9k_get_inquiry_value (s, i + 11, 2, &junk);
    }

  /*== Get Window (all)  ===================================================*/
  cs9k_get_all_windows (s);	/* Exposure + */

  /*========================================================================*/
  /* Derived static initialization */

  if (!s->init_done)
    {
      /* Bit depth list generation */
      s->depth_list[0] = 0;
      for (i = 0; i < 8; i++)
	if (cs9k_bit_depths[i])
	  {
	    if (depth_mask & (1 << i))
	      {
		s->depth_list[0]++;
		s->depth_list[s->depth_list[0]] = cs9k_bit_depths[i];
		DBG (4,
		     "%s: scanner has %d bit depth support\n", __func__,
		     s->depth_list[s->depth_list[0]]);
		s->depth = cs9k_bit_depths[i];
	      }
	  }
	else if (depth_mask & (1 << i))
	  DBG (4, "%s: scanner has unkown bit depth support\n", __func__);

      /* Color map Look Up Table generation */
      if (s->n_lut && s->n_lut_out &&
	  (s->sendread_cap1 & CS9K_SENDREAD_03_WRITE_GAMMA_FUNC))
	{
	  s->lut_r = (cs9k_pixel_t *) cs9k_xrealloc (s->lut_r,
						     s->n_lut *
						     sizeof (cs9k_pixel_t));
	  s->lut_g =
	    (cs9k_pixel_t *) cs9k_xrealloc (s->lut_g,
					    s->n_lut * sizeof (cs9k_pixel_t));
	  s->lut_b =
	    (cs9k_pixel_t *) cs9k_xrealloc (s->lut_b,
					    s->n_lut * sizeof (cs9k_pixel_t));
	  s->lut_neutral =
	    (cs9k_pixel_t *) cs9k_xrealloc (s->lut_neutral,
					    s->n_lut * sizeof (cs9k_pixel_t));

	  if (!s->lut_r || !s->lut_g || !s->lut_b || !s->lut_neutral)
	    {
	      cs9k_xfree (s->lut_r);
	      cs9k_xfree (s->lut_g);
	      cs9k_xfree (s->lut_b);
	      cs9k_xfree (s->lut_neutral);
	      s->lut_r = s->lut_g = s->lut_b = s->lut_neutral = NULL;
	      return SANE_STATUS_NO_MEM;
	    }
	  s->option_list[CS9K_OPTION_LUT_R].size = s->n_lut * WSIZE;
	  s->option_list[CS9K_OPTION_LUT_G].size = s->n_lut * WSIZE;
	  s->option_list[CS9K_OPTION_LUT_B].size = s->n_lut * WSIZE;

	  lut_scale = (s->n_lut == s->n_lut_out) ? 0 :
		      (s->n_lut / s->n_lut_out) > 0 ?
			  -s->n_lut / s->n_lut_out :
			  s->n_lut_out / s->n_lut;	/* Amount to shift value */
	  for (pixel = 0; pixel < s->n_lut; pixel++)
	    {
	      if (!(pixel % 2048))
		DBG (10, "Pixel %d -> %d\n", pixel,
		     (lut_scale == 0) ? pixel :
		     (lut_scale > 0) ? pixel << lut_scale :
		     pixel >> -lut_scale);
	      s->lut_r[pixel] = s->lut_g[pixel] = s->lut_b[pixel] =
		s->lut_neutral[pixel] = (lut_scale == 0) ? pixel :
		(lut_scale > 0) ? pixel << lut_scale : pixel >> -lut_scale;
	    }
	  tflags |= cs9k_set_range (s, CS9K_OPTION_LUT_R, 0, s->n_lut - 1, 1);
	  tflags |= cs9k_set_range (s, CS9K_OPTION_LUT_G, 0, s->n_lut - 1, 1);
	  tflags |= cs9k_set_range (s, CS9K_OPTION_LUT_B, 0, s->n_lut - 1, 1);
	}

      /* generate resolution list for x */
      s->resx_list = (SANE_Word *) cs9k_xmalloc ((s->linegapcount +
						  1) * sizeof (SANE_Word));
      s->resx_list[0] = 0;
      for (pitch = 1; pitch <= s->linegapcount; pitch++)
	if ((s->linegapcount % pitch) == 0 && s->resx_max / pitch >= resx_min)
	  s->resx_list[++s->resx_list[0]] = s->resx_max / pitch;

      /* generate resolution list for y */
      s->resy_list = (SANE_Word *) cs9k_xmalloc ((s->linegapcount +
						  1) * sizeof (SANE_Word));
      s->resy_list[0] = 0;
      for (pitch = 1; pitch <= s->linegapcount; pitch++)
	if ((s->linegapcount % pitch) == 0 && s->resy_max / pitch >= resy_min)
	  s->resy_list[++s->resy_list[0]] = s->resy_max / pitch;

      if (s->scankinds & CS9K_SCANKIND_THUMBNAIL)
	{
	  pitch_max = res_thumb_max / res_thumb_min;
	  s->res_thumb_list = (SANE_Word *) cs9k_xmalloc ((pitch_max +
							   1) *
							  sizeof (SANE_Word));
	  s->res_thumb_list[0] = pitch_max;
	  for (pitch = 1; pitch <= pitch_max; pitch++)
	    s->res_thumb_list[pitch] = res_thumb_max / pitch;
	  s->res_thumb = res_thumb_max;
	}

      s->unit_mm = 25.4 / (double) s->unit_dpi;

      s->option_list[CS9K_OPTION_RES].constraint.word_list = s->resx_list;
      s->option_list[CS9K_OPTION_RESX].constraint.word_list = s->resx_list;
      s->option_list[CS9K_OPTION_RESY].constraint.word_list = s->resy_list;
      s->option_list[CS9K_OPTION_PREVIEW_RESOLUTION].constraint.word_list =
	s->resx_list;
      s->option_list[CS9K_OPTION_THUMB_RESOLUTION].constraint.word_list =
	s->res_thumb_list;

      s->res_independent = SANE_FALSE;
      s->res = s->resx = resx_optical;
      cs9k_adjustWord2list (&s->res, s->resx_list, "full_inquiry res");
      cs9k_adjustWord2list (&s->resx, s->resx_list, "full_inquiry resx");
      s->resy = resy_optical;
      cs9k_adjustWord2list (&s->resy, s->resy_list, "full_inquiry resy");
      s->res_preview = s->resx_max / 10;
      cs9k_adjustWord2list (&s->res_preview, s->resx_list,
			    "full_inquiry res_preview");

      exposure_range_max =
	log2 ((double) (s->exposure_max - s->exposure_min));
      if (exposure_range_max > 18.0)
	exposure_c_range_max = 3.0;
      else
	exposure_c_range_max = exposure_range_max / 6.0;

      s->exposure_zero = exposure_c_range_max;

      exposure_range_max -= exposure_c_range_max * 2.0;

      tflags |=
	cs9k_set_range (s, CS9K_OPTION_EXPOSURE, SANE_FIX (0.0),
			SANE_FIX (exposure_range_max), SANE_FIX (0.01));
      tflags |=
	cs9k_set_range (s, CS9K_OPTION_EXPOSURE_R,
			SANE_FIX (-exposure_c_range_max),
			SANE_FIX (exposure_c_range_max), SANE_FIX (0.01));
      tflags |=
	cs9k_set_range (s, CS9K_OPTION_EXPOSURE_G,
			SANE_FIX (-exposure_c_range_max),
			SANE_FIX (exposure_c_range_max), SANE_FIX (0.01));
      tflags |=
	cs9k_set_range (s, CS9K_OPTION_EXPOSURE_B,
			SANE_FIX (-exposure_c_range_max),
			SANE_FIX (exposure_c_range_max), SANE_FIX (0.01));
      tflags |=
	cs9k_set_range (s, CS9K_OPTION_EXPOSURE_I,
			SANE_FIX (-exposure_c_range_max),
			SANE_FIX (exposure_range_max + -exposure_c_range_max),
			SANE_FIX (0.01));
      tflags |=
	cs9k_set_range (s, CS9K_OPTION_EXPOSURE_PER_FRAME,
			SANE_FIX (0.0),
			SANE_FIX (3.0),
			SANE_FIX (0.01));

      tflags |=
	cs9k_set_range (s, CS9K_OPTION_ADJUST_EXPOSURE_TARGET, SANE_FIX (0.0),
			SANE_FIX (exposure_range_max), SANE_FIX (0.01));
      tflags |=
	cs9k_set_range (s, CS9K_OPTION_ADJUST_EXPOSURE_TARGET_IR,
			SANE_FIX (-exposure_c_range_max),
			SANE_FIX (exposure_range_max + -exposure_c_range_max),
			SANE_FIX (0.01));

      s->adj_exp_tgt = SANE_FIX (0.25);
      s->adj_exp_tgt_ir = SANE_FIX (1.8);

      if (s->exe_ax & CS9K_EXE_A0_AF)
	s->autofocus = SANE_TRUE;

      cs9k_set_adjust_message (s, CS9K_OPTION_ADJUST_EXPOSURE, SANE_FIX (0.0),
			       0);
      cs9k_set_adjust_message (s, CS9K_OPTION_ADJUST_EXPOSURE_IR,
			       SANE_FIX (0.0), 0);

      DBG (4, " maximum depth:	%d\n", s->maxbits);
      DBG (4, " focus range:	%d:%d\n", s->focus_min, s->focus_max);
      DBG (4, " resolution (x):	%d (%d-%d)\n", resx_optical,
	   resx_min, s->resx_max);
      DBG (4, " resolution (y):	%d (%d-%d)\n", resy_optical,
	   resy_min, s->resy_max);
      DBG (4, " frames:		%d\n", s->n_frames);

      DBG (4,
	   " Exposure range: %f; control %.2f:%.2f; per-color %.2f, %.2f; center %.2f\n",
	   exposure_range_max,
	   SANE_UNFIX (s->ranges[CS9K_OPTION_EXPOSURE].min),
	   SANE_UNFIX (s->ranges[CS9K_OPTION_EXPOSURE].max),
	   SANE_UNFIX (s->ranges[CS9K_OPTION_EXPOSURE_R].min),
	   SANE_UNFIX (s->ranges[CS9K_OPTION_EXPOSURE_R].max),
	   s->exposure_zero);

      s->init_done = SANE_TRUE;
      s->sf_width = s->sf_stride_x = s->frame_width;
      s->sf_height = s->sf_stride_y = s->frame_height;

      cs9k_set_visible (s, CS9K_OPTION_NEGATIVE,
			s->caps & CS9K_CAPS_NEGATIVE);
      cs9k_set_visible (s, CS9K_OPTION_GROUP_INFRARED,
			s->scancolors & CS9K_CLRCMP_DROPOUT);
      cs9k_set_visible (s, CS9K_OPTION_INFRARED,
			s->scancolors & CS9K_CLRCMP_DROPOUT);
      if (s->depth_list[0] == 1)
	{
	  s->option_list[CS9K_OPTION_DEPTH].cap |= SANE_CAP_ADVANCED;
	  s->option_list[CS9K_OPTION_DEPTH].cap &= ~SANE_CAP_SOFT_SELECT;
	}
      cs9k_set_visible (s, CS9K_OPTION_GROUP_LUT, s->lut_r || s->lut_g
			|| s->lut_b);
      cs9k_set_visible (s, CS9K_OPTION_LUT_R, s->lut_r != NULL);
      cs9k_set_visible (s, CS9K_OPTION_LUT_G, s->lut_g != NULL);
      cs9k_set_visible (s, CS9K_OPTION_LUT_B, s->lut_b != NULL);
      cs9k_set_visible (s, CS9K_OPTION_RESET, s->exe_8x & CS9K_EXE_80_RESET);
      cs9k_set_visible (s, CS9K_OPTION_LOAD, s->exe_dx & CS9K_EXE_D1_LOAD);
      cs9k_set_visible (s, CS9K_OPTION_AUTOLOAD,
			s->exe_dx & CS9K_EXE_D1_LOAD);
      cs9k_set_visible (s, CS9K_OPTION_EJECT, s->exe_dx & CS9K_EXE_D0_UNLOAD);
      cs9k_set_visible (s, CS9K_OPTION_RES_INDEPENDENT,
			s->caps & CS9K_CAPS_INDEP_XY);
      cs9k_set_visible (s, CS9K_OPTION_SCAN_AE,
			s->scankinds & CS9K_SCANKIND_AESCAN);
      cs9k_set_visible (s, CS9K_OPTION_SCAN_AE_WB,
			s->scankinds & CS9K_SCANKIND_AEWBSCAN);
      cs9k_set_visible (s, CS9K_OPTION_PREVIEW_RESOLUTION,
			s->resx_list[0] != 1);
      cs9k_set_visible (s, CS9K_OPTION_HOLDER, s->caps & CS9K_CAPS_HOLDER);
      cs9k_set_visible (s, CS9K_OPTION_THUMB_RESOLUTION,
			s->scankinds & CS9K_SCANKIND_THUMBNAIL);
      if (s->res_thumb_list[0] == 1)
	s->option_list[CS9K_OPTION_THUMB_RESOLUTION].cap &=
	  ~SANE_CAP_SOFT_SELECT;
      cs9k_set_visible (s, CS9K_OPTION_SCAN_THUMBNAILS,
			s->scankinds & CS9K_SCANKIND_THUMBNAIL);
      cs9k_set_visible (s, CS9K_OPTION_AUTOFOCUS, s->exe_ax & CS9K_EXE_A0_AF);


      cs9k_set_range (s, CS9K_OPTION_SUBFRAME_X, SANE_FIX (-5.),
		      SANE_FIX (5.), SANE_FIX (0.));
      cs9k_set_range (s, CS9K_OPTION_SUBFRAME_Y, SANE_FIX (-5.),
		      SANE_FIX (5.), SANE_FIX (0.));

      i = -1;
      do
	{
	  i++;
	  cs9k_frame_preset_list[i] = cs9k_frame_presets[i].name;
	}
      while (cs9k_frame_preset_list[i]);

      s->infrared = CS9K_INFRARED_OFF;
      s->subframe_x = s->subframe_y = 0.;
      s->xmin = 0;
      s->xmax = s->frame_width - 1;
      s->ymin = 0;
      s->ymax = s->frame_height - 1;
      s->focus_on_centre = SANE_TRUE;

      status = cs9k_get_exposure (s);
      if (status)
	return status;
    }

  /* Set option visibility */
  tflags |= cs9k_set_visible (s, CS9K_OPTION_INFRARED_FMT, s->infrared);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_EXPOSURE_I, s->infrared);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_EXPOSURE_PER_FRAME, s->n_frames > 1
				     && s->thumbnail && s->source);

  tflags |= cs9k_set_visible (s, CS9K_OPTION_RES, SANE_TRUE);	/* Future: !s->res_independent */
  tflags |= cs9k_set_visible (s, CS9K_OPTION_RESX, SANE_FALSE);	/* Future: s->res_independent */
  tflags |= cs9k_set_visible (s, CS9K_OPTION_RESY, s->res_independent);

  tflags |= cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE,
		      fabs (SANE_UNFIX (s->adj_exp)) > 0.02);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_TEXT_IR,
		      (s->infrared != CS9K_INFRARED_OFF));
  tflags |= cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_IR,
		      (s->infrared != CS9K_INFRARED_OFF)
		      && (fabs (SANE_UNFIX (s->adj_exp_ir)) > 0.02));
  tflags |= cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_TARGET_IR,
		      s->infrared != CS9K_INFRARED_OFF);

  tflags |= cs9k_set_visible (s, CS9K_OPTION_FRAME, s->n_frames > 1
			      && (!s->thumbnail || s->source));

  tflags |= cs9k_set_visible (s, CS9K_OPTION_SUBFRAME_X, !s->sf_user);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_SUBFRAME_Y, !s->sf_user);

  tflags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_ORIGIN_X, s->sf_user);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_ORIGIN_Y, s->sf_user);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_STRIDE_X, s->sf_user);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_STRIDE_Y, s->sf_user);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_WIDTH, s->sf_user);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_HEIGHT, s->sf_user);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_SET_FRAME_PRESET, s->sf_user);

  tflags |= cs9k_set_visible (s, CS9K_OPTION_FOCUS,
		      s->exe_cx & CS9K_EXE_C1_FOCUS_MOVE);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_FOCUS_ON_CENTRE,
		      s->exe_cx & CS9K_EXE_C1_FOCUS_MOVE);
  tflags |= cs9k_set_visible (s, CS9K_OPTION_FOCUSX, !s->focus_on_centre
			      && (s->exe_cx & CS9K_EXE_C1_FOCUS_MOVE));
  tflags |= cs9k_set_visible (s, CS9K_OPTION_FOCUSY, !s->focus_on_centre
			      && (s->exe_cx & CS9K_EXE_C1_FOCUS_MOVE));
  if (flags)
    *flags |= tflags;

  switch (s->holder_id)
    {
    case 0x12:
    case 0x13:
    case 0x14:
    case 0x15:
    case 0x16:
    case 0x19:
    case 0x1a:
    case 0x1b:
    case 0x1c:
    case 0x1d:
    case 0x1e:
    case 0x1f:
      newval = SANE_FALSE;
      cs9k_option_set_value (s, CS9K_OPTION_SET_FRAME_USER, &newval);
      break;
    case 0x17:
    case 0x18:
      newval = SANE_TRUE;
      cs9k_option_set_value (s, CS9K_OPTION_SET_FRAME_USER, &newval);
      break;
    default:
      DBG (1, "%s: Unrecognized holder id %d.\n", __func__, s->holder_id);
      break;
    }
  return SANE_STATUS_GOOD;
}

static SANE_Status
cs9k_execute (cs9k_t * s)
{
  DBG (16, "%s\n", __func__);

  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "c1 00 00 00 00 00");
  return cs9k_issue_cmd (s);
}

static SANE_Status
cs9k_issue_and_execute (cs9k_t * s)
{
  SANE_Status status;

  DBG (10, "%s, opcode = %02x\n", __func__, s->send_buf[0]);

  status = cs9k_issue_cmd (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  return cs9k_execute (s);
}

static SANE_Status
cs9k_mode_select (cs9k_t * s)
{
  SANE_Status status;

  DBG (4, "%s\n", __func__);

  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);

  cs9k_parse_cmd (s,
		  "15 10 00 00 14 00 00 00 00 08 00 00 00 00 00 00 00 01 03 06 00 00");
  cs9k_pack_word (s, s->unit_dpi);
  cs9k_parse_cmd (s, "00 00");

  status = cs9k_issue_cmd (s);
  if (status && s->sense_code == 0x052c0000)
    {
      /* This is the first non-info command issued */
      DBG (1,
	   "%s: Scanner found in middle of sequence.  Retrying mode select.\n",
	   __func__);
      return cs9k_issue_cmd (s);
    }
  return status;
}

static SANE_Status
cs9k_load (cs9k_t * s)
{
  SANE_Status status;

  DBG (6, "%s\n", __func__);

  if ((s->exe_dx & CS9K_EXE_D1_LOAD) == 0)
    return SANE_STATUS_UNSUPPORTED;
  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "e0 00 d1 00 00 00 00 00 0d 00");
  s->n_send += 13;

  status = cs9k_grow_send_buffer (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  return cs9k_issue_and_execute (s);
}

static SANE_Status
cs9k_eject (cs9k_t * s)
{
  SANE_Status status;

  if ((s->exe_dx & CS9K_EXE_D0_UNLOAD) == 0)
    return SANE_STATUS_UNSUPPORTED;
  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "e0 00 d0 00 00 00 00 00 0d 00");
  s->n_send += 13;

  status = cs9k_grow_send_buffer (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  cs9k_seq_begin (s);
  s->medium_changed = 1;
  return cs9k_issue_and_execute (s);
}

static SANE_Status
cs9k_reset (cs9k_t * s)
{
  SANE_Status status;

  if ((s->exe_8x & CS9K_EXE_80_RESET) == 0)
    return SANE_STATUS_UNSUPPORTED;
  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "e0 00 80 00 00 00 00 00 0d 00");
  s->n_send += 13;

  status = cs9k_grow_send_buffer (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  return cs9k_issue_and_execute (s);
}


static SANE_Status
cs9k_reserve_unit (cs9k_t * s)
{
  DBG (10, "%s\n", __func__);

  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "16 00 00 00 00 00");
  return cs9k_issue_cmd (s);
}

static SANE_Status
cs9k_release_unit (cs9k_t * s)
{
  DBG (10, "%s\n", __func__);

  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "17 00 00 00 00 00");
  return cs9k_issue_cmd (s);
}


static SANE_Status
cs9k_set_focus (cs9k_t * s)
{
  if ((s->exe_cx & CS9K_EXE_C1_FOCUS_MOVE) == 0)
    return SANE_STATUS_UNSUPPORTED;
  DBG (6, "%s: setting focus to %d\n", __func__, s->focusz);

  cs9k_scanner_ready (s, CS9K_STATUS_READY);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "e0 00 c1 00 00 00 00 00 09 00 00");
  cs9k_pack_long (s, s->focusz);
  cs9k_parse_cmd (s, "00 00 00 00");

  return cs9k_issue_and_execute (s);
}

static SANE_Status
cs9k_read_focus (cs9k_t * s)
{
  SANE_Status status;

  if ((s->exe_cx & CS9K_EXE_C1_FOCUS_MOVE) == 0)
    return SANE_STATUS_UNSUPPORTED;
  cs9k_scanner_ready (s, CS9K_STATUS_READY);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "e1 00 c1 00 00 00 00 00 0d 00");
  s->n_recv = 13;

  status = cs9k_issue_cmd (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  s->focusz = cs9k_field2val (s->recv_buf + 1, 4);

  DBG (4, "%s: focus at %d\n", __func__, s->focusz);

  return status;
}

static SANE_Status
cs9k_autofocus (cs9k_t * s)
{
  SANE_Status status;

  if ((s->exe_ax & CS9K_EXE_A0_AF) == 0)
    return SANE_STATUS_INVAL;

  if ((s->old_focusx == s->real_focusx) && (s->old_focusy == s->real_focusy))
    {
      DBG (6, "%s: skipping autofocus - already done at %d,%d\n", __func__,
	   s->old_focusx, s->old_focusy);
      return SANE_STATUS_GOOD;
    }

  DBG (6, "%s: focusing at %d,%d\n", __func__,
       s->real_focusx, s->real_focusy);

  cs9k_convert_options (s);

  status = cs9k_read_focus (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  /* set parameter, autofocus */
  cs9k_scanner_ready (s, CS9K_STATUS_READY);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "e0 00 a0 00 00 00 00 00 09 00 00");
  cs9k_pack_long (s, s->real_focusx);
  cs9k_pack_long (s, s->real_focusy);
  /*cs9k_parse_cmd(s, "00 00 00 00"); */

  status = cs9k_issue_and_execute (s);
  if (status != SANE_STATUS_GOOD)
    return status;
  s->old_focusx = s->real_focusx;
  s->old_focusy = s->real_focusy;
  return cs9k_read_focus (s);
}

static SANE_Status
cs9k_autoexposure (cs9k_t * s, int wb)
{
  SANE_Status status;

  DBG (6, "%s, wb = %d\n", __func__, wb);

  if (wb)
    {
      if ((s->scankinds & CS9K_SCANKIND_AEWBSCAN) == 0)
	{
	  DBG (1, "Autoexposure with white balance not supported.\n");
	  return SANE_STATUS_INVAL;
	}
    }
  else
    {
      if ((s->scankinds & CS9K_SCANKIND_AESCAN) == 0)
	{
	  DBG (1, "Autoexposure not supported.\n");
	  return SANE_STATUS_INVAL;
	}
    }
  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  status = cs9k_scan (s, wb ? CS9K_SCAN_AE_WB : CS9K_SCAN_AE);
  if (status != SANE_STATUS_GOOD)
    return status;

  status = cs9k_get_exposure (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  return status;
}


static SANE_Status
cs9k_get_all_windows (cs9k_t * s)
{
  SANE_Status status;
  SANE_Word tmp_val, window;
  size_t i;

  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "25 00 00 00 00 00 00 00 08 00");
  s->n_recv = 8;
  status = cs9k_issue_cmd (s);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (4,
	   "Error: cs9k_page_inquiry(): Inquiry of window descriptor size failed: %s.\n",
	   sane_strstatus (status));
      return status;
    }
  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  cs9k_init_buffer (s);
  cs9k_parse_cmd (s, "25 00 00 00 00 00");
  cs9k_pack_3byte (s, cs9k_field2val (s->recv_buf, 2) + 2);
  cs9k_parse_cmd (s, "00");
  s->n_recv = cs9k_field2val (s->recv_buf, 2) + 2;
  status = cs9k_issue_cmd (s);
  if (status != SANE_STATUS_GOOD)
    {
      DBG (4,
	   "Error: %s: get_window(all) failed: %s.\n", __func__,
	   sane_strstatus (status));
    }
  else
    for (i = 8; i < s->n_recv; i += cs9k_field2val (s->recv_buf + 6, 2))
      {
	window = cs9k_field2val (s->recv_buf + i + 0, 1);
	if (window < 0 || window >= CS9K_COLOR_MAX)
	  {
	    DBG (4, "Window identifier %d out of range.\n", window);
	    continue;
	  }
	DBG (10,
	     "Window %d: resolution %dx%d, upper left %d,%d, size %dx%d\n",
	     cs9k_field2val (s->recv_buf + i + 0, 1),
	     cs9k_field2val (s->recv_buf + i + 2, 2),
	     cs9k_field2val (s->recv_buf + i + 4, 2),
	     cs9k_field2val (s->recv_buf + i + 6, 4),
	     cs9k_field2val (s->recv_buf + i + 10, 4),
	     cs9k_field2val (s->recv_buf + i + 14, 4),
	     cs9k_field2val (s->recv_buf + i + 18, 4));
	DBG (10, "  Brightness %d, Threshold %d, contrast %d.\n",
	     cs9k_field2val (s->recv_buf + i + 22, 1),
	     cs9k_field2val (s->recv_buf + i + 23, 1),
	     cs9k_field2val (s->recv_buf + i + 24, 1));
	tmp_val = cs9k_field2val (s->recv_buf + i + 25, 1);
	DBG (10,
	     "  Image: %s, %dbpp, Halftone %d, Padding Type = %d, Reverse = %d.\n",
	     tmp_val == 0 ? "bi-level B&W" : tmp_val ==
	     1 ? "dithered/halftone B&W" : tmp_val ==
	     2 ? "multi-level B&W" : tmp_val ==
	     3 ? "bi-level RGB color" : tmp_val ==
	     4 ? "dithered/halftone RGB color" : tmp_val ==
	     5 ? "multi-level RGB color" : "Reserved",
	     cs9k_field2val (s->recv_buf + i + 26, 1),
	     cs9k_field2val (s->recv_buf + i + 27, 2),
	     cs9k_field2val (s->recv_buf + i + 29, 1) & 7,
	     (cs9k_field2val (s->recv_buf + i + 29, 1) >> 7) & 1);
	DBG (10,
	     "  Bit ordering %d, Compression type %d, Compression arg %d, Reserved 0x%08x\n",
	     cs9k_field2val (s->recv_buf + i + 30, 2),
	     cs9k_field2val (s->recv_buf + i + 32, 1),
	     cs9k_field2val (s->recv_buf + i + 33, 1),
	     cs9k_field2val (s->recv_buf + i + 34, 4));
	tmp_val = cs9k_field2val (s->recv_buf + i + 41, 1);
	DBG (10,
	     "  Color order %d, Additional passes %d, %s, setup mode %d%s%s%s%s.\n",
	     cs9k_field2val (s->recv_buf + i + 40, 1) & 15,
	     (cs9k_field2val (s->recv_buf + i + 40, 1) >> 4) & 15,
	     tmp_val & 1 ? "Positive" : "Negative", (tmp_val >> 1) & 7,
	     (tmp_val >> 4) & 1 ? ", reserved bit 4 is set" : "",
	     (tmp_val >> 5) & 1 ? ", filter" : "",
	     (tmp_val >> 6) & 1 ? ", matrix" : "",
	     (tmp_val >> 7) & 1 ? ", averaging" : "");
	tmp_val = cs9k_field2val (s->recv_buf + i + 42, 1);
	DBG (10, "  Scanning Kind: %s%s%s%s%s%s%s%s\n",
	     tmp_val & 1 ? "image scanning, " : "",
	     (tmp_val >> 1) & 1 ? "thumbnail scanning, " : "",
	     (tmp_val >> 2) & 1 ? "set up scanning, " : "",
	     (tmp_val >> 3) & 1 ? "set up scanning2, " : "",
	     (tmp_val >> 4) & 1 ? "reserved bit 4 is set, " : "",
	     (tmp_val >> 5) & 1 ? "auto exposure scanning, " : "",
	     (tmp_val >> 6) & 1 ? "AE with WB scanning, " : "",
	     (tmp_val >> 7) & 1 ? "reserved bit 7 is set" : "");
	tmp_val = cs9k_field2val (s->recv_buf + i + 43, 1);
	DBG (10, "  Scan Mode: %s%s%s%s%s%s%s%s\n",
	     tmp_val & 1 ? "high quality scan, " : "",
	     (tmp_val >> 1) & 1 ? "normal quality scan, " : "",
	     (tmp_val >> 2) & 1 ? "high speed scan, " : "",
	     (tmp_val >> 3) & 1 ? "reserved3 scan, " : "",
	     (tmp_val >> 4) & 1 ? "multiple reading scan, " : "",
	     (tmp_val >> 5) & 1 ? "reserved5 scan, " : "",
	     (tmp_val >> 6) & 1 ? "reverse direction scanning, " : "",
	     (tmp_val >> 7) & 1 ? "reserved7" : "");
	tmp_val = cs9k_field2val (s->recv_buf + i + 44, 1);
	DBG (10, "  Scan Mode: %s%s%s%s%s%s%s%s\n",
	     tmp_val & 1 ? "pixel without CCD distance, " : "",
	     (tmp_val >> 1) & 1 ? "line without CCD distance, " : "",
	     (tmp_val >> 2) & 1 ? "plane, " : "",
	     (tmp_val >> 3) & 1 ? "reserved3, " : "",
	     (tmp_val >> 4) & 1 ? "pixel with CCD distance, " : "",
	     (tmp_val >> 5) & 1 ? "line with CCD distance, " : "",
	     (tmp_val >> 6) & 1 ? "3 line simultaneous reading, " : "",
	     (tmp_val >> 7) & 1 ? "reserved7" : "");
	s->real_exposure[window] = cs9k_field2val (s->recv_buf + i + 46, 4);
	DBG (10, "  AE Target Value: %d, exposure 0x%x (%d)\n",
	     cs9k_field2val (s->recv_buf + i + 45, 1),
	     s->real_exposure[window], s->real_exposure[window]);
      }
  return status;
}

static SANE_Status
cs9k_get_exposure (cs9k_t * s)
{
  SANE_Status status;
  int i_color;
  double exp_low, exp_high, exp_level;

  DBG (6, "%s\n", __func__);

  cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);

  /* Get s->real_exposure[] from device */
  status = cs9k_get_all_windows (s);

  /* Calculate exposure center */
  exp_low = (double) s->real_exposure[cs9k_colors[0]];
  exp_high = (double) s->real_exposure[cs9k_colors[0]];
  for (i_color = 1; i_color < 3; i_color++)
    {
      if ((double) s->real_exposure[cs9k_colors[i_color]] < exp_low)
	exp_low = (double) s->real_exposure[cs9k_colors[i_color]];
      if ((double) s->real_exposure[cs9k_colors[i_color]] > exp_high)
	exp_high = (double) s->real_exposure[cs9k_colors[i_color]];
    }
  exp_level = log2 ((exp_high + exp_low) / 2.0);
  DBG (4, "avg %f, %f => %f\n", exp_low, exp_high, exp_level);

  s->exposure = SANE_FIX (exp_level - s->exposure_zero);
  s->exposure_r = SANE_FIX (log2 ((double) s->real_exposure[cs9k_colors[0]])
			    - exp_level);
  s->exposure_g = SANE_FIX (log2 ((double) s->real_exposure[cs9k_colors[1]])
			    - exp_level);
  s->exposure_b = SANE_FIX (log2 ((double) s->real_exposure[cs9k_colors[2]])
			    - exp_level);
  s->exposure_i = SANE_FIX (log2 ((double) s->real_exposure[cs9k_colors[3]])
			    - s->exposure_zero);

  cs9k_adjustWord2optrange (s, &s->exposure_r, CS9K_OPTION_EXPOSURE_R);
  cs9k_adjustWord2optrange (s, &s->exposure_g, CS9K_OPTION_EXPOSURE_G);
  cs9k_adjustWord2optrange (s, &s->exposure_b, CS9K_OPTION_EXPOSURE_B);
  cs9k_adjustWord2optrange (s, &s->exposure_i, CS9K_OPTION_EXPOSURE_I);

  DBG (4,
       "%s: Exposure 0x%07x 0x%07x 0x%07x 0x%07x -> (%.2f) %.2f %.2f %.2f %.2f\n",
       __func__, s->real_exposure[cs9k_colors[0]],
       s->real_exposure[cs9k_colors[1]], s->real_exposure[cs9k_colors[2]],
       s->real_exposure[cs9k_colors[3]], SANE_UNFIX (s->exposure),
       SANE_UNFIX (s->exposure_r), SANE_UNFIX (s->exposure_g),
       SANE_UNFIX (s->exposure_b), SANE_UNFIX (s->exposure_i));
  return SANE_STATUS_GOOD;
}

/* Keep some digits during integer math */
#define CS9K_SCALE_FIXED (1000)

/* Convert selection on thumbnail to frame selection */
static SANE_Status
cs9k_thumb2real (cs9k_t * s, SANE_Int xmin, SANE_Int xmax, SANE_Int ymin,
		 SANE_Int ymax)
{
  SANE_Int f, sel_lrx, sel_lry, frame_lrx, frame_lry;
  SANE_Int thumb_scalex, thumb_scaley;

  /* Thumb preview and expected preview are likely different shapes.
   * Pick larger scale (multiplier).  The FIXED multiplier is to retain at
   * least a couple of signficant figures.
   */
  thumb_scalex =
    CS9K_SCALE_FIXED * (s->swax_max - s->swax_min + 1) / s->frame_width;
  thumb_scaley =
    CS9K_SCALE_FIXED * (s->sway_max - s->sway_min + 1) / s->frame_height;
  if (thumb_scalex < thumb_scaley)
    thumb_scalex = thumb_scaley;

  /* xmin,ymin reference does not include frame and user offsets */
  s->real_xoffset = xmin * thumb_scalex / CS9K_SCALE_FIXED +
                    s->frame_bounds[0].offset_x +
                    (s->sf_user ? 0 : s->subframe_x / s->unit_mm);
  s->real_yoffset = ymin * thumb_scalex / CS9K_SCALE_FIXED +
                    s->frame_bounds[0].offset_y +
                    (s->sf_user ? 0 : s->subframe_y / s->unit_mm);

  /* User may select area that's part of preview but not thumbnail */
  cs9k_adjustWord2range (&s->real_xoffset, s->swax_min, s->swax_max - 1,
			 "thumb_scale real_xoffset");
  cs9k_adjustWord2range (&s->real_yoffset, s->sway_min, s->sway_max - 1,
			 "thumb_scale real_yoffset");

  s->logical_width = (xmax - xmin + 1) * thumb_scalex /
		     CS9K_SCALE_FIXED / s->real_pitchx;
  s->logical_height = (ymax - ymin + 1) * thumb_scalex /
		     CS9K_SCALE_FIXED / s->real_pitchy;

  /* Determine virtual frame number */
  s->thumb_sel_descr = CS9K_THUMB_SEL_NOFRAME;
  sel_lrx = s->real_xoffset + s->logical_width * s->real_pitchx;
  sel_lry = s->real_yoffset + s->logical_height * s->real_pitchy;

  DBG (12, "Selection %d,%d -> %d,%d scales to %d,%d -> %d,%d\n",
       xmin, xmax, ymin, ymax, s->real_xoffset, s->real_yoffset, sel_lrx,
       sel_lry);

  /* Look for selection completely enclosed by frame */
  for (f = 0; f < s->n_frames; f++)
    {
      frame_lrx =
	s->frame_bounds[f].offset_x + s->frame_bounds[f].width +
	(s->sf_user ? 0 : s->subframe_x / s->unit_mm);
      frame_lry =
	s->frame_bounds[f].offset_y + s->frame_bounds[f].height +
	(s->sf_user ? 0 : s->subframe_y / s->unit_mm);
      if ((s->real_xoffset >= s->frame_bounds[f].offset_x)
	  && (s->real_xoffset < frame_lrx)
	  && (sel_lrx >= s->frame_bounds[f].offset_x) && (sel_lrx < frame_lrx)
	  && (s->real_yoffset >= s->frame_bounds[f].offset_y)
	  && (s->real_yoffset < frame_lry)
	  && (sel_lry >= s->frame_bounds[f].offset_y)
	  && (sel_lry < frame_lry))
	{
	  s->t_frame = f + 1;
	  s->thumb_sel_descr = CS9K_THUMB_SEL_INFRAME;
	  DBG (22, "%s: Selection inside frame %d.\n", __func__, s->t_frame);
	  break;		/* Found a frame completely enclosing selection */
	}
    }

  if (s->thumb_sel_descr == CS9K_THUMB_SEL_NOFRAME)
    /* Look for selection origin enclosed by frame */
    for (f = 0; f < s->n_frames; f++)
      {
	frame_lrx = s->frame_bounds[f].offset_x + s->frame_bounds[f].width;
	frame_lry = s->frame_bounds[f].offset_y + s->frame_bounds[f].height;
	if ((s->real_xoffset >= s->frame_bounds[f].offset_x)
	    && (s->real_xoffset < frame_lrx)
	    && (s->real_yoffset >= s->frame_bounds[f].offset_y)
	    && (s->real_yoffset < frame_lry))
	  {
	    s->t_frame = f + 1;
	    s->thumb_sel_descr = CS9K_THUMB_SEL_CROSSFRAME;
	    break;
	  }
      }

  if (s->source)
    {
      /* ADF mode - if selection is entirely within a window, 
       * translate to current window.  If it crosses the border,
       * use full frame
       */
      if (s->thumb_sel_descr == CS9K_THUMB_SEL_INFRAME)
	{
	  DBG (4, "%s: shifting selection from frame %d to %d\n",
	       __func__, s->t_frame, s->c_frame);
	  s->real_xoffset +=
	    s->frame_bounds[s->c_frame - 1].offset_x -
	    s->frame_bounds[s->t_frame - 1].offset_x;
	  s->real_yoffset +=
	    s->frame_bounds[s->c_frame - 1].offset_y -
	    s->frame_bounds[s->t_frame - 1].offset_y;
	}
      else
	{
	  if (s->thumb_sel_descr == CS9K_THUMB_SEL_CROSSFRAME)
	    {
	      DBG (4, "%s: upper left corner of selection in frame %d\n",
		   __func__, s->t_frame);
	    }

	  /* CS9K_THUMB_SEL_CROSSFRAME or CS9K_THUMB_SEL_NOFRAME */
	  DBG (4,
	       "%s: selection crosses border or not within frame - using frame definition %d\n",
	       __func__, s->c_frame);
	  s->real_xoffset =
	    s->frame_bounds[s->c_frame - 1].offset_x +
	    (s->sf_user ? 0 : s->subframe_x / s->unit_mm);
	  s->real_yoffset =
	    s->frame_bounds[s->c_frame - 1].offset_y +
	    (s->sf_user ? 0 : s->subframe_y / s->unit_mm);
	  s->logical_width =
	    s->frame_bounds[s->c_frame - 1].width / s->real_pitchx;
	  s->logical_height =
	    s->frame_bounds[s->c_frame - 1].height / s->real_pitchy;
	}
    }
  DBG (10, "%s: real_xoffset = %d, real_yoffset = %d\n", __func__,
       s->real_xoffset, s->real_yoffset);
  DBG (10, "%s: logical_width = %d, logical_height = %d\n", __func__,
       s->logical_width, s->logical_height);

  return SANE_STATUS_GOOD;
}

static SANE_Status
cs9k_convert_options (cs9k_t * s)
{
  int i_color;
  SANE_Int xmin, xmax, ymin, ymax;
  SANE_Int thumb_scaled_x, thumb_scaled_y;
  double exp_per_frame;

  DBG (4, "%s\n", __func__);

  s->real_depth = (s->preview ? 8 : s->depth);
  cs9k_adjustWord2list (&s->real_depth, s->depth_list,
			"convert_options real_depth");
  s->bytes_per_pixel = (s->real_depth > 8 ? 2 : 1);
  s->shift_bits = 8 * s->bytes_per_pixel - s->real_depth;

  DBG (12, "%s:   depth = %d, bpp = %d, shift = %d\n", __func__,
       s->real_depth, s->bytes_per_pixel, s->shift_bits);

  if (s->preview)
    if (s->thumbnail)
      {
	s->real_resx = s->res_thumb;
	s->real_resy = s->res_thumb;
      }
    else
      {
	s->real_resx = s->res_preview;
	s->real_resy = s->res_preview;
      }
  else if (s->res_independent)
    {
      s->real_resx = s->resx;
      s->real_resy = s->resy;
    }
  else
    {
      s->real_resx = s->res;
      s->real_resy = s->res;
    }

  s->real_pitchx = s->resx_max / s->real_resx;
  s->real_pitchy = s->resy_max / s->real_resy;

  s->real_resx = s->resx_max / s->real_pitchx;
  s->real_resy = s->resy_max / s->real_pitchy;

  DBG (12, "%s:   resx = %d, resy = %d, pitchx = %d, pitchy = %d\n", __func__,
       s->real_resx, s->real_resy, s->real_pitchx, s->real_pitchy);

  if (s->xmin < s->xmax)
    {
      xmin = s->xmin;
      xmax = s->xmax;
    }
  else
    {
      xmin = s->xmax;
      xmax = s->xmin;
    }

  if (s->ymin < s->ymax)
    {
      ymin = s->ymin;
      ymax = s->ymax;
    }
  else
    {
      ymin = s->ymax;
      ymax = s->ymin;
    }

  /* The prefix "real_" refers to data in device units (1/maxdpi),
   * "logical_" refers to resolution-dependent data.
   */
  s->thumb_padx = 0;
  s->thumb_pady = 0;
  if (s->preview && s->thumbnail)
    {
      DBG (12, "%s: Thumbnail preview - whole strip\n", __func__);
      /* Preview thumbnails - do whole area */
      xmin = s->swax_min;
      xmax = s->swax_max;
      ymin = s->sway_min;
      ymax = s->sway_max;
      s->real_xoffset =
	s->frame_bounds[0].offset_x +
	(s->sf_user ? 0 : s->subframe_x / s->unit_mm);
      s->real_yoffset =
	s->frame_bounds[0].offset_y +
	(s->sf_user ? 0 : s->subframe_y / s->unit_mm);
      cs9k_adjustWord2range (&s->real_xoffset, s->swax_min, s->swax_max,
			     "Thumbnail preview width");
      cs9k_adjustWord2range (&s->real_yoffset, s->sway_min, s->sway_max,
			     "Thumbnail preview height");

      /* Full area at thumbnail resolution */
      s->logical_width = (xmax - xmin + 1) / s->real_pitchx;
      s->logical_height = (ymax - ymin + 1) / s->real_pitchy;

      cs9k_adjustWord2range (&s->logical_width, 1,
			     (s->ccd_width -
			      s->real_xoffset) / s->real_pitchx,
			     "convert_options pad_width");
      cs9k_adjustWord2range (&s->logical_height, 1,
			     (s->sway_max - s->real_yoffset) / s->real_pitchy,
			     "convert_options pad_height");

      /* Each axis if frame shape and other axis is retained */
      thumb_scaled_x = s->logical_height * s->frame_width / s->frame_height;
      thumb_scaled_y = s->logical_width * s->frame_height / s->frame_width;

      s->thumb_padx = thumb_scaled_x - s->logical_width;
      s->thumb_pady = thumb_scaled_y - s->logical_height;

      if (s->thumb_padx > 0)
	s->thumb_pady = 0;
      else
	s->thumb_padx = 0;

      s->logical_width += s->thumb_padx;
      s->logical_height += s->thumb_pady;

      DBG (12, "%s: Padding for thumbnail preview = %d,%d\n", __func__,
	   s->thumb_padx, s->thumb_pady);
    }

  else if (!s->preview && s->thumbnail)
    {
      /* Selection made in thumbnail mode */
      cs9k_thumb2real (s, xmin, xmax, ymin, ymax);
      cs9k_adjustWord2range (&s->logical_width, 1,
			     (s->ccd_width -
			      s->real_xoffset) / s->real_pitchx,
			     "convert_options logical_width");
      cs9k_adjustWord2range (&s->logical_height, 1,
			     (s->sway_max - s->real_yoffset) / s->real_pitchy,
			     "convert_options logical height");
    }

  else
    {
      DBG (12, "%s: User specified window %d,%d - %d,%d, frame %d\n",
	   __func__, s->xmin, s->ymin, s->xmax, s->ymax, s->c_frame);
      if ((s->c_frame > 0) && (s->c_frame <= s->n_frames) && s->frame_bounds)
	{
	  cs9k_adjustWord2range (&xmax, 1,
				 xmin + s->frame_bounds[s->c_frame - 1].width,
				 "convert_options frame_width");
	  cs9k_adjustWord2range (&ymax, 1,
				 ymin + s->frame_bounds[s->c_frame -
							1].height,
				 "convert_options frame_height");
	  s->real_xoffset =
	    s->frame_bounds[s->c_frame - 1].offset_x + xmin +
	    (s->sf_user ? 0 : s->subframe_x / s->unit_mm);
	  s->real_yoffset =
	    s->frame_bounds[s->c_frame - 1].offset_y + ymin +
	    (s->sf_user ? 0 : s->subframe_y / s->unit_mm);
	  cs9k_adjustWord2range (&s->real_xoffset, s->swax_min, s->swax_max,
				 "convert_options real_offsetx");
	  cs9k_adjustWord2range (&s->real_yoffset, s->sway_min, s->sway_max,
				 "convert_options real_offsety");
	  s->logical_width = (xmax - xmin + 1) / s->real_pitchx;	/* XXX use mm units */
	  s->logical_height = (ymax - ymin + 1) / s->real_pitchy;
	  cs9k_adjustWord2range (&s->logical_width, 1,
				 (s->ccd_width -
				  s->real_xoffset) / s->real_pitchx,
				 "convert_options logical_width");
	  cs9k_adjustWord2range (&s->logical_height, 1,
				 (s->sway_max -
				  s->real_yoffset) / s->real_pitchy,
				 "convert_options logical height");
	}
      else
	{
	  DBG (12, "%s: ERROR: frame %d of 1-%d @ %p\n", __func__, s->c_frame,
	       s->n_frames, (void *) &s->frame_bounds[s->c_frame - 1]);
	  return SANE_STATUS_INVAL;
	}
    }

  DBG (12, "%s:   xmin = %d, xmax = %d\n", __func__, xmin, xmax);
  DBG (12, "%s:   ymin = %d, ymax = %d\n", __func__, ymin, ymax);
  DBG (12, "%s:   real_xoffset = %d, real_yoffset = %d\n", __func__,
       s->real_xoffset, s->real_yoffset);

  s->real_width = (s->logical_width - s->thumb_padx) * s->real_pitchx;
  s->real_height = (s->logical_height - s->thumb_pady) * s->real_pitchy;

  DBG (12, "%s:   lw = %d, lh = %d, rw = %d, rh = %d\n", __func__,
       s->logical_width, s->logical_height, s->real_width, s->real_height);

  s->odd_padding = 0;
  if ((s->bytes_per_pixel & 1) && ((s->logical_width - s->thumb_padx) & 0x01)
      && (s->caps & CS9K_CAPS_PAD_WORD))
    s->odd_padding = 1;

  if (s->focus_on_centre)
    {
      if (s->preview && s->thumbnail)
	{
	  /* Choose frame 1 */
	  s->real_focusx = (s->framex_pos + s->frame_width) / 2;
	  s->real_focusy = (s->framey_pos + s->frame_height) / 2;
	}
      else
	{
	  s->real_focusx = s->real_xoffset + s->real_width / 2;
	  s->real_focusy = s->real_yoffset + s->real_height / 2;
	}
    }
  else
    {
      s->real_focusx = s->frame_bounds[s->c_frame - 1].offset_x + s->focusx +
	(s->sf_user ? 0 : s->subframe_x / s->unit_mm);
      s->real_focusy = s->frame_bounds[s->c_frame - 1].offset_y + s->focusy +
	(s->sf_user ? 0 : s->subframe_y / s->unit_mm);
    }
  cs9k_adjustWord2range (&s->real_focusx, 0, s->frame_width,
			 "convert_options real_focusx");
  cs9k_adjustWord2range (&s->real_focusy, 0, s->frame_height,
			 "convert_options real_focusy");

  DBG (12, "%s:   focusx = %d, focusy = %d\n", __func__, s->real_focusx,
       s->real_focusy);

  if (s->source && s->thumbnail && 
      (s->frame_bounds[s->c_frame - 1].flags & CS9K_FBF_PRESENT) &&
      (fabs(s->frame_bounds[s->c_frame - 1].adj_exp) <= s->exposure_per_frame))
    {
      exp_per_frame = SANE_UNFIX(s->frame_bounds[s->c_frame - 1].exposure);
      DBG (4,
	"%s: Using per-frame exposure of %.2f instead of %.2f set point.\n",
	__func__, exp_per_frame, SANE_UNFIX(s->exposure));
    }
  else
    exp_per_frame = SANE_UNFIX(s->exposure);

  s->real_exposure[1] = (SANE_Word) pow (2.0, s->exposure_zero +
					 exp_per_frame +
					 SANE_UNFIX (s->exposure_r));
  s->real_exposure[2] = (SANE_Word) pow (2.0, s->exposure_zero +
					 exp_per_frame +
					 SANE_UNFIX (s->exposure_g));
  s->real_exposure[3] = (SANE_Word) pow (2.0, s->exposure_zero +
					 exp_per_frame +
					 SANE_UNFIX (s->exposure_b));
  s->real_exposure[9] = (SANE_Word) pow (2.0, s->exposure_zero +
					 SANE_UNFIX (s->exposure_i));

  for (i_color = 0; i_color < 4; i_color++)
    cs9k_adjustWord2range (&s->real_exposure[cs9k_colors[i_color]],
			   s->exposure_min, s->exposure_max,
			   "convert_options exposure");

  if (s->scancolors & CS9K_CLRCMP_RGB)
    {
      s->n_colors_in = 3;
    }
  else if (s->scancolors & CS9K_CLRCMP_GRAYSCALE)
    {
      s->n_colors_in = 1;
    }
  else
    {
      DBG (4, "%s:   RGB support not indicated; using it anyway.\n",
	   __func__);
      s->n_colors_in = 3;
    }
  s->n_colors_out = s->n_colors_in;

  /* Do infrared even in preview to enable IR exposure analysis */
  switch (s->infrared)
    {
      case CS9K_INFRARED_4PLANE:
	s->n_colors_in++;
	s->n_colors_out++;
	break;

      case CS9K_INFRARED_IN:
	s->n_colors_in++;
	break;

      case CS9K_INFRARED_OUT:
	s->n_colors_in = 0;
	s->n_colors_out = 1;
	break;

      case CS9K_INFRARED_ONLY:
	s->n_colors_in++;
	s->n_colors_out = 1;
	break;

      default:
	break;
    }

  DBG (12, "%s:   Colors %d:%d, Exposure 0x%06d 0x%06d 0x%06d 0x%06d\n",
       __func__, s->n_colors_in, s->n_colors_out,
       s->real_exposure[1], s->real_exposure[2],
       s->real_exposure[3], s->real_exposure[9]);

  return SANE_STATUS_GOOD;
}

static SANE_Status
cs9k_send_lut (cs9k_t * s)
{
  int color;
  SANE_Status status;
  cs9k_pixel_t *lut, pixel;

  DBG (6, "%s\n", __func__);

  if (!s->lut_r)
    {
      DBG (4, "%s: LUT unsupported - skipping.\n", __func__);
      return SANE_STATUS_GOOD;
    }

  for (color = 0; color < s->n_colors_in; color++)
    {
      cs9k_scanner_ready(s, CS9K_STATUS_READY);

      switch (color)
        {
        case 0:
          lut = s->lut_r;
          break;
        case 1:
          lut = s->lut_g;
          break;
        case 2:
          lut = s->lut_b;
          break;
        case 3:
          lut = s->lut_neutral;
          break;
        default:
          DBG (1,
      	 "BUG: %s: Unknown color number for LUT download.\n",
      	 __func__);
          return SANE_STATUS_INVAL;
          break;
        }

      cs9k_init_buffer (s);
      cs9k_parse_cmd (s, "2a 00 03 00");
      cs9k_pack_byte (s, cs9k_colors[color]);
      cs9k_pack_byte (s, 2 - 1);	/* XXX number of bytes per data point - 1 */
      cs9k_pack_byte (s, ((2 * s->n_lut) >> 16) & 0xff);	/* XXX 2 bytes per point */
      cs9k_pack_byte (s, ((2 * s->n_lut) >> 8) & 0xff);	/* XXX 2 bytes per point */
      cs9k_pack_byte (s, (2 * s->n_lut) & 0xff);	/* XXX 2 bytes per point */
      cs9k_pack_byte (s, 0x00);

      for (pixel = 0; pixel < s->n_lut; pixel++)
        {			/* XXX 2 bytes per point */
          cs9k_pack_word (s, lut[pixel]);
        }

      status = cs9k_issue_cmd (s);
      if (status != SANE_STATUS_GOOD)
        return status;
    }

  return status;
}

static SANE_Status
cs9k_set_window (cs9k_t * s, cs9k_scan_t type)
{
  int color, colors;
  SANE_Status status = SANE_STATUS_INVAL;

  DBG (4, "%s: %d colors in\n", __func__, s->n_colors_in);

  colors = (s->scancolors & CS9K_CLRCMP_DROPOUT) ? 4 : 3;
  /* SET WINDOW */
  for (color = 0; color < colors; color++)
    {

      DBG (8, "%s: window %d, type %d\n", __func__, cs9k_colors[color], type);

      cs9k_scanner_ready (s, CS9K_STATUS_READY);

      cs9k_init_buffer (s);
      /* Some devices require vendor specific bits always be 10b? */
      /* Coolscan 9000 requires it be set for dropout channel. */
      if ((s->caps & CS9K_CAPS_CTL_VS10) || (color == 3))
	cs9k_parse_cmd (s, "24 00 00 00 00 00 00 00 3a 80");
      else
	cs9k_parse_cmd (s, "24 00 00 00 00 00 00 00 3a 00");

      cs9k_parse_cmd (s, "00 00 00 00 00 00 00 32");

      cs9k_pack_byte (s, cs9k_colors[color]);

      cs9k_pack_byte (s, 0x00);

      cs9k_pack_word (s, s->real_resx);
      cs9k_pack_word (s, s->real_resy);
      cs9k_pack_long (s, s->real_xoffset);
      cs9k_pack_long (s, s->real_yoffset);
      cs9k_pack_long (s, s->real_width);
      cs9k_pack_long (s, s->real_height);

      cs9k_pack_byte (s, 0x00);	/* brightness, etc. */
      cs9k_pack_byte (s, 0x00);
      cs9k_pack_byte (s, 0x00);
      cs9k_pack_byte (s, 0x05);	/* image composition CCCCCCC */
      cs9k_pack_byte (s, s->real_depth);	/* pixel composition */
      cs9k_parse_cmd (s, "00 00 00 00 00 00 00 00 00 00 00 00 00");
      cs9k_pack_byte (s, 0x00);	/* multiread, ordering */

      cs9k_pack_byte (s, 0x80 | (s->negative ? 0 : 1));	/* averaging, pos/neg */

      switch (type)
	{			/* scanning kind */
	case CS9K_SCAN_NORMAL:
	  if (s->real_resx < s->resx_list[s->resx_list[0]])
	    cs9k_pack_byte (s, 0x02);	/* Thumbnail scan */
	  else
	    cs9k_pack_byte (s, 0x01);
	  break;
	case CS9K_SCAN_AE:
	  cs9k_pack_byte (s, 0x20);
	  break;
	case CS9K_SCAN_AE_WB:
	  cs9k_pack_byte (s, 0x40);
	  break;
	default:
	  DBG (1, "BUG: cs9k_scan(): Unknown scanning type.\n");
	  return SANE_STATUS_INVAL;
	}
      cs9k_pack_byte (s, 0x02);	/* scanning mode */
      cs9k_pack_byte (s, 0x02);	/* color interleaving */
      cs9k_pack_byte (s, 0xff);	/* (ae) */

      DBG (4, "%s: exposure = %d * 10ns\n", __func__,
	   s->real_exposure[cs9k_colors[color]]);
      cs9k_pack_long (s, s->real_exposure[cs9k_colors[color]]);

      status = cs9k_issue_cmd (s);

      if (status != SANE_STATUS_GOOD)
	return status;

    }	/* for each color */
  return status;
}

static SANE_Status
cs9k_scan (cs9k_t * s, cs9k_scan_t type)
{
  SANE_Status status;

  s->block_padding = 0;

  DBG (6, "%s, type = %d, colors = %d:%d\n", __func__, type,
       s->n_colors_in, s->n_colors_out);

  switch (type)
    {
    case CS9K_SCAN_NORMAL:
      DBG (16, "%s: normal scan\n", __func__);
      break;
    case CS9K_SCAN_AE:
      if ((s->scankinds & CS9K_SCANKIND_AESCAN) == 0)
	{
	  DBG (1, "Autoexposure not supported.\n");
	  return SANE_STATUS_INVAL;
	}
      DBG (16, "%s: ae scan\n", __func__);
      break;
    case CS9K_SCAN_AE_WB:
      if ((s->scankinds & CS9K_SCANKIND_AEWBSCAN) == 0)
	{
	  DBG (1, "Autoexposure with white balance not supported.\n");
	  return SANE_STATUS_INVAL;
	}
      DBG (16, "%s: ae wb scan\n", __func__);
      break;
    }

  /* wait for device to be ready with document, and set device unit */
  status = cs9k_scanner_ready (s, CS9K_STATUS_NO_DOCS);
  if (status != SANE_STATUS_GOOD)
    return status;

  if (s->status & CS9K_STATUS_NO_DOCS)
    return SANE_STATUS_NO_DOCS;

  status = cs9k_convert_options (s);
  if (status != SANE_STATUS_GOOD)
    return status;

  if (s->preview && s->thumbnail && s->thumbnail_done)
    DBG (4, "%s: Using data in thumbnail buffer.\n", __func__);
  else if (s->infrared == CS9K_INFRARED_OUT)
    DBG (4, "%s: Using data in infrared buffer.\n", __func__);
  else
    {
      cs9k_set_focus (s);

      cs9k_scanner_ready (s, CS9K_STATUS_READY);

      if (type == CS9K_SCAN_NORMAL)
	cs9k_send_lut (s);

      status = cs9k_set_window (s, type);
      if (status != SANE_STATUS_GOOD)
	return status;

      status = cs9k_get_exposure (s);
      if (status != SANE_STATUS_GOOD)
	return status;

      cs9k_scanner_ready (s, CS9K_STATUS_READY);

      cs9k_init_buffer (s);
      switch (s->n_colors_in)
	{
	case 1:
	  if (s->infrared == CS9K_INFRARED_ONLY)
	    cs9k_parse_cmd (s, "1b 00 00 00 04 00 01 02 03 09");
	  else
	    cs9k_parse_cmd (s, "1b 00 00 00 01 00 00");
	  break;
	case 3:
	  cs9k_parse_cmd (s, "1b 00 00 00 03 00 01 02 03");
	  break;
	case 4:
	  cs9k_parse_cmd (s, "1b 00 00 00 04 00 01 02 03 09");
	  break;
	default:
	  DBG (0, "BUG: %s: Unsupported number of input colors.\n", __func__);
	  break;
	}

      status = cs9k_issue_cmd (s);
      if (status != SANE_STATUS_GOOD)
	{
	  DBG (6, "scan setup failed\n");
	  return status;
	}

      if (s->status == CS9K_STATUS_REISSUE)
	{
	  status = cs9k_issue_cmd (s);
	  if (status != SANE_STATUS_GOOD)
	    return status;
	}
    }

  return SANE_STATUS_GOOD;
}

/* get exposure advice */
SANE_Status
cs9k_analyze_histogram (SANE_Int ** hists, SANE_Int layers,
			SANE_Int levels, double edge_target,
			cs9k_hist_alg_t * type, SANE_Fixed * offset)
{
  SANE_Int layer, level, total, threshold, count, lcount, edge, max;
  double adjust;
  SANE_Int tcount, l;

  *type = 0;
  if (!hists)
    {
      DBG (8, "%s called with null histogram.\n", __func__);
      return SANE_STATUS_INVAL;
    }

  DBG (10, "%s: Analyzing %d wide x %d plane; histogram base at %p.\n",
       __func__, levels, layers, (void *) hists);

  max = levels - 1;

  /* Count the total number of pixel elements */
  total = 0;
  for (layer = 0; layer < layers; layer++)
    {
      if (!hists[layer])
	return SANE_STATUS_INVAL;
      DBG (12, "%s: Layer %d histogram at %p.\n", __func__, layer,
	   (void *) (hists[layer]));
      for (level = 0; level < levels; level++)
	total += hists[layer][level];
    }
  if (!total)
    {
      DBG (4, "%s: Histogram contains no data.\n", __func__);
      return SANE_STATUS_INVAL;
    }

  threshold = total / 512;	/* 
				 * Set point is no more than 0.2% of pixels
				 * should be brighter than <edge_target> stops
				 * below saturation
				 */

  /* Check for highly saturated scan - likely no film */
  count = 0;
  for (layer = 0; layer < layers; layer++)
    count += hists[layer][max];
  if (count > (total / 2))
    {
      *offset = SANE_FIX (0.0);
      *type = CS9K_HIST_ALG_HIGH_SAT;
    }

  /* Find high exposure edge */
  count = 0;
  edge = max;
  for (level = max; level >= 0; level--)
    {
      if (count < threshold)
	edge = level;
      else
	break;
      for (layer = 0; layer < layers; layer++)
	count += hists[layer][level];
    }

  if (0)	/* Print raw histogram data */
    for (level = 0; level < levels; level++)
      {
	fprintf (stderr, "%5.2f", (double) level / (double) HIST_STEPS);
	for (layer = 0; layer < layers; layer++)
	  fprintf (stderr, "\t%6d", hists[layer][level]);
	if (edge == level)
	  {
	    tcount = 0;
	    for (l = level; l < levels; l++)
	      for (layer = 0; layer < layers; layer++)
		tcount += hists[layer][l];
	    fprintf (stderr,
		     " <== %4.1f%% of pixels (%d/%d) at or above %.3f",
		     (double) tcount / (double) total * 100.0, tcount, total,
		     (double) edge / HIST_STEPS);
	  }
	fprintf (stderr, "\n");
      }

  if (*type == 0)
   {
      /* Determine exposure adjustment */
      if (edge > 0)
	{
	  if (edge == max)
	    {
	      /* Overexposed - try from low side */
	      lcount = 0;
	      edge = 0;
	      for (level = 0; level <= max; level++)
		{
		  if (lcount < threshold)
		    edge = level;
		  else
		    break;
		  for (layer = 0; layer < layers; layer++)
		    lcount += hists[layer][level];
		}

	      if (edge == 0)
		{
		  /* over and under - balance the saturation */
		  adjust = log2 ((double) lcount / (double) count);
		  DBG (20, "%s: Both ends filled - compromise at %.2f.\n",
		       __func__, adjust);
		*type = CS9K_HIST_ALG_HIGH_RANGE;
		}
	      else
		{
		  /* Have some footroom */
		  adjust = -((double) edge - 16.0) / HIST_STEPS + edge_target;
		  DBG (20, "%s: Found %.2f of room to decrease exposure.\n",
		       __func__, adjust);
		  if (adjust > 0.0)
		    adjust = 0.0;
		*type = CS9K_HIST_ALG_LEDGE_DET;
		}
	    }
	  else
	    {
	      adjust =
		((double) (max) - (double) edge) / HIST_STEPS - edge_target;
	      DBG (1,
		   "%s: Edge of high exposure = %.2f,  Target = %.2f.  Adjust exposure by %.2f stops\n",
		   __func__, (double) edge / HIST_STEPS,
		   (double) max / HIST_STEPS - edge_target, adjust);
	      *type = CS9K_HIST_ALG_HEDGE_DET;
	 }
	  *offset = SANE_FIX (adjust);
	}
      else
	{				/* All pixels dark? */
	  *offset = SANE_FIX (5.0);
	*type = CS9K_HIST_ALG_UNDEREXPOSED;
	}
    }
  return SANE_STATUS_GOOD;
}

/* Draw on thumbnails */
void
cs9k_set_pixel (cs9k_t * s, SANE_Int x, SANE_Int y, SANE_Int color)
{
  unsigned offs;

  offs = (y * s->logical_width + x) * s->n_colors_out * s->bytes_per_pixel;


  if (offs < s->xfer_bytes_total)
    switch (s->bytes_per_pixel)
      {
      case 1:
	memcpy (s->thumbnail_obuf + offs, bordercolors1[color], 3);
	break;
      case 2:
	memcpy (s->thumbnail_obuf + offs, bordercolors2[color], 6);
	break;
      }
  else
    DBG (1,
	 "%s: Pixel at %d,%d is out of range (width = %d, colors = %d, total size = %d\n",
	 __func__, x, y, s->logical_width, s->n_colors_out,
	 s->bytes_per_pixel);
}

/* Draw frame boundaries on thumbnail view and extract exposure information. */
SANE_Status
cs9k_draw_bounds (cs9k_t * s)
{
  SANE_Int x, y, ps, n, ulx, uly, lrx, lry;
  size_t offs;
  SANE_Int pixelcolor, c, xmargin, ymargin;
  double exp_err, exp_cnt;

  if (s->thumbnail_done)
    memcpy (s->thumbnail_obuf, s->thumbnail_ibuf, s->xfer_bytes_total);
  else
    memcpy (s->thumbnail_obuf + s->thumb_pos_out,
	    s->thumbnail_ibuf + s->thumb_pos_out,
	    s->thumb_pos_in - s->thumb_pos_out);

  exp_err = 0.0;
  exp_cnt = 0.0;
  for (n = 0; n < s->n_frames; n++)
    {
      pixelcolor = (s->negative ? 2 : 0) + (n & 1);
      ulx =
	(s->frame_bounds[n].offset_x -
	 s->frame_bounds[0].offset_x) * s->res_thumb / s->unit_dpi;
      uly =
	(s->frame_bounds[n].offset_y -
	 s->frame_bounds[0].offset_y) * s->res_thumb / s->unit_dpi;
      lrx = ulx + (s->frame_bounds[n].width) * s->res_thumb / s->unit_dpi;
      lry = uly + (s->frame_bounds[n].height) * s->res_thumb / s->unit_dpi;
      cs9k_adjustWord2range (&ulx, 0, s->logical_width - s->thumb_padx - 1,
			     "draw_bounds ul-x");
      cs9k_adjustWord2range (&uly, 0, s->logical_height - s->thumb_pady - 1,
			     "draw_bounds ul-y");
      cs9k_adjustWord2range (&lrx, 0, s->logical_width - s->thumb_padx - 1,
			     "draw_bounds lr-x");
      cs9k_adjustWord2range (&lry, 0, s->logical_height - s->thumb_pady - 1,
			     "draw_bounds lr-y");

      for (y = uly; y <= lry; y += lry - uly)
	for (x = ulx; x <= lrx; x++)
	  cs9k_set_pixel (s, x, y, pixelcolor);

      for (x = ulx; x <= lrx; x += lrx - ulx)
	for (y = uly; y <= lry; y++)
	  cs9k_set_pixel (s, x, y, pixelcolor);

      if (s->thumbnail_done)
	{
	  s->histogram_size = s->real_depth * HIST_STEPS;
	  if (s->histogram_size > s->histogram_alloc)
	    {
	      DBG (1, "%s: Need %d histogram entries, have only %d.\n",
		   __func__, s->histogram_size, s->histogram_alloc);
	      s->histogram_size = 0;
	    }
	  else
	    {
	      ps = s->n_colors_out * s->bytes_per_pixel;
	      memset (s->histogram[3], 0,
		      s->histogram_size * sizeof (SANE_Int));
	      xmargin = (lrx - ulx) / HIST_FRAME_MARGIN;
	      ymargin = (lry - uly) / HIST_FRAME_MARGIN;
	      for (x = ulx + xmargin; x < lrx - xmargin; x++)
		for (y = uly + ymargin; y < lry - ymargin; y++)
		  for (c = 0;
		       c < (s->n_colors_out * s->bytes_per_pixel);
		       c += s->bytes_per_pixel)
		    {
		      offs = (y * s->logical_width + x) * ps;
		      switch (s->bytes_per_pixel)
			{
			case 1:
			  s->
			    histogram[3][(int)
					 (log2
					  (*(u_int8_t *)
					   (s->thumbnail_ibuf + offs +
					    c)) * HIST_STEPS)]++;
			  break;
			case 2:
			  s->
			    histogram[3][(int)
					 (log2
					  (*(u_int16_t *)
					   (s->thumbnail_ibuf + offs +
					    c)) * HIST_STEPS)]++;
			  break;
			}
		    }

	      cs9k_analyze_histogram (s->histogram + 3, 1, s->histogram_size,
				      SANE_UNFIX (s->adj_exp_tgt),
				      &s->frame_bounds[n].
				      adj_exp_type,
				      &s->frame_bounds[n].adj_exp);
	      if (s->frame_bounds[n].adj_exp_type == CS9K_HIST_ALG_HIGH_SAT)
		{
		  DBG (1,
		       "%s: Frame %2d: Over half of pixels saturated; marking frame non-existent.\n",
		       __func__, n + 1);
		  s->frame_bounds[n].flags &= ~CS9K_FBF_PRESENT;
		}
	      else
		{
		  if ((s->frame_bounds[n].adj_exp_type ==
		       CS9K_HIST_ALG_HEDGE_DET)
		      || (s->frame_bounds[n].adj_exp_type ==
			  CS9K_HIST_ALG_LEDGE_DET))

		    {
		      /* Only include good estimates in average */
		      exp_err += SANE_UNFIX (s->frame_bounds[n].adj_exp);
		      exp_cnt += 1.0;
		    }
		  DBG (10,
		       "%s: Frame %d, should be adjusted %.2f stops (type %d); sum = %.2f, count = %f\n",
		       __func__, n + 1,
		       SANE_UNFIX (s->frame_bounds[n].adj_exp),
		       s->frame_bounds[n].adj_exp_type, exp_err, exp_cnt);
		  s->frame_bounds[n].exposure = s->exposure +
			                        s->frame_bounds[n].adj_exp;
		  s->frame_bounds[n].flags |= CS9K_FBF_PRESENT;
		}
	      s->frame_bounds[n].flags |= CS9K_FBF_EXAMINED;
	    }

	  if ((s->frame_bounds[n].flags & CS9K_FBF_EXAMINED)
	      && ((s->frame_bounds[n].flags & CS9K_FBF_PRESENT) == 0))
	    {
	      /* Draw an X over frames that will be skipped in ADF mode */
	      if ((lrx - ulx) > (lry - uly))
		/* Wide frame */
		for (x = ulx; x <= lrx; x++)
		  {
		    cs9k_set_pixel (s, x,
				    uly + (x - ulx) * (lry - uly + 1) / (lrx -
									 ulx +
									 1),
				    pixelcolor);
		    cs9k_set_pixel (s, x,
				    lry - (x - ulx) * (lry - uly + 1) / (lrx -
									 ulx +
									 1),
				    pixelcolor);
		  }
	      else
		/* Tall frame */
		for (y = uly; y <= lry; y++)
		  {
		    cs9k_set_pixel (s,
				    ulx + (y - uly) * (lrx - ulx) / (lry -
								     uly), y,
				    pixelcolor);
		    cs9k_set_pixel (s,
				    lrx - (y - uly) * (lrx - ulx) / (lry -
								     uly), y,
				    pixelcolor);
		  }
	    }
	}
    }

  if (s->thumb_padx)
    {
      for (x = s->logical_width - s->thumb_padx; x < s->logical_width; x++)
	{
	  cs9k_set_pixel (s, x, 0, 0);
	  cs9k_set_pixel (s, x, s->logical_height - 1, 0);
	}

      ulx = s->logical_width - s->thumb_padx;
      lrx = s->logical_width - 1;
      for (y = 0; y < s->logical_height; y++)
	{
	  cs9k_set_pixel (s, ulx, y, 0);
	  cs9k_set_pixel (s, lrx, y, 0);
	  if (s->thumbnail_done)
	    {
	      for (x = ulx + 31 - (y & 31); x < lrx; x += 32)
		{
		  cs9k_set_pixel (s, x, y, 0);
		}
	    }
	  else
	    {
	      cs9k_set_pixel (s, ulx + y * s->thumb_padx / s->logical_height,
			      y, 0);
	      cs9k_set_pixel (s, lrx - y * s->thumb_padx / s->logical_height,
			      y, 0);
	    }
	}
    }

  if (exp_cnt > 0.5)
    {
      s->adj_exp = SANE_FIX (exp_err / exp_cnt);
      s->adj_exp_type = CS9K_HIST_ALG_MULTI_FRAME;
      cs9k_set_adjust_message (s,
			       CS9K_OPTION_ADJUST_EXPOSURE,
			       s->adj_exp, s->adj_exp_type);
      DBG (20, "%s: Multi-frame adjustment %.2f\n", __func__,
	   SANE_UNFIX (s->adj_exp));
    }

  if (s->thumbnail_done)
    if (s->infrared == CS9K_INFRARED_IN ||
	s->infrared == CS9K_INFRARED_4PLANE ||
	s->infrared == CS9K_INFRARED_OUT)
      {
	DBG (4, "%s: Start Infrared analysis\n", __func__);
	memset (s->histogram[3], 0, s->histogram_size * sizeof (SANE_Int));
	for (n = 0; n < s->n_frames; n++)
	  {
	    if ((s->frame_bounds[n].flags & CS9K_FBF_EXAMINED)
		&& (s->frame_bounds[n].flags & CS9K_FBF_PRESENT))
	      {
		ulx =
		  (s->frame_bounds[n].offset_x -
		   s->frame_bounds[0].offset_x) * s->res_thumb / s->unit_dpi;
		uly =
		  (s->frame_bounds[n].offset_y -
		   s->frame_bounds[0].offset_y) * s->res_thumb / s->unit_dpi;
		lrx =
		  ulx +
		  (s->frame_bounds[n].width) * s->res_thumb / s->unit_dpi;
		lry =
		  uly +
		  (s->frame_bounds[n].height) * s->res_thumb / s->unit_dpi;
		cs9k_adjustWord2range (&ulx, 0,
				       s->logical_width - s->thumb_padx - 1,
				       "draw_bounds ul-x");
		cs9k_adjustWord2range (&uly, 0,
				       s->logical_height - s->thumb_pady - 1,
				       "draw_bounds ul-y");
		cs9k_adjustWord2range (&lrx, 0,
				       s->logical_width - s->thumb_padx - 1,
				       "draw_bounds lr-x");
		cs9k_adjustWord2range (&lry, 0,
				       s->logical_height - s->thumb_pady - 1,
				       "draw_bounds lr-y");

		xmargin = (lrx - ulx) / HIST_FRAME_MARGIN;
		ymargin = (lry - uly) / HIST_FRAME_MARGIN;
		for (x = ulx + xmargin; x < lrx - xmargin; x++)
		  for (y = uly + ymargin; y < lry - ymargin; y++)
		    {
		      offs = (y * s->logical_width + x) * s->bytes_per_pixel;
		      switch (s->bytes_per_pixel)
			{
			case 1:
			  s->histogram[3][(int)
					  (log2
					   (*(u_int8_t *)
					    (s->infrared_buf + offs +
					     c)) * HIST_STEPS)]++;
			  break;
			case 2:
			  s->
			    histogram[3][(int)
					 (log2
					  (*(u_int16_t *)
					   (s->infrared_buf + offs +
					    c)) * HIST_STEPS)]++;
			  break;
			}
		    }

		cs9k_analyze_histogram (s->histogram + 3, 1,
					s->histogram_size,
					SANE_UNFIX (s->adj_exp_tgt_ir),
					&s->adj_exp_type_ir, &s->adj_exp_ir);
		cs9k_set_adjust_message (s, CS9K_OPTION_ADJUST_EXPOSURE_IR,
					 s->adj_exp_ir, s->adj_exp_type_ir);
		DBG (20, "%s: Multi-frame IR adjustment %.2f\n", __func__,
		     SANE_UNFIX (s->adj_exp_ir));
	      }
	  }
      }
  return SANE_STATUS_GOOD;
}

/* Return to origin frame, non-IR frame, etc. */
SANE_Status
cs9k_seq_begin (cs9k_t * s)
{
  int i;
  s->thumbnail_done = 0;
  s->thumb_pos_in = 0;
  s->infrared_index = 0;
  if (s->infrared == CS9K_INFRARED_OUT)
    s->infrared = CS9K_INFRARED_IN;
  s->c_frame = s->i_frame;	/* Reset current frame pointer */
  for (i = 0; i < s->n_frames; i++)
    {
      s->frame_bounds[i].adj_exp = SANE_FIX (0.0);
      s->frame_bounds[i].adj_exp_type = 0;
      s->frame_bounds[i].flags = 0;
    }
  return SANE_STATUS_GOOD;
}

SANE_Status
cs9k_set_adjust_message (cs9k_t * s, SANE_Int option,
			 SANE_Fixed val, SANE_Word adjtype)
{
  char *dest;
  double mult;

  switch (option)
    {
    case CS9K_OPTION_ADJUST_EXPOSURE:
      dest = s->adjust_exposure_title;
      s->option_flags |=
	cs9k_set_visible (s, option, (fabs (SANE_UNFIX (val)) > 0.015));
      mult = (s->negative ? -1.0 : 1.0);
      break;
    case CS9K_OPTION_ADJUST_EXPOSURE_IR:
      dest = s->adjust_exposure_ir_title;
      s->option_flags |=
	cs9k_set_visible (s, option, (fabs (SANE_UNFIX (val)) > 0.015)
			  && (s->infrared != CS9K_INFRARED_OFF));
      s->option_flags |=
	cs9k_set_visible (s, CS9K_OPTION_ADJUST_EXPOSURE_TEXT_IR,
			  s->infrared != CS9K_INFRARED_OFF);
      mult = (s->negative ? -1.0 : 1.0);
      break;
    default:
      return SANE_STATUS_INVAL;
    }
  snprintf (dest, CS9K_ADJUST_EXPOSURE_STRING_MAX,
	    "Adjust %s exposure %.2f stops [%s]",
	    (option == CS9K_OPTION_ADJUST_EXPOSURE) ? "base" : (option ==
								CS9K_OPTION_ADJUST_EXPOSURE_IR)
	    ? "infrared" : "unknown", mult * SANE_UNFIX (val),
	    (adjtype ==
	     CS9K_HIST_ALG_UNDEREXPOSED) ? "Very Underexposed" : (adjtype ==
								  CS9K_HIST_ALG_HIGH_RANGE)
	    ? "Both edges saturated" : (adjtype ==
					CS9K_HIST_ALG_HIGH_SAT) ?
	    "Very Overexposed" : (adjtype ==
				  CS9K_HIST_ALG_HEDGE_DET) ? "High Edge"
	    : (adjtype == CS9K_HIST_ALG_LEDGE_DET) ? "Low Edge" : (adjtype ==
								   CS9K_HIST_ALG_MULTI_FRAME)
	    ? "Multi-frame average" : (adjtype ==
				       0) ? "Unset" : "Bad programmer");
  DBG (22, "%s: Set option %d text to '%s'\n", __func__, option, dest);
  return SANE_STATUS_GOOD;
}

void *
cs9k_xmalloc (size_t size)
{
  register void *value = malloc (size);
  if (value == NULL)
    {
      DBG (0, "error: %s: failed to malloc() %lu bytes.\n",
	   __func__, (unsigned long) size);
    }
  return value;
}

static void *
cs9k_xrealloc (void *p, size_t size)
{
  register void *value;
  if (!size)
    return p;
  value = realloc (p, size);
  if (value == NULL)
    {
      DBG (0, "error: %s: failed to realloc() %lu bytes.\n",
	   __func__, (unsigned long) size);
    }

  return value;
}

static void
cs9k_xfree (const void *p)
{
  free (p);
}

static void
cs9k_decode_sense (const unsigned long sk,
		   const unsigned long asc,
		   const unsigned long ascq,
		   const unsigned long info, const unsigned long sks)
{
  switch (sk)
    {
    case 0x00:			/* No sense */
      return;
    case 0x02:
      switch (asc)
	{
	case 0x04:
	  switch (ascq)
	    {
	    case 0x01:
	      DBG (10, "Logical Unit is in process of becoming ready\n");
	      switch (info)
		{
		case 0x00:
		  DBG (10,
		       "...during the execution of the operation activation command\n");
		  return;
		case 0x01:
		  DBG (10,
		       "...during loading/ejection of the object to be scanned\n");
		  return;
		case 0x02:
		  DBG (10,
		       "...during the measurement of the correction data\n");
		  return;
		default:
		  return;
		}
	    case 0x02:
	      DBG (10, "Logical Unit not ready, cause not reportable\n");
	      return;
	    case 0x03:
	      DBG (10,
		   "Logical Unit not ready, manual intervention required\n");
	      switch (info)
		{
		case 0x06:
		  DBG (10, "...FH-869GR: The mask is not set\n");
		  return;
		case 0x07:
		  DBG (10, "...Undefined holder\n");
		  return;
		default:
		  return;
		}
	    default:
	      return;
	    }			/* ascq for asc == 0x04 */
	case 0x05:
	  switch (ascq)
	    {
	    case 0x00:
	      switch (info)
		{
		case 0x00:
		  DBG (10, "Logical Unit does not respond to selection\n");
		  return;
		default:
		  return;
		}
	    default:
	      return;
	    }
	default:
	  return;
	}
    case 0x05:			/* Sense Key == 5 */
      switch (asc)
	{
	case 0x24:
	  switch (ascq)
	    {
	    case 0x00:
	      switch (info)
		{
		case 0x00:
		  DBG (10, "Invalid field in CDB\n");
		  if (sks & 0x800000)
		    DBG (10, "Field pointer: Byte %ld, %sbit %ld in %s\n",
			 (long) ((sks >> 16) & 7),
			 sks & 0x080000 ? "" : "invalid ",
			 (long) (sks & 0xffff),
			 sks & 0x400000 ? "CDB" : "Data");
		  return;
		default:
		  return;
		}
	    default:
	      return;
	    }			/* ascq for asc == 0x04 */
	case 0x25:
	  switch (ascq)
	    {
	    case 0x00:
	      switch (info)
		{
		case 0x00:
		  DBG (10,
		       "Logical Unit not supported; a LUN other than 0 was specified\n");
		  return;
		default:
		  return;
		}
	    default:
	      return;
	    }			/* ascq for asc == 0x04 */
	case 0x26:
	  switch (ascq)
	    {
	    case 0x00:
	      switch (info)
		{
		case 0x00:
		  DBG (10, "Invalid field in parameter list\n");
		  if (sks & 0x800000)
		    DBG (10, "Field pointer: Byte %ld, %sbit %ld in %s\n",
			 (long) ((sks >> 16) & 7),
			 sks & 0x080000 ? "" : "invalid ",
			 (long) (sks & 0xffff),
			 sks & 0x400000 ? "CDB" : "Data");
		  return;
		default:
		  return;
		}
	    default:
	      return;
	    }			/* ascq for asc == 0x04 */
	case 0x2c:
	  switch (ascq)
	    {
	    case 0x00:
	      DBG (10, "Command Sequence Error\n");
	      return;
	    case 0x02:
	      DBG (10, "Invalid combination of windows specified\n");
	      return;
	    default:
	      return;
	    }

	case 0x06:		/* Sense Key == 6 */
	  switch (asc)
	    {
	    case 0x28:
	      switch (ascq)
		{
		case 0x00:
		  switch (info)
		    {
		    case 0x00:
		      DBG (10,
			   "Not ready to ready change, medium may have changed\n");
		      return;
		    default:
		      return;
		    }
		default:
		  return;
		}
	    case 0x29:
	      switch (ascq)
		{
		case 0x00:
		  switch (info)
		    {
		    case 0x00:
		      DBG (10,
			   "Power on, reset, or bus device reset occurred\n");
		      return;
		    default:
		      return;
		    }
		default:
		  return;
		}
	    case 0x3f:
	      DBG (10, "Target operating conditions have changed\n");
	      switch (ascq)
		{
		case 0x03:
		  DBG (10, " - Inquiry data has changed\n");
		  switch (info)
		    {
		    default:
		      return;
		    }
		case 0x04:
		  DBG (10, " - 04 specific ???\n");
		  switch (info)
		    {
		    default:
		      return;
		    }
		default:
		  return;
		}
	    default:
	      return;
	    }

	case 0x0b:		/* Sense Key == b */
	  switch (asc)
	    {
	    case 0x4b:
	      switch (ascq)
		{
		case 0x00:
		  switch (info)
		    {
		    case 0x00:
		      DBG (10,
			   "Data Phase error - unexpected error during data phase\n");
		      return;
		    default:
		      return;
		    }
		default:
		  return;
		}
	    case 0x4e:
	      switch (ascq)
		{
		case 0x00:
		  switch (info)
		    {
		    case 0x00:
		      DBG (10,
			   "Overlapped commands attempted - the unit is selected by the same initiator while disconnecting\n");
		      return;
		    default:
		      return;
		    }
		default:
		  return;
		}
	    default:
	      return;
	    }
	default:
	  return;
	}
    }
}

/* vim: set noet cino=>2: */
